#############################################################################
# !make
#
# !Makefile name: MOD_PR09.mk
#
# description: This is the makefile for MOD_PR09
#
# !Variables		
#   INC           Include directories
#   LIB           Library directories
#   OBJ           Object files
#   TARGET        Target executable
#   ADD_CFLAGS    Additional compiler settings, specified in this make
#   INC_FILES     Additional include files
#
# !Environment variables
#   BRAND         system type (sgi32, sgi64, or linux)
#   CC		  Compiler environment variable
#   CFLAGS        Additional compiler settings in the environment
#   HDFINC        Include directory of HDF set by the script
#   HDFLIB        Library of HDF set by the script
#   HDFEOS_INC    Include directory of HDFEOS set by the script
#   HDFEOS_LIB    Library of HDFEOS set by the script
#   MAKE          make utility
#   PGSINC        Include directory of PGS toolkit set by the script
#   PGSLIB        Library of PGS toolkit set by the script
#
# !Team-Unique Header:
#   MODIS Land Science Team for the National 
#   Aeronautics and Space Administration, Goddard Space Flight 
#   Center, under NASA Task 92-012-05.
#
# !Revision History:
# 	Original development: October 2002	
#		Jim Ray (SSAI/NASA's GSFC Code 922)
# 
#	Revision: 
# 
# !END
#############################################################################

# Compiler options are set based on the environment.
target:
	case "$(BRAND)" in \
          sgi32 | sgi64 ) \
            $(MAKE) $(TARGET) -f MOD_PR09.mk \
                    ADD_CFLAGS="-O2 -DIRIS4 -D_POSIX_SOURCE -DOLDSDSNAMES " \
                    ADD_FFLAGS="-O2 -DIRIS4 -D_POSIX_SOURCE" \
                    LIB="-L$(HDFLIB) -lmfhdf -ldf -ljpeg -lz \
                         -L$(PGSLIB) -lPGSTK \
                         -L$(HDFEOS_LIB) -lhdfeos \
			 -L$(HDFEOS5_LIB) -lhe5_hdfeos -lGctp \
                         -L$(HDF5LIB) -lhdf5 \
                         -lm -lftn " \
            ;; \
          linux ) \
            $(MAKE) $(TARGET) -f MOD_PR09.mk \
                    ADD_CFLAGS="-O0 -Df2cFortran -DLINUX -D_POSIX_SOURCE -DOLDSDSNAMES "\
                    ADD_FFLAGS="-O0 -D_POSIX_SOURCE -Minform,inform" \
                    LIB="-L$(PGSLIB) -lPGSTK \
                         -L$(HDFLIB) -lmfhdf -ldf -ljpeg -lz \
                         -L$(HDFEOS_LIB) -lhdfeos \
			 -L$(HDFEOS5_LIB) -lhe5_hdfeos -lGctp -lm \
                         -L$(HDF5LIB) -lhdf5  \
			 -L$(PGI)/linux86/lib -lpgc -lpgftnrtl " \
            ;; \
          hp ) \
            $(MAKE) $(TARGET) -f MOD_PR09.mk \
                    ADD_CFLAGS="-DHPUX -D_POSIX_SOURCE -DOLDSDSNAMES " \
                    ADD_FFLAGS="-DHPUX -D_POSIX_SOURCE" \
                    LIB="-L$(PGSLIB) -lPGSTK \
                         -L$(HDFLIB) -lmfhdf -ldf -lz \
                         -L$(HDFEOS_LIB) -lhdfeos -lGctp -lm -lf" \
            ;; \
          * ) \
            echo " ERROR: Environmental variable BRAND (set to  \"$(BRAND)\") not recognized." >&2 ; \
            exit 1 ;; \
        esac



INC = -I$(HDFINC) -I$(HDF5INC) -I$(HDFEOS_INC) -I$(PGSINC)
VPATH = $(HDFINC):$(HDF5INC):$(HDFEOS_INC):$(PGSINC)

OBJ = modis_srefl.o modis_srefl_io.o meta.o param.o atmcor.o grib.o smf.o ancillary.o utils.o polarization.o sra.o atmos.o sunglint.o aerosols.o ids.o qa.o hdflib.o newLUT_subr.o 

TARGET = MOD_PR09.exe

$(TARGET): $(OBJ)
	$(CC) $(CFLAGS) $(ADD_CFLAGS) -o $(TARGET) $(OBJ) $(LIB) 

modis_srefl.o:  modis_srefl.h protos.h param.h  global.h struct.h versions.h
modis_srefl_io.o:  modis_srefl.h protos.h struct.h
meta.o:  meta.h struct.h versions.h
param.o:  modis_srefl.h param.h struct.h 
atmcor.o:  modis_srefl.h protos.h struct.h global.h 
hdflib.o:  modis_srefl.h protos.h struct.h global.h 
aerosols.o:  modis_srefl.h protos.h struct.h
qa.o:  modis_srefl.h protos.h struct.h
grib.o:  grib.h 
smf.o:  smf.h 
utils.o:  smf.h 
ancillary.o:  meta.h protos.h
ids.o:  ids.h protos.h

clean:
	rm -f $(OBJ) core.* $(TARGET)
.c.o:
	$(CC) $(CFLAGS) $(ADD_CFLAGS) $(INC) -c $< -o $@

.f.o:
	$(F77) $(F77FLAGS) $(ADD_FFLAGS) -c $< -o $@
