#include <ctype.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "mfhdf.h"
#include "modis_srefl.h"
#include "protos.h"
#ifdef LINUX
#include <byteswap.h>
#endif
/* maximum # of dimensions in any file we will be opening */
#define HDF_MAXRANK 3

void AOT_interpolate(uint8 *qa, int16 *data, uint8 *fail, int n_row, int n_col, int TIERMAX);
int16 *downsample(int16 *data, int *n_row, int *n_col, int factor);
uint8 *downsampleqa(uint8 *qa, int *n_row, int *n_col, int factor);
void upsample(int16 *data, uint8 *qa, int n_lines, int n_cols, int16 *avg_data, int factor);

#ifdef NBSPOTS_1KM
#define COLWIDTH NBSPOTS_1KM
#else
#define COLWIDTH 1354
#endif

/* This is in global.h */
extern aerosol_retr_rej_t aotr_struct;

#ifdef SUBSET
extern float32 T2M_value, NCEPH2O_value;
#endif

/* New, 09-NOV-06 */
#define LATLIMIT 89.0

enum {REFFILE=1, GEOFILE, CRSREFFILE, FILETYPES};
#ifndef SUBSET
enum {B31, B32, WVAP, QA, CM, SZ, SA, VZ, VA, LATT, LONT, T2M, NCEPWV, PARAMS};
#else
enum {B31, B32, WVAP, QA, CM, SZ, SA, VZ, VA, LATT, LONT, PARAMS};
#endif
typedef struct{
   int32 sds;    /* initialized to the -(file source) value */
   int32 dims[2], type;
   void *data;
   char sdsname[50];
} hdf_param;
hdf_param parameters[PARAMS] = {
#ifndef OLDSDSNAMES 
    /* Use these for the rewrite only */
    { -REFFILE, {-1, -1}, DFNT_UINT16, (void *)NULL, "1km Brightness Temperature Band 31\0"},
    { -REFFILE, {-1, -1}, DFNT_UINT16, (void *)NULL, "1km Brightness Temperature Band 32\0"},
    { -REFFILE, {-1, -1}, DFNT_UINT16, (void *)NULL, "1km Water Vapor\0"},
#else
   /*  and use these for the old MOD_PR09 code, or for purposes of comparison to products of old code:  */
    { -REFFILE, {-1, -1}, DFNT_UINT16, (void *)NULL, "BAND31\0"},
    { -REFFILE, {-1, -1}, DFNT_UINT16, (void *)NULL, "BAND32\0"},
    { -REFFILE, {-1, -1}, DFNT_UINT16, (void *)NULL, "1km water_vapor\0"},
#endif    
    { -REFFILE, {-1, -1}, DFNT_UINT16, (void *)NULL, "1km Reflectance Data State QA\0"},
    { -REFFILE, {-1, -1}, DFNT_UINT16, (void *)NULL, "1km Atmospheric Optical Depth Band QA\0"}, /* 5.0.2 */
    { -GEOFILE, {-1, -1}, DFNT_INT16,  (void *)NULL, "SolarZenith\0"},
    { -GEOFILE, {-1, -1}, DFNT_INT16,  (void *)NULL, "SolarAzimuth\0"},
    { -GEOFILE, {-1, -1}, DFNT_INT16,  (void *)NULL, "SensorZenith\0"},
    { -GEOFILE, {-1, -1}, DFNT_INT16,  (void *)NULL, "SensorAzimuth\0"},
    { -GEOFILE, {-1, -1}, DFNT_FLOAT,  (void *)NULL, "Latitude\0"},
    { -GEOFILE, {-1, -1}, DFNT_FLOAT,  (void *)NULL, "Longitude\0"}
#ifndef SUBSET
    ,{ -CRSREFFILE, {-1, -1}, DFNT_UINT16, (void *)NULL, "Coarse Resolution Air Temperature (2m)\0"},
    { -CRSREFFILE, {-1, -1}, DFNT_UINT8,  (void *)NULL, "Coarse Resolution NCEP Water Vapor\0"}
#endif
    };
void *get_image(int32 id, int32 *dims, int32 type, int *size);

int aot_filter(int refout, dims_struct ***outputs)
/*
!C******************************************************************

!Description: This function filters retrieved aerosols.

!Input Parameters:
refout:	              the maximum number of files with aerosols to be interpolated
ref_product_file:     a list of files with the aerosols.  Some of these strings may be
                      NULL-valued; these are to be skipped (obviously).


!Output Parameters:
none; returns 0 for success, nonzero for error.


!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), 07-JUN-01.


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
     This routine is only meant to filter aerosols generated with routine 
     aerosol_retrieval(), in file atmcor.c.  1-km versions of the retrieved aerosols
     are written to MOD09 files during the aerosol-retrieval part of PGE11 (the first 
     pass).  These are to be read from the MOD09 files (listed in ref_product_files)
     and filtered.  
     
     This routine looks for an SDS with a "cm" in its title, and filters the rest
     of the SDSs according to the values found in the "cm" SDS.
     
     This filtration occurs on a per-granule basis.  Algorithm is like this:
       for all pixels where AOT band 1 > AOT band 3,
           if pixel is within 20 pixels of snow, set AOT1 and AOT3 to fill value, set qa.

!END

*****************************************************************************/
{
#define WITHINX 20
#define FILL    10
int WITHIN;
int ifile,i,j,eye,ii,jj,yey,snowy;
int got_qa, got_cm, got_band1, got_band3, n_sds, n_lines, n_cols;
int32 sd, sds;
int32 start[2], dims[2];
int32 n_sets1, n_gattr1, rank1, dims1[2], number_type1, nattr1;
char msg[256];
char name[256];
char location[100]="aerosols.c, aot_filter\0";
int16 *band1, *band3;
uint16 *cm;
uint8 *qa, *subpixel_cloud;
char *ref_product_file;

for (ifile=0;ifile<refout;ifile++) {
   got_qa = 0;
   got_cm = 0;
   got_band1 = 0;
   got_band3 = 0;
   n_sds = 0;
   if (outputs[0][ifile]->filename != NULL) {
      ref_product_file = (char *)malloc(strlen(outputs[0][ifile]->filename)+1);
      strcpy(ref_product_file, outputs[0][ifile]->filename);
      }
   else ref_product_file = (char *)NULL;
   
   if (ref_product_file != NULL) {

       if ( (sd = SDstart(ref_product_file, DFACC_RDWR)) == -1) {
	  sprintf(msg, "'%s' (aerosol optical thickness)", ref_product_file);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
	  return(1);
       }
      SDfileinfo(sd, &n_sets1, &n_gattr1);
      if (n_sets1 == 0) {
	  sprintf(msg,  "file %s (aerosol optical thickness) is unreadable ", ref_product_file);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
          SDend(sd);
	  return(1);
       }
      
      /* get the QA, band1 aot, band3 aot */
      for (j=0;j<n_sets1;j++) {
          sds = SDselect(sd, j);
          if (sds == -1) {
	     sprintf(msg,  "file %s (aerosol optical thickness) has unreadable SDS ", ref_product_file);
	     modsmf(MODIS_E_FILE_ACCESS, msg, location);
             SDend(sd);
	     return(1);
          }
          SDgetinfo(sds, name, &rank1, dims1, &number_type1, &nattr1);
 	  if (strstr(name, "1km Atmospheric Optical Depth Band QA")) {  /* read it */ /* 5.0.2 */
	     got_cm = 1;
             if (  (cm = (uint16 *)malloc(dims1[0]*dims1[1]*sizeof(uint16))) == NULL)  {
                 sprintf(msg,  " cannot allocate %d uint16's (%d by %d) ", 
	                      dims1[0]*dims1[1], dims1[0], dims1[1]);
                 modsmf(MODIS_E_MEM_FAILURE, msg, location);
                 SDendaccess(sds); SDend(sd); return(4);
                 }
             if ( HDF_Read_2D(sds, 0, dims1[0], dims1[1], (VOID *)cm) == 0) {
                 SDendaccess(sds); SDend(sd); return(4);
	     }
	  }
 	  else if (strstr(name, "1km Atmospheric Optical Depth Band CM"))  {  /* read it */  /* 5.0.2 */
	     got_qa = 1;
             if (  (qa = (uint8 *)malloc(dims1[0]*dims1[1]*sizeof(uint8))) == NULL)  {
                 sprintf(msg,  " cannot allocate %d uint8's (%d by %d) ", 
	                      dims1[0]*dims1[1], dims1[0], dims1[1]);
                 modsmf(MODIS_E_MEM_FAILURE, msg, location);
                 SDendaccess(sds); SDend(sd); return(4);
                 }
             if ( HDF_Read_2D(sds, 0, dims1[0], dims1[1], (VOID *)qa) == 0) {
                 SDendaccess(sds); SDend(sd); return(4);
	     }
	  }
 	  else if (strstr(name, "1km Atmospheric Optical Depth Band 1")) {  /* read it */
	     got_band1 = 1;
             if (  (band1 = (int16 *)malloc(dims1[0]*dims1[1]*sizeof(int16))) == NULL)  {
                 sprintf(msg,  " cannot allocate %d int16's (%d by %d) ", 
	                      dims1[0]*dims1[1], dims1[0], dims1[1]);
                 modsmf(MODIS_E_MEM_FAILURE, msg, location);
                 SDendaccess(sds); SDend(sd); return(4);
                 }
             if ( HDF_Read_2D(sds, 0, dims1[0], dims1[1], (VOID *)band1) == 0) {
                 SDendaccess(sds); SDend(sd); return(4);
	     }
	  }
 	  else if (strstr(name, "1km Atmospheric Optical Depth Band 3")) {  /* read it */
	     got_band3 = 1;
             if (  (band3 = (int16 *)malloc(dims1[0]*dims1[1]*sizeof(int16))) == NULL)  {
                 sprintf(msg,  " cannot allocate %d int16's (%d by %d) ", 
	                      dims1[0]*dims1[1], dims1[0], dims1[1]);
                 modsmf(MODIS_E_MEM_FAILURE, msg, location);
                 SDendaccess(sds); SDend(sd); return(4);
                 }
             if ( HDF_Read_2D(sds, 0, dims1[0], dims1[1], (VOID *)band3) == 0) {
                 SDendaccess(sds); SDend(sd); return(4);
	     }
	  }
	  SDendaccess(sds);
      }  /* for (j=0;j<n_sets1;j++) */

     if (  (subpixel_cloud = (uint8 *)calloc(dims1[0]*dims1[1],sizeof(uint8))) == NULL)  {
          sprintf(msg,  " cannot allocate %d uint8's (%d by %d) ", dims1[0]*dims1[1], dims1[0], dims1[1]);
          modsmf(MODIS_E_MEM_FAILURE, msg, location);return(4);
      }


      if ((!got_qa)||(!got_cm)||(!got_band1)||(!got_band3)) {
	  sprintf(msg,  "file %s (aerosol optical thickness) is missing QA, CM, AOT_BAND1 or AOT_BAND3 ", ref_product_file);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
          SDend(sd);
	  return(1);
      }
      else {  /* filter */
	   /* Try to get this working more efficiently */
	   
	   WITHIN=WITHINX;
	   for(i=WITHIN;i<dims1[0]-WITHIN;i++) {
	       for(j=WITHIN;j<dims1[1]-WITHIN;j++) {
		  eye=i*dims1[1] + j;
		  if ((band1[eye] > band3[eye])&&(band1[eye] > 200)) {
		      /*printf("%s %d %d ", ref_product_file, i,j);*/
		      
		      /* See if there is a snowy pixel within WITHIN of here */
		      snowy=0;
		      for(ii=i-WITHIN;ii<i+WITHIN;ii++) {
		         for(jj=j-WITHIN;jj<j+WITHIN;jj++) {
		             yey=ii*dims1[1] + jj;
			     if ( (cm[yey]&0x0010)>>4 == 1 ) {
			         snowy++;
				 break;
			     }
		         }
			 if (snowy) break;
		      }
		      if (snowy) {
		         band1[eye] = FILL;
			 band3[eye] = FILL;
			 qa[eye] = 8;
		         /*printf(" SNOWY = %d\n", snowy);*/
		      }
		      /*else   printf("\n");*/
		      
		  }
	       }
	    }
	    
           /* 28-SEP-01: subpixel cloud removal */
	   WITHIN=1;
	   for(i=WITHIN;i<dims1[0]-WITHIN;i++) {
	       for(j=WITHIN;j<dims1[1]-WITHIN;j++) {
		  eye=i*dims1[1] + j;
		  if ((cm[eye]&0x0080 == 0) &&   /* 23-OCT-01: don't take dust as subpixel cloud! */
		      (band3[eye] > 700)) {
		      snowy=0;
		      for(ii=i-WITHIN;ii<i+WITHIN;ii++) {
		         for(jj=j-WITHIN;jj<j+WITHIN;jj++) {
		             yey=ii*dims1[1] + jj;
			     if ( (yey != eye) && (qa[yey] == 0)) {   /* assume qa[yey] == 0 means good retrieval */
			         snowy++;
				 /*break;*/
			     }
		         }
			 /*if (snowy) break;*/
		      }
		      if (snowy == 0)   /* if not even one good retrieval... */
		        subpixel_cloud[eye] = 1;    /* it's a cloud */

		  }
	    
	       }
	    } 
	     
	   for(eye=0;eye<dims1[0]*dims1[1];eye++) 
	      if (subpixel_cloud[eye] == 1) qa[eye] = 12;

       }
       free(subpixel_cloud);
       
       /* now write qa, band1, band3 back to where they started */
       for (j=0;j<n_sets1;j++) {
          sds = SDselect(sd, j);
          SDgetinfo(sds, name, &rank1, dims1, &number_type1, &nattr1);
 	  if (strstr(name, "1km Atmospheric Optical Depth Band 1")) {  /* write it */
              HDF_Write_2D(sds, 0, dims1[0], dims1[1], (VOID *)band1);
              free(band1);
	  }
 	  else if (strstr(name, "1km Atmospheric Optical Depth Band CM")) {  /* write it */ /* 5.0.2 */
              HDF_Write_2D(sds, 0, dims1[0], dims1[1], (VOID *)qa);
              free(qa);
	  }
 	  else if (strstr(name, "1km Atmospheric Optical Depth Band 3")) {  /* write it */
              HDF_Write_2D(sds, 0, dims1[0], dims1[1], (VOID *)band3);
              free(band3);
	  }
	  SDendaccess(sds);
       }
       free(cm);
       SDend(sd);
       
       free(ref_product_file);   
   }
   
      }  /* for (ifile=0;ifile<refout;ifile++) */

return(0);
}

int cld_shadow_mask(int refout, dims_struct ***inputs, dims_struct ***outputs)
/*
!C******************************************************************

!Description: This function gets cloud shadow from the MOD09 products.

!Input Parameters:
refout:	              the maximum number of files
ref_product_file:     a list of files with the cloudmask/angle data.  Some of these 
                      strings may be NULL-valued; these are to be skipped (obviously).
geo_for_ref_product_file:     the geo files that match the ref_product_files.
crsref_product_file:          the CRS files that match the ref_product_files.

!Output Parameters:
none; returns 0 for success, nonzero for error.


!Revision History:

Original version
   From PFII code, file cloud_c.c, routine shadowmask(): adapted for MODIS March, 2002,
   J. Ray (jim@cratmos.gsfc.nasa.gov).


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes     
     This cloud shadow mask determination is done on a full swath: i. e., all 
     granules are read in sequence into a swath, which is then interpolated.
     Required MOD09 SDSs: band31, band32, 1km water_vapor, 1km Surface Reflectance Band 1,
       1km Surface Reflectance Band 2, 1km Surface Reflectance Band 3, State QA.
     Required MOD03 SDSs: solar zenith, solar azimuth, sensor zenith, sensor azimuth,
       lat, lon.
     Required MOD09CRS SDSs: Coarse Resolution Air Temperature (2m), Coarse Resolution NCEP 
       Water Vapor.
     
!END

*****************************************************************************/
{
#include "PGS_PC.h"
#define ERAD 6378.137 /* km */
#define TWOPI 6.28318530717958647692
#define NGRAN 30
float *ts_clear;
int i,j,ii,m, pixs, lorbit, line, pixel;
int n_sds, n_lines, total_lines, n_cols, ds_lines, ds_cols, row, col, nbsamp;
int l,k;
int kk,ll,X;
int newindex;
int32 sd, sds;
int32 start[2], dims[2];
int32 n_sets1, n_gattr1, rank1, dims1[2], number_type1, nattr1;
int32 each_files_lines[NGRAN], max_lines;
int gridheight,gridwidth;
float64 fts,ffs,ftv,ffv,flat,flon,flat_save,flon_save;
float ch1,ch2,ch3,t4,t5,ndvi,tbs,tempv;
float tscld,*tsclear,z,zz;
float64 tmpflt64;
float t,v;
float conv_factor;
double dx,dy,dlat,dlon;
uint16 *cm, intcm, testintcm;
char geofile[256];
char msg[256];
char name[256];
char location[50]="aerosols.c, cld_shadow_mask\0";
int16 *b1, *b2, *b3;
float *latitude, *longitude, *tmpflt;
int16 *solarzenith, *solarazimuth, *sensorzenith, *sensorazimuth;
float wv, zz2;
uint16 *b31, *b32, *qa, *t2mp, *cloudmask;
uint16 *wvp;
uint8 *ncepp;
float *resampled_ncepp;
uint16 *tmpuint16;
float *wv_clear;
int32 sd_srcfile[NGRAN][FILETYPES];
int32 sdid, ret;
int n_lines_in_orbit, num_lines[NGRAN], OK_to_process[NGRAN];
uint16 *orbitstateqa;
float *orbitlatitude, *orbitlongitude;
int16 iterations;
int16 *all_cloudy;
char *ref_product_file[20];
char *crsref_product_file[20];
char *geo_for_ref_product_file[20];

int xxx = 0;

void geo_sat(float *lat_buffer, float *lon_buffer, int total_lines, int in_line, int in_pix,
             float outlat, float outlon, int *out_line, int *out_pix, int16 *iterations);
int get_filename_from_input(char *MOD09_filename, char *ESDT, char *outfilename);
void get_parameters(int32 sdid, hdf_param *parameter, int nlines);
int get_parameter_info(int32 sdid, hdf_param parameter, int *n_lines);
void get_orbital_data_float(int32 sdid, hdf_param parameter, float *space, int *n_lines, int8 line_increment);
void get_orbital_data_int(int32 sdid, hdf_param parameter, uint16 *space, int *n_lines, int8 line_increment);

n_cols = 0;
max_lines = 0;
n_lines = 0;

modsmf(MODIS_A_CLDSHD_START, " ", location);   /* New, 23-APR-04 */

/* Make daaammmed-sure we aren't going to process more granules than we can handle (NGRAN or 30) 
 ************************************************************************************************/
if (refout >= NGRAN) { 
   sprintf(msg, "Files from more than %d granules have been staged; will limit processing to the first %d granules\0", NGRAN, NGRAN);
   refout = NGRAN-1;
   modsmf(MODIS_W_FUNCTION_ERROR, msg, location);
          }


/* First pass: make sure all parameters are there, get the number of lines in 'em 
 ***********************************************************************************/
n_lines_in_orbit = 0;
for (i=0;i<refout;i++) {

   OK_to_process[i] = 0;
   num_lines[i] = 0;
   if (outputs[0][i]->filename != NULL) {
      ref_product_file[i] = (char *)malloc(strlen(outputs[0][i]->filename)+1);
      strcpy(ref_product_file[i], outputs[0][i]->filename);
      }
   else ref_product_file[i] = (char *)NULL;
   if (outputs[1][i]->filename != NULL) {
      crsref_product_file[i] = (char *)malloc(strlen(outputs[1][i]->filename)+1);
      strcpy(crsref_product_file[i], outputs[1][i]->filename);
      }
   else crsref_product_file[i] = (char *)NULL;
   if (inputs[GEO_FILE][i]->filename != NULL) {
      geo_for_ref_product_file[i] = (char *)malloc(strlen(inputs[GEO_FILE][i]->filename)+1);
      strcpy(geo_for_ref_product_file[i], inputs[GEO_FILE][i]->filename);
      }
   else geo_for_ref_product_file[i] = (char *)NULL;

   /*printf("***** %s %s %s\n", crsref_product_file[i], ref_product_file[i], geo_for_ref_product_file[i] );*/
   
   for(j=REFFILE;j<FILETYPES;j++) sd_srcfile[i][j] = -1;


   if (crsref_product_file[i] != NULL) { 
       if ( (sd_srcfile[i][CRSREFFILE] = SDstart(crsref_product_file[i], DFACC_RDONLY)) == -1) {
	  sprintf(msg, "'%s' (aerosol optical thickness)\0", crsref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
	  continue;
          }
       }
#ifndef SUBSET
   else continue;
#endif
   if (ref_product_file[i] != NULL) { 
       if ( (sd_srcfile[i][REFFILE] = SDstart(ref_product_file[i], DFACC_RDONLY)) == -1) {
	  sprintf(msg, "'%s' (aerosol optical thickness)\0", ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
          SDend(sd_srcfile[i][CRSREFFILE]);
	  continue;
          }
       }
   else continue;
   if (geo_for_ref_product_file[i] != NULL) { 
       if ( (sd_srcfile[i][GEOFILE] = SDstart(geo_for_ref_product_file[i], DFACC_RDONLY)) == -1) {
	  sprintf(msg, "'%s' (aerosol optical thickness)\0", geo_for_ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
          SDend(sd_srcfile[i][CRSREFFILE]);
          SDend(sd_srcfile[i][REFFILE]);
	  continue;
          }
       }
   else continue;
   ret = 0;
   for(j=0;j<PARAMS;j++) {
         ret += get_parameter_info(sd_srcfile[i][-(parameters[j].sds)], parameters[j], &n_lines); 
	 if (j == 0) {
	     n_lines_in_orbit += n_lines;
	     num_lines[i] = n_lines;
	     }
	 }
   if (ret == 0)  OK_to_process[i] = 1;

   /* presumably, if we are HERE, then these steps will WORK... */
   SDend(sd_srcfile[i][CRSREFFILE]);
   SDend(sd_srcfile[i][REFFILE]);
   SDend(sd_srcfile[i][GEOFILE]);

}


/* Allocate memory for a whole swath of latitude, longitude, and StateQA.  Assume COLWIDTH columns.  
 **************************************************************************************************/
orbitlatitude = (float *)malloc(n_lines_in_orbit*COLWIDTH*sizeof(float));
orbitlongitude = (float *)malloc(n_lines_in_orbit*COLWIDTH*sizeof(float));
orbitstateqa = (uint16 *)malloc(n_lines_in_orbit*COLWIDTH*sizeof(uint16));
if ((orbitlatitude == NULL)||(orbitlongitude == NULL)||(orbitstateqa == NULL)){
    sprintf(msg, "Memory problems, cannot allocate %d by %d floats", n_lines_in_orbit, COLWIDTH);
    modsmf(MODIS_E_MEM_FAILURE, msg, location);
    return(4);
     }

/* Assemble the lats/lons/stateQAs into an orbit 
 ************************************************/
n_lines = 0;
for (i=0;i<refout;i++) {
     if (OK_to_process[i]) {
        sdid = SDstart(geo_for_ref_product_file[i], DFACC_RDONLY);
        get_orbital_data_float(sdid, parameters[LATT], orbitlatitude, &n_lines, 0); 
        get_orbital_data_float(sdid, parameters[LONT], orbitlongitude, &n_lines, 0); 
        SDend(sdid);
        sdid = SDstart(ref_product_file[i], DFACC_RDONLY);
        get_orbital_data_int(sdid, parameters[QA], orbitstateqa, &n_lines, 1); 
        SDend(sdid);
        }
     }  /* for (i=1;i<=refout;i++)) */

total_lines = n_lines;

/* Now read all the per-granule stuff and DO THE CLOUD SHADOW DETECTION 
 *************************************************************************/
n_lines = 0;
for (i=0;i<refout;i++) {

    if (OK_to_process[i]) {
    
    if (crsref_product_file[i] != NULL) { 
       if ( (sd_srcfile[i][CRSREFFILE] = SDstart(crsref_product_file[i], DFACC_RDONLY)) == -1) {
	  sprintf(msg, "'%s' (aerosol optical thickness)\0", crsref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
	  return(1);  /* this is REAL bad, get out */
          }
       }
    if (ref_product_file[i] != NULL) { 
       if ( (sd_srcfile[i][REFFILE] = SDstart(ref_product_file[i], DFACC_RDONLY)) == -1) {
	  sprintf(msg, "'%s' (aerosol optical thickness)\0", ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
	  return(1);
          }
       }
    if (geo_for_ref_product_file[i] != NULL) { 
       if ( (sd_srcfile[i][GEOFILE] = SDstart(geo_for_ref_product_file[i], DFACC_RDONLY)) == -1) {
	  sprintf(msg, "'%s' (aerosol optical thickness)\0", geo_for_ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
	  return(1);
          }
       }
        
    for(j=0;j<PARAMS;j++) {
        get_parameters(sd_srcfile[i][-(parameters[j].sds)], &(parameters[j]), num_lines[i]);    
	}
	
    SDend(sd_srcfile[i][CRSREFFILE]);
    SDend(sd_srcfile[i][REFFILE]);
    SDend(sd_srcfile[i][GEOFILE]);

    b31           = (uint16 *)parameters[B31].data;
    b32           = (uint16 *)parameters[B32].data;
    wvp           = (uint16 *)parameters[WVAP].data;
    qa            = (uint16 *)parameters[QA].data;
    cloudmask     = (uint16 *)parameters[CM].data;
    latitude      = (float *)parameters[LATT].data;
    longitude     = (float *)parameters[LONT].data;
    solarzenith   = (int16 *)parameters[SZ].data;
    solarazimuth  = (int16 *)parameters[SA].data;
    sensorzenith  = (int16 *)parameters[VZ].data;
    sensorazimuth = (int16 *)parameters[VA].data;
#ifndef SUBSET
    t2mp          = (uint16 *)parameters[T2M].data;
    ncepp         = (uint8 *)parameters[NCEPWV].data;
#endif      
       tmpuint16 = orbitstateqa + (n_lines*COLWIDTH);

 /*      sprintf(name, "xxxx_%d.hdf\0", i);
       if ( (sd = SDstart(name, DFACC_CREATE)) == -1) {  / * delete for final version  * /
          sprintf(msg, "'%s' cannot open (5) ", name);
          modsmf(MODIS_E_FILE_ACCESS, msg, location);
          return(2);
          }

       dims1[0] = num_lines[i]/5;
       dims1[1] = COLWIDTH/5 + 1;

       sds = SDcreate(sd, "t2m", DFNT_UINT16, 2, dims1);
       printf("** %d\n", HDF_Write_2D(sds, 0, num_lines[i]/5, COLWIDTH/5+1, (VOID *)t2mp));
       SDendaccess(sds);
       sds = SDcreate(sd, "ncep", DFNT_UINT8, 2, dims1);
       printf("** %d\n", HDF_Write_2D(sds, 0, num_lines[i]/5, COLWIDTH/5+1, (VOID *)ncepp));
       SDendaccess(sds);

       dims1[0] = num_lines[i];
       dims1[1] = COLWIDTH;
       
       sds = SDcreate(sd, "band 31", DFNT_UINT16, 2, dims1);
       printf("** %d\n", HDF_Write_2D(sds, 0, num_lines[i], COLWIDTH, (VOID *)b31));
       SDendaccess(sds);
       sds = SDcreate(sd, "band 32", DFNT_UINT16, 2, dims1);
       printf("** %d\n", HDF_Write_2D(sds, 0, num_lines[i], COLWIDTH, (VOID *)b32));
       SDendaccess(sds);
       sds = SDcreate(sd, "water", DFNT_UINT16, 2, dims1);
       printf("** %d\n", HDF_Write_2D(sds, 0, num_lines[i], COLWIDTH, (VOID *)wvp));
       SDendaccess(sds);
       SDend(sd);
 */
 
       /* Calculate an "all_cloudy" mask, a 1km mask which says whether at least one pixel 
          within 20 pixels is cloudless (0) or not (1).  */
       all_cloudy=(int16 *) malloc(num_lines[i]*COLWIDTH*sizeof(int16));
       if (all_cloudy==NULL) {
            sprintf(msg,  " cannot allocate %d int16's (%d by %d) ", num_lines[i]*COLWIDTH, num_lines[i], COLWIDTH);
            modsmf(MODIS_E_MEM_FAILURE, msg, location);
            return(4);
           }

       for (row=0;row<num_lines[i];row++) {
         lorbit = row*COLWIDTH;
         for (col=0;col<COLWIDTH;col++) {
	   nbsamp = 0;
	   all_cloudy[lorbit+col]=0;
	   
	   for (k=(row-20);k<=(row+20);k++) {
	     if ((k<0)||(k>=num_lines[i])) continue;
             m = k*COLWIDTH;
	     for (l=(col-20);l<=(col+20);l++) {
	        if ((l<0)||(l>=COLWIDTH)) continue;
		
	           nbsamp++;                                   /* total number of pixels that map here */
		   intcm = (tmpuint16[m+l]&0x0400)>>10;
		   if (intcm == 1) all_cloudy[lorbit+col]++;   /* ... and total number that are cloudy */
		   
	       }
	     }
	    if (all_cloudy[lorbit+col] == nbsamp) all_cloudy[lorbit+col] = 1;  /* if they match, all are cloudy */
	    else                                  all_cloudy[lorbit+col] = 0;
           }
         }
 /*      sds = SDcreate(sd, "all_cloudy", DFNT_INT16, 2, dims1);
       printf("** %d\n", HDF_Write_2D(sds, 0, num_lines[i], COLWIDTH, (VOID *)all_cloudy));
       SDendaccess(sds);
       SDend(sd);
  */
       /**** CLOUD SHADOW DETECTION: UPDATE CM:  ***********/

       pixs=NBSCANS_1KM*2; /* neighborhood defined by 20x20 array for now */
       gridwidth=COLWIDTH/pixs+1;
       gridheight=(num_lines[i]+1)/pixs+1;
       tsclear=(float *) malloc(gridwidth*gridheight*sizeof(float));
       ts_clear=(float *) malloc(gridwidth*gridheight*sizeof(float));
       wv_clear=(float *) malloc(gridwidth*gridheight*sizeof(float));
       resampled_ncepp=(float *) malloc(gridwidth*gridheight*sizeof(float));
       if ((tsclear==NULL)||(ts_clear==NULL)||(wv_clear==NULL)||(resampled_ncepp==NULL)) {
            sprintf(msg,  " cannot allocate %d float's (%d by %d) ", 
	         gridwidth*gridheight, gridwidth, gridheight);
            modsmf(MODIS_E_MEM_FAILURE, msg, location);
            return(4);
           }
       for (row=0;row<gridheight;row++) {
         for (col=0;col<gridwidth;col++) {

#ifndef SUBSET
          for (k=row*4;((k<(row+1)*4)&&(k<(num_lines[i]/5)));k++) {  /* because t2m is at 5km (coarse) resolution; 4 = 20/5 */
             for (l=col*4;((l<(col+1)*4)&&(l<((COLWIDTH/5)+1)));l++) {
                m=k*((COLWIDTH/5)+1)+l;
                tsclear[row*gridwidth+col]=(float)t2mp[m]/100.0F;
                resampled_ncepp[row*gridwidth+col]=(float)ncepp[m]/50.0F;
              }
           } /* for (k=i;((k<i+pixs)&&(k<=granule.last_scan));k++) */
#else
                tsclear[row*gridwidth+col]=(float)T2M_value;   /* UNSCALED!! */
                resampled_ncepp[row*gridwidth+col]=(float)NCEPH2O_value;
#endif

           /* 16-SEP-02 */
           nbsamp=0;
	   t4=0.;
	   wv=0.;
           for (k=row*pixs;((k<(row+1)*pixs)&&(k<num_lines[i]));k++) {
             for (l=col*pixs;((l<(col+1)*pixs)&&(l<COLWIDTH));l++) {
                m=k*COLWIDTH+l;
		intcm = (tmpuint16[m]&0x0400)>>10;   /* internal CM from State QA */
                if (intcm == 0) {  
                   t4+=((float)b31[m]/100.0F);
                   wv+=((float)wvp[m]/100.0F);
                   nbsamp++;
                   }
             }
           } /* for (k=i;((k<i+pixs)&&(k<=granule.last_scan));k++) */
           if (nbsamp) {
              ts_clear[row*gridwidth+col]=t4/nbsamp;
              wv_clear[row*gridwidth+col]=wv/nbsamp;
            } else {
              ts_clear[row*gridwidth+col]=305.00F;
              wv_clear[row*gridwidth+col]=resampled_ncepp[row*gridwidth+col];
	     }

           }
         } /* for (row=0;row<gridheight;row++) */


/*       dims1[0] = gridheight;
       dims1[1] = gridwidth;
       sds = SDcreate(sd, "tsclear", DFNT_FLOAT, 2, dims1);
       printf("** %d\n", HDF_Write_2D(sds, 0, gridheight, gridwidth, (VOID *)tsclear));
       SDendaccess(sds);
       sds = SDcreate(sd, "ts_clear", DFNT_FLOAT, 2, dims1);
       printf("** %d\n", HDF_Write_2D(sds, 0, gridheight, gridwidth, (VOID *)ts_clear));
       SDendaccess(sds);
       sds = SDcreate(sd, "wv_clear", DFNT_FLOAT, 2, dims1);
       printf("** %d\n", HDF_Write_2D(sds, 0, gridheight, gridwidth, (VOID *)wv_clear));
       SDendaccess(sds);
       sds = SDcreate(sd, "resampled_ncepp", DFNT_FLOAT, 2, dims1);
       printf("** %d\n", HDF_Write_2D(sds, 0, gridheight, gridwidth, (VOID *)resampled_ncepp));
       SDendaccess(sds);
       SDend(sd);
 */

       /* 15-JUN-05; Will still set these bits but just for cloud-shadow purposes, and will UNSET them below.  */
if (xxx == 1) {
       for (m=0;m<(num_lines[i]*COLWIDTH);m++) {
	   /* 16-DEC-02 */
           if ((cloudmask[m]&0x0200) != 0)  {               /* if CM's a cloud edge... */
                tmpuint16[m] |= QAINTCM;                    /* flag the State QA as cloud */
		}
           }
}


       for (row=0;row<num_lines[i];row++) {
         for (col=0;col<COLWIDTH;col++) {	     
             lorbit=row*COLWIDTH+col;

             k=row/pixs;
             l=col/pixs;
	     intcm = (tmpuint16[lorbit]&QAINTCM)>>10;   /* internal CM from State QA */
	     
	     /* How about we only do this if the pixel here is cloudy AND at least one pixel within 20 
	        pixels of here is NOT cloudy; but we DON'T if ALL the pixels within 20 pixels of here are
		cloudy? */
             if ((intcm != 0) && (tsclear[k*gridwidth+l]!=0.) &&   /* cloudy */
	         (all_cloudy[lorbit]!=1)  ) { /* there is at least one non-cloudy pixel somewhere around it */
                flat = (float64)orbitlatitude[(row+n_lines)*COLWIDTH+col];

                if ((flat > LATLIMIT) || (flat < -LATLIMIT)) continue;  /* new, v5.3.12 */
		
                flat = (float64)orbitlatitude[(row+n_lines)*COLWIDTH+col]*DEG2RAD;
                flon = (float64)orbitlongitude[(row+n_lines)*COLWIDTH+col]*DEG2RAD;
		
                if ((solarzenith[lorbit]/100.0F) > 75.000F) continue;
		
                fts = ((float64)solarzenith[lorbit]/100.0F)*DEG2RAD;
                ffs = ((float64)solarazimuth[lorbit]/100.0F)*DEG2RAD;
                ftv = ((float64)sensorzenith[lorbit]/100.0F)*DEG2RAD;
                ffv = ((float64)sensorazimuth[lorbit]/100.0F)*DEG2RAD;

                t4=((float)b31[lorbit]/100.0F);
		
		if ((t4 < 100.0)||(t4 > 500.0)) continue;  /* new, v5.3.12; crazy values were throwing cldshd off. */  

                tmpflt64=(float64)t4;


                flat_save=flat;
                flon_save=flon;
		zz=(tsclear[k*gridwidth+l]-(float)tmpflt64)/6.0F;  /* z-temp diff is not 1:1, more like 6 km per degree */

                if (( fabsf(wv_clear[k*gridwidth+l]) > 1e-8) && (wvp[lorbit] != 0))    /* if wvp[lorbit] is zero, zz2 will be inf */
		   zz2 = -10.0*(log(((float)wvp[lorbit]/100.0F)/wv_clear[k*gridwidth+l]));
		else 
		   zz2 = zz;
		   
		if (zz2 < zz) zz2 = zz;
		for(z=(zz-(0.300F*zz)-1.50F);z<=(zz2+(0.300F*zz)+1.50F);z+=0.5F) {
                   flat=flat_save;
                   flon=flon_save;
                   if (z>0.) {
                       dx=cos(TWOPI-ffv)*tan(ftv)*z;
                       dy=sin(TWOPI-ffv)*tan(ftv)*z;
                       dlat=dx/(ERAD);
                       dlon=dy/(ERAD*cos(flat));
                       flat+=dlat;
                       flon-=dlon;
                       dx=cos(TWOPI-ffs)*tan(fts)*z;
                       dy=sin(TWOPI-ffs)*tan(fts)*z;
                       dlat=dx/(ERAD);
                       dlon=dy/(ERAD*cos(flat));
                       flat-=dlat;
                       flon+=dlon;
                       dlat=flat;
                       dlon=flon;
		       line = 0;
		       pixel = 0;
                       geo_sat(orbitlatitude, orbitlongitude, n_lines_in_orbit, n_lines+row, col, 
		               (float)dlat/DEG2RAD, (float)dlon/DEG2RAD, &line, &pixel, &iterations);
			       
		       if ((line==(n_lines+row))&&(pixel==col))continue;  /* 29-JAN-04: to prevent 3x3-related cldshd on the wrong side of the cloud */

                       if ((line>=0)&&(line<n_lines_in_orbit)&&(pixel>=0)&&(pixel<COLWIDTH)) {
			  
		          /* Here I read from orbitstateqa and not tmpuint16 because geo_sat may have
			     propogated line beyond the limits of the granule */
			  X = 1;
			  for (ll=(line-X);ll<=(line+X);ll++) {                    /* this does a 3 by 3 */
			      if ((ll < 0) || (ll > (n_lines_in_orbit-1))) continue;
			      for (kk=(pixel-X);kk<=(pixel+X);kk++) {
			          if ((kk < 0) || (kk > (COLWIDTH-1))) continue;
			          newindex = ll*COLWIDTH+kk;
	                          intcm = (orbitstateqa[newindex]&0x0400)>>10;
                                  if (intcm == 0) orbitstateqa[newindex] |= IISKMSHADOW; 
			         }
			      }



                          }
                       }  /* if z > 0 */
                     }  /* while (conv_factor ...) */
                  }   /* if cloudy */
                }  /* for j */
             }  /* for i */
       free(tsclear);
       free(ts_clear);
       free(wv_clear);
       free(resampled_ncepp);
       free(all_cloudy);

/*       sprintf(name, "xxxx_%d.hdf\0", i);
       if ( (sd = SDstart(name, DFACC_CREATE)) == -1) {  / * delete for final version * /
          sprintf(msg, "'%s' cannot open (6) ", name);
          modsmf(MODIS_E_FILE_ACCESS, msg, location);
          return(2);
          }
 
       dims1[0] = num_lines[i];
       dims1[1] = COLWIDTH;
       sds = SDcreate(sd, "iterations", DFNT_INT16, 2, dims1);
       printf("*** %d\n", HDF_Write_2D(sds, 0, num_lines[i], COLWIDTH, (VOID *)whatever));
       SDendaccess(sds);
       SDend(sd); 
 */
	
        if ( (sd = SDstart(ref_product_file[i], DFACC_RDWR)) == -1) {
          sprintf(msg, "'%s' cannot open (7) ", ref_product_file[i]);
          modsmf(MODIS_E_FILE_ACCESS, msg, location);
	  return(2);
          }
        
       
       for (m=0;m<(num_lines[i]*COLWIDTH);m++) {
	   if ( ICM_TEST(tmpuint16[m], IISKMSHADOW) != 0)   /* if StateQA's a cloud shadow... */
	        cloudmask[m] |= ICM_CLDSMASK;               /* flag the CM too. */
	   
           /* 15-JUN-05: Will do this: here, I will UNSET the StateQA intcm (bit 10). 
	       Bit 13 of the State QA is set elsewhere (toward the end of atmcorrect()). */
	       
if (xxx == 1) {
           if ((cloudmask[m]&0x0200) != 0) {
                tmpuint16[m] = tmpuint16[m] & ~QAINTCM;

            }
}
            
            }


  
       sds = SDselect(sd, SDnametoindex(sd, "1km Atmospheric Optical Depth Band QA\0"));  /* 5.0.2 */
       HDF_Write_2D(sds, 0, num_lines[i], COLWIDTH, (VOID *)cloudmask);
       SDendaccess(sds);

       sds = SDselect(sd, SDnametoindex(sd, "1km Reflectance Data State QA\0"));
       HDF_Write_2D(sds, 0, num_lines[i], COLWIDTH, (VOID *)tmpuint16);
       SDendaccess(sds);
          
       SDend(sd);
       
       n_lines += num_lines[i];

       if (b32 != NULL) free(b32);
       if (b31 != NULL) free(b31);
       if (wvp != NULL) free(wvp);
       if (qa != NULL) free(qa);
       if (cloudmask != NULL) free(cloudmask);
       if (latitude != NULL) free(latitude);
       if (longitude != NULL) free(longitude);
#ifndef SUBSET
       if (t2mp != NULL) free(t2mp);
       if (ncepp != NULL) free(ncepp);
#endif
       if (solarzenith != NULL) free(solarzenith);
       if (solarazimuth != NULL) free(solarazimuth);
       if (sensorzenith != NULL) free(sensorzenith);
       if (sensorazimuth != NULL) free(sensorazimuth);

   }   /* OK_to_process */
}  /*  for (i=1;i<=refout;i++)  */
       
/* Cleanup */
free(orbitlatitude);
free(orbitlongitude);
free(orbitstateqa);

modsmf(MODIS_A_CLDSHD_END, " ", location);   /* New, 23-APR-04 */

return(0);
}

void get_parameters(int32 sdid, hdf_param *parameter, int nlines)
/*
!C******************************************************************

!Description: This function reads parameters for the cloud shadow algorithm.

!Input Parameters:
   sdid               HDF file to read from
   n_lines            number of lines to read
   
!Output Parameters:
  none; 

!Input/Output Parameters:
  parameters          structure into which the data is read
  
!Returns:
   nothing.


!Revision History:

!Original version

!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes     
     
!END

*****************************************************************************/
{
int32 index, sds, rank, dims[3], number_type, nattr;
int size;
char name[256];

sds = SDselect(sdid, SDnametoindex(sdid, parameter->sdsname));
if (sds < 0) return;

SDgetinfo(sds, name, &rank, dims, &number_type, &nattr);
dims[2] = 0;
parameter->data = (void *)get_image(sds, dims, number_type, &size);
SDendaccess(sds);
parameter->dims[0] = dims[0];
parameter->dims[1] = dims[1];
return;
}

int get_parameter_info(int32 sdid, hdf_param parameter, int *n_lines)
/*
!C******************************************************************

!Description: This function reads information about parameters for the cloud 
              shadow algorithm -- the information is the number of lines.

!Input Parameters:
   sdid               HDF file to read from
   parameters         structure with information about the parameter
  
!Output Parameters:
   n_lines            number of lines to read

!Input/Output Parameters:
   none; 
  
!Returns:
   1   failure
   0   success


!Revision History:

!Original version

!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes     
     
!END

*****************************************************************************/
{
int32 index, sds, rank, dims[2], number_type, nattr;
char name[256];
char msg[256];
char location[100] = "aerosols.c, get_parameter_info\0";

index = SDnametoindex(sdid, parameter.sdsname);
if (index == -1) {
    sprintf(msg, "Can't find sds named '%s' in file\0", parameter.sdsname);
    modsmf(MODIS_E_HDF_ERROR, msg, location);
    return(1);
    }
    
sds = SDselect(sdid, index);
if (sds == -1) {
    sprintf(msg, "Can't run SDselect() on sds named '%s' in file\0", parameter.sdsname);
    modsmf(MODIS_E_HDF_ERROR, msg, location);
    return(1);
    }

SDgetinfo(sds, name, &rank, dims, &number_type, &nattr);
if ((rank != 2) || (number_type !=  parameter.type)){
    sprintf(msg, "Something wrong with sds '%s' in file\0", parameter.sdsname);
    modsmf(MODIS_E_HDF_ERROR, msg, location);
    return(1);
    }
SDendaccess(sds);

*n_lines = dims[0];

return(0);
} 

void get_orbital_data_float(int32 sdid, hdf_param parameter, float *space, int *n_lines, int8 line_increment)
/*
!C******************************************************************

!Description: Reads floating-point parameters and assembles it into a swath (for cloud shadow algorithm).

!Input Parameters:
   sdid               HDF file to read from
   parameters         structure with information about the parameter
   n_lines            number of lines to read
   line_increment     whether or not to increment n_lines 
 
!Output Parameters:
   space              the swathspace to be filled with the data
   
!Input/Output Parameters:
   n_lines            number of lines that have been read
  
!Returns:
   nothing

!Revision History:

!Original version

!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes     
     
!END

*****************************************************************************/
{
int32 i, sds, rank, dims[3], number_type, nattr;
int size;
char name[256];
float *data;

sds = SDselect(sdid, SDnametoindex(sdid, parameter.sdsname));
SDgetinfo(sds, name, &rank, dims, &number_type, &nattr);
dims[2] = 0;   /* get_image() has a little quirk... */
data = (float *)get_image(sds, dims, number_type, &size);
SDendaccess(sds);

for(i=0;i<dims[0]*dims[1];i++) space[(*n_lines*COLWIDTH)+i] = data[i];
free(data);
if (line_increment) *n_lines += dims[0];
}


void get_orbital_data_int(int32 sdid, hdf_param parameter, uint16 *space, int *n_lines, int8 line_increment)
/*
!C******************************************************************

!Description: Reads uint16 parameters and assembles it into a swath (for cloud shadow algorithm).

!Input Parameters:
   sdid               HDF file to read from
   parameters         structure with information about the parameter
   n_lines            number of lines to read
   line_increment     whether or not to increment n_lines 
 
!Output Parameters:
   space              the swathspace to be filled with the data
   
!Input/Output Parameters:
   n_lines            number of lines that have been read
  
!Returns:
   nothing

!Revision History:

!Original version

!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes     
     
!END

*****************************************************************************/
{
int32 i, sds, rank, dims[3], number_type, nattr;
int size;
char name[256];
uint16 *data = NULL;

sds = SDselect(sdid, SDnametoindex(sdid, parameter.sdsname));
SDgetinfo(sds, name, &rank, dims, &number_type, &nattr);
dims[2] = 0;
data = (uint16 *)get_image(sds, dims, number_type, &size);
SDendaccess(sds);

for(i=0;i<dims[0]*dims[1];i++) space[(*n_lines*COLWIDTH)+i] = data[i];
free(data);
if (line_increment) *n_lines += dims[0];
}


int get_filename_from_input(char *MOD09_filename, char *ESDT, char *outfilename)
/*
!C*****************************************************************************
!Description:
	Takes the MOD09 file, reads from its INPUTPOINTER any file that contains the
	input 'ESDT'; returns as 'outfilename' whatever entry in the PCF matches that
	filename (so I can have the path as well).   CURRENTLY NOT USED.

!Input Parameters:
        MOD09_filename     name of the MOD09 output file (same as in PCF, not the localgranuleid)
	                   File must have metadata written to it already.
        ESDT               What to look for in the INPUTPOINTER of MOD09_filename
	                   (Anything unique to the file you're looking for will do).

!Output Parameters:
        outfilename        the resultant file name, plus path.
	
!Returns:
        0     success
	non0  failure

!Revision History:

	Original version September 2002, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:
!References and Credits
!Design Notes

    1) memory for outfilename must be allocated by the calling routine.
    2) 'get_filename_from_input' is a misnomer, in that the PCF is also required.
	
!END***************************************************************************
*/
{
char tmp[PGSd_PC_FILE_PATH_MAX] = {"\0"};
char *attr[50];
char msg[PGSd_PC_FILE_PATH_MAX] = {"\0"};
PGSt_integer MOD09Versions, target_version, get_version, local_version;
char location[50]="aerosols.c, get_filename_from_input\0";
int i;
void addpath(char *file);

/* Why can't the toolkit read metadata from a file when given that file's NAME?  Wouldn't that
   be easy?  No, we need the file's PCF version number.  Look for it from the PCF. */
target_version = (PGSt_integer)-1;

if (PGS_PC_GetNumberOfFiles((PGSt_PC_Logical)REF, &MOD09Versions)!=PGS_S_SUCCESS) return(-1);
if (MOD09Versions > 1) {
  for (i=1;i<=MOD09Versions;i++) {
     get_version = (PGSt_integer)i;
     
     /* get MOD09 filename */
     if (PGS_PC_GetReference((PGSt_PC_Logical)REF, &get_version, tmp)!=PGS_S_SUCCESS) return(-3);
     rmpath(tmp);
     if (!strcmp(tmp, MOD09_filename))    /* if they match */
        target_version = (PGSt_integer)i;
     }
}
else if (MOD09Versions == 1) {
   target_version = (PGSt_integer)1;
}
else return(-2);

if (target_version == (PGSt_integer)-1)return(-4);

/* now that we have the version, use PGS_MET_GetPCAttr() to get the INPUTPOINTER */
local_version = target_version;
for(i=0;i<50;i++) {
   attr[i] = (char *)malloc(50);
   strcpy(attr[i], "");
   }
if(PGS_MET_GetPCAttr((PGSt_PC_Logical)REF, local_version, "CoreMetadata.0", "INPUTPOINTER\0", attr)!=PGS_S_SUCCESS){
    sprintf(msg, "Warning: INPUTPOINTER metadata field was not found within the file with lun %ld and version %d", 
	   REF, (int32)local_version);
    modsmf(MODIS_W_GET_METADATA, msg, location);
    for(i=0;i<50;i++) free(attr[i]);
    return(MODIS_W_GET_METADATA);
  } 

/* Now search through the INPUTPOINTER for the ESDT */
for(i=0;i<50;i++) {
   if (strlen(attr[i])) {
        if (strstr(attr[i], ESDT)) {
	     strcpy(outfilename, attr[i]);
	    }
      }
       
      /*printf("%d %p %s\n", i, strstr(attr[i], ESDT), outfilename);*/
  }
  
/* now -- *groan* -- check the whole PCF's inputs for this file and get path/file */
addpath(outfilename);

/* clean up */
for(i=0;i<50;i++) free(attr[i]);
printf("%s\n", outfilename);
return(0);
}

void addpath(char *file)
/*
!C*****************************************************************************
!Description:
	Looks through the PCF for the path for the filename in the argument

!Input Parameters:
        none   

!Input/Output Parameters:
        file              IN: the filename, alone
	                  OUT: the file name, plus path.
!Output Parameters:
        none
	
!Returns:
        nothing

!Revision History:

	Original version September 2002, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:
!References and Credits
!Design Notes

	
!END

****************************************************************************/
{
char tmp[PGSd_PC_FILE_PATH_MAX] = {"\0"};
PGSt_integer n_versions, get_version;
int i,j;
PGSt_integer LUN[6] = { RAD1, RAD500, RAD250, GRIB_ANCWVT, EPT_OZONEFILE, GEO};

for (i=0;i<6;i++) {
    if (PGS_PC_GetNumberOfFiles((PGSt_PC_Logical)LUN[i], &n_versions)!=PGS_S_SUCCESS) continue;
    if (n_versions<=0) continue;
    else {
        for (j=1;j<=n_versions;j++) {
             get_version = (PGSt_integer)j;
             if (PGS_PC_GetReference((PGSt_PC_Logical)LUN[i], &get_version, tmp)==PGS_S_SUCCESS) {
	        if (strstr(tmp, file)) {
	            strcpy(file, tmp);
	            return;
		     }
	        }
	    }
        }
   }  /* for (i=0;i<6;i++) */
}

/* 06-JUN-06: new cloud adjacency algorithm */

int cloud_adjacency(int refout, dims_struct ***outputs)
/*
!C*****************************************************************************
!Description:
	Routine which is the main() routine's interface for flagging QA 
	bits with cloud adjacency information.

!Input Parameters:
        none   

!Input/Output Parameters:
        refout
	outputs
	
!Output Parameters:
        none
	
!Returns:
        0 success
	1 file problem

!Revision History:

	Original version June 2006, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:
!References and Credits
!Design Notes

	
!END

****************************************************************************/
{
int32 sds,dims[2],start[2], sdindex;
uint8 *qa;
uint16 *cm;
uint16 *sqa;
uint8 *lqa;
uint8 *nqa;
char location[40]="aerosols.c, cloud_adjacency\0";
char *ref_product_file[30];
int n_1kmlines_file[30], n_lines, ln_lines, nn_lines;
int32 all_sd[30];
int i, j, startfile, endfile;
extern char sdsname[ALLBANDS-BANDPROCESS+1][50];
char name[40];
int cldadj_routine(int n_lines, uint8 *qa, int ln_lines, uint8 *lqa, 
                   int nn_lines, uint8 *nqa, uint16 *cm, uint16 *sqa);

startfile = endfile = -1;
n_lines = ln_lines = nn_lines = 0;

for (i=0;i<refout;i++) {

   n_1kmlines_file[i] = -1;   /* 17-APR-06 */
   
   if ((outputs[0][i]->filename != NULL) &&
      ((outputs[0][i]->nblines > 2000) && 
       (outputs[0][i]->nblines < 2080)))  { 
   
      if (startfile == -1)startfile = i;
      if (endfile < i) endfile = i;
      
      ref_product_file[i] = (char *)malloc(strlen(outputs[0][i]->filename)+1);
      strcpy(ref_product_file[i], outputs[0][i]->filename);
      n_1kmlines_file[i] = outputs[0][i]->nblines; 
      
      if ( (all_sd[i] = SDstart(ref_product_file[i], DFACC_RDWR)) == -1) {
	  sprintf(msg, "'%s' ", ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
	  return(1);
         }
      
      }
      else {
         sprintf(msg, "Possible error: file '%s' has unanticipated number of lines (%d), cannot do cloud adjacency on that file", 
	     outputs[0][i]->filename, outputs[0][i]->nblines);
         modsmf(MODIS_E_FILE_ACCESS, msg, location);        
         }
   }   

if (startfile == -1 || endfile == -1) {
    sprintf(msg, "Programming error, line %d; cannot do cloud adjacency", __LINE__);
    modsmf(MODIS_E_FILE_ACCESS, msg, location);
    return(2);
   }

for (i=startfile;i<=endfile;i++) { 

   if (n_1kmlines_file[i] == -1) continue;   /* 17-APR-06 */

   /* Initialize */
   n_lines = ln_lines = nn_lines = 0;
   qa = (uint8 *)NULL;
   cm = (uint16 *)NULL;
   lqa = (uint8 *)NULL;
   nqa = (uint8 *)NULL;

   /* Get memory for the current granule */
   qa = (uint8 *)AllocMem(n_1kmlines_file[i]*NBSPOTS_1KM*sizeof(uint8), location);   
   cm = (uint16 *)AllocMem(n_1kmlines_file[i]*NBSPOTS_1KM*sizeof(uint16), location);   
   sqa = (uint16 *)AllocMem(n_1kmlines_file[i]*NBSPOTS_1KM*sizeof(uint16), location);   
      
   if ((i > startfile) && (n_1kmlines_file[i-1] > 0)) {
      /* Get memory for the preceeding granule */
      lqa = (uint8 *)AllocMem(n_1kmlines_file[i-1]*NBSPOTS_1KM*sizeof(uint8), location);   
      }
   if ((i != endfile) && (n_1kmlines_file[i+1] > 0)) {
      /* Get memory for the next granule */
      nqa = (uint8 *)AllocMem(n_1kmlines_file[i+1]*NBSPOTS_1KM*sizeof(uint8), location);   
      }

   start[0] = start[1] = 0;
   dims[1] = NBSPOTS_1KM;

   sprintf(msg, "reading '1km Atmospheric Optical Depth Band QA', file %s\0", ref_product_file[i]); 
   modsmf(MODIS_W_INFORMATION, msg, location);

   /* 31-MAY-06: Read the cm */  /* at one time the cm was actually called "1km Atmospheric Optical Depth Band CM";
                                     but we reversed the names... */
   sdindex = SDnametoindex(all_sd[i], "1km Atmospheric Optical Depth Band QA");
   sds = SDselect(all_sd[i], sdindex);
   if ((sdindex == -1) || (sds == -1)) {
        sprintf(msg,  "SDS '1km Atmospheric Optical Depth Band QA' not found in file %s ", 
	      ref_product_file[i]);
	 modsmf(MODIS_E_FILE_ACCESS, msg, location);
         for (j=startfile;j<=endfile;j++) SDend(all_sd[j]);
	 return(1);
       }
      
   dims[0] = n_1kmlines_file[i];
   SDreaddata(sds, start, NULL, dims, (VOID *)cm);
   SDendaccess(sds);

   /* 05-JUN-06: Read the state qa. */
   sdindex = SDnametoindex(all_sd[i], "1km Reflectance Data State QA");
   sds = SDselect(all_sd[i], sdindex);
   if ((sdindex == -1) || (sds == -1)) {
        sprintf(msg,  "SDS '1km Reflectance Data State QA' not found in file %s ", 
	      ref_product_file[i]);
	 modsmf(MODIS_E_FILE_ACCESS, msg, location);
         for (j=startfile;j<=endfile;j++) SDend(all_sd[j]);
	 return(1);
       }
      
   dims[0] = n_1kmlines_file[i];
   SDreaddata(sds, start, NULL, dims, (VOID *)sqa);
   SDendaccess(sds);
   
   /* Read the qa's */
   sdindex = SDnametoindex(all_sd[i], "1km Atmospheric Optical Depth Band CM");
   sds = SDselect(all_sd[i], sdindex);
   if ((sdindex == -1) || (sds == -1)) {
        sprintf(msg,  "SDS '1km Atmospheric Optical Depth Band CM' not found in file %s ", 
	      ref_product_file[i]);
	 modsmf(MODIS_E_FILE_ACCESS, msg, location);
         for (j=startfile;j<=endfile;j++) SDend(all_sd[j]);
	 return(1);
       }
      
   dims[0] = n_1kmlines_file[i];
   SDreaddata(sds, start, NULL, dims, (VOID *)qa);
   SDendaccess(sds);
   n_lines = n_1kmlines_file[i];

   if ((i > startfile) && (n_1kmlines_file[i-1] > 0)) {
      sdindex = SDnametoindex(all_sd[i-1], "1km Atmospheric Optical Depth Band CM");
      sds = SDselect(all_sd[i-1], sdindex);
      if ((sdindex == -1) || (sds == -1)) {
           sprintf(msg,  "SDS '1km Atmospheric Optical Depth Band CM' not found in file %s ", 
	      ref_product_file[i-1]);
	    modsmf(MODIS_E_FILE_ACCESS, msg, location);
            for (j=startfile;j<=endfile;j++) SDend(all_sd[j]);
	    return(1);
          }
      dims[0] = n_1kmlines_file[i-1];
      SDreaddata(sds, start, NULL, dims, (VOID *)lqa);
      SDendaccess(sds);
      ln_lines = n_1kmlines_file[i-1];
      }
   if ((i != endfile) && (n_1kmlines_file[i+1] > 0)) {
      sdindex = SDnametoindex(all_sd[i+1], "1km Atmospheric Optical Depth Band CM");
      sds = SDselect(all_sd[i+1], sdindex);
      if ((sdindex == -1) || (sds == -1)) {
           sprintf(msg,  "SDS '1km Atmospheric Optical Depth Band CM' not found in file %s ", 
	      ref_product_file[i+1]);
	    modsmf(MODIS_E_FILE_ACCESS, msg, location);
            for (j=startfile;j<=endfile;j++) SDend(all_sd[j]);
	    return(1);
          }
      dims[0] = n_1kmlines_file[i+1];
      SDreaddata(sds, start, NULL, dims, (VOID *)nqa);
      SDendaccess(sds);
      nn_lines = n_1kmlines_file[i+1];
      }

    
   /* Call the cldadj routine */

   cldadj_routine(n_lines, qa, ln_lines, lqa, nn_lines, nqa, cm, sqa);

   /* 31-MAY-06: NOW WRITE THE CM */
   sprintf(msg, "writing '1km Atmospheric Optical Depth Band QA', file %s\0", ref_product_file[i]); 
   modsmf(MODIS_W_INFORMATION, msg, location);
   sdindex = SDnametoindex(all_sd[i], "1km Atmospheric Optical Depth Band QA");
   sds = SDselect(all_sd[i], sdindex);
   if ((sdindex == -1) || (sds == -1)) {
        sprintf(msg,  "SDS '1km Atmospheric Optical Depth Band QA' not found in file %s ", 
	      ref_product_file[i]);
	 modsmf(MODIS_E_FILE_ACCESS, msg, location);
       }
      
   dims[0] = n_1kmlines_file[i];
   SDwritedata(sds, start, NULL, dims, (VOID *)cm);
   SDendaccess(sds);

   /* 31-MAY-06: NOW WRITE THE State QA */
   sdindex = SDnametoindex(all_sd[i], "1km Reflectance Data State QA");
   sds = SDselect(all_sd[i], sdindex);
   if ((sdindex == -1) || (sds == -1)) {
        sprintf(msg,  "SDS '1km Reflectance Data State QA' not found in file %s ", 
	      ref_product_file[i]);
	 modsmf(MODIS_E_FILE_ACCESS, msg, location);
       }

   dims[0] = n_1kmlines_file[i];
   SDwritedata(sds, start, NULL, dims, (VOID *)sqa);
   SDendaccess(sds);

 
   /* 06-JUN-06: write the QA also... */
   sdindex = SDnametoindex(all_sd[i], "1km Atmospheric Optical Depth Band CM");
   sds = SDselect(all_sd[i], sdindex);
   if ((sdindex == -1) || (sds == -1)) {
        sprintf(msg,  "SDS '1km Atmospheric Optical Depth Band CM' not found in file %s ", 
	      ref_product_file[i]);
	 modsmf(MODIS_E_FILE_ACCESS, msg, location);
       }
      
   dims[0] = n_1kmlines_file[i];
   SDwritedata(sds, start, NULL, dims, (VOID *)qa);
   SDendaccess(sds);

      
   /* Free memory */
   
   free(cm);
   free(qa);
   free(sqa);
   if (i > startfile) free(lqa);
   if (i != endfile) free(nqa);
        
   }   /*  for (i=startfile;i<=endfile;i++) */

/* Oh yes, by the way... */
for (i=startfile;i<=endfile;i++) SDend(all_sd[i]);
   

return(0);
}


int cldadj_routine(int n_lines, uint8 *qa, int ln_lines, uint8 *lqa, 
                   int nn_lines, uint8 *nqa, uint16 *cm, uint16 *sqa)
/*
!C*****************************************************************************
!Description:
	Cloud adjacency routine
	
!Input Parameters:
        none   

!Input/Output Parameters:
        n_lines      The number of lines in qa
	qa           The QA
        ln_lines     SAME FOR THE PRECEDING GRANULE
	lqa         
        nn_lines     SAME FOR THE FOLLOWING GRANULE
	nqa         
	cm           contents of "1km Atmospheric Optical Depth Band QA" SDS, to be updated.
	             Only for the qa/data granule, not for preceding/following...        
	sqa          contents of "state qa SDS, to be updated.
	             Only for the qa/data granule, not for preceding/following...        
	
!Output Parameters:
        none
	
!Returns:
         0 success
        -1 error

!Revision History:

	Original version June 2006, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:
!References and Credits
!Design Notes

	
!END

****************************************************************************/
{
#define NNN 10
int i, j, k, total_pixels, do_qa, n_interp, row, col, n_cols;
long l, ll, in_dex, start, realstart;
uint8 *localqa;
int ds_lines, ds_cols, total_lines;
char location[256] = "aerosols.c, cldadj_routine\0";
uint8 qaval;

n_cols = NBSPOTS_1KM;

if (n_lines == 0) return(-1);

total_lines = n_lines;
if (nn_lines > 0) total_lines += NNN;
if (ln_lines > 0) total_lines += NNN;

total_pixels = total_lines*n_cols;  

localqa = (uint8 *)AllocMem(total_lines*NBSPOTS_1KM*sizeof(uint8), location);   

/* Make a swathlette 
 ************************/
realstart = start = 0L;
if (total_lines == n_lines) {   /* one granule being processed; do essentially nothing */
   for (l=0L;l<n_lines*n_cols;l++) {
      localqa[l] = qa[l];
      }
   }
else {

   if (ln_lines > 0) {  /* Take their last NNN */
       l = 0;
       for (i=ln_lines-NNN;i<ln_lines;i++) {
          in_dex = i*n_cols;
	  for (k=0;k<n_cols;k++) localqa[l++] = lqa[in_dex+k];
          }
       start = l;
       realstart = l;
      }  /* if (ln_lines > 0) */
      
    for (l=0L;l<n_lines*n_cols;l++) {
         localqa[start+l] = qa[l];
         }
    start += n_lines*n_cols;	
     
    if (nn_lines > 0) {  /* Take their first NNN */
       l = 0;
       for (i=0;i<NNN;i++) {
          in_dex = i*n_cols;
	  for (k=0;k<n_cols;k++,l++) localqa[start+l] = nqa[in_dex+k];
          }
         
      }  /* if (nn_lines > 0) */
      
      
    }
   
  /* Do cloud adjacency */
 
  for(row=1;row<total_lines-1;row++){
        for(col=1;col<n_cols-1;col++){
           l = row*n_cols + col;
           qaval = localqa[l];
	   if ( qaval == 4 ) {
	       if (localqa[l-1] != 4) localqa[l-1] = 19;
	       if (localqa[l+1] != 4) localqa[l+1] = 19;
	       ll = l - n_cols;
	       if (localqa[ll-1] != 4) localqa[ll-1] = 19;
	       if (localqa[ll]   != 4) localqa[ll]   = 19;
	       if (localqa[ll+1] != 4) localqa[ll+1] = 19;
	       ll = l + n_cols;
	       if (localqa[ll-1] != 4) localqa[ll-1] = 19;
	       if (localqa[ll]   != 4) localqa[ll]   = 19;
	       if (localqa[ll+1] != 4) localqa[ll+1] = 19;
	      }
          }
       }

  for (l=0L;l<(n_lines*NBSPOTS_1KM);l++) {   /* now update the rest */
        qaval = localqa[realstart+l];
        if (qaval == 19) {
	    cm[l] |= 512;                         /* cloud adjacent */
	    cm[l] |= QACLIMATOLOGY;               /* climatological AOT */
            sqa[l] = sqa[l] & ~QAHIGHAEROSOL;     /* 04-FEB-05: CLEAR THESE AEROSOL QUANTITY BITS */
            sqa[l] = sqa[l] | OISADJACENT;
	    }
       }

/* Now snip out the qa arrays from localqa 
 ********************************************/
for (l=0L;l<n_lines*n_cols;l++) {
   qa[l] = localqa[realstart+l];
         }

/* and free localqa 
 *******************/
free(localqa);
   
   
return(0);
}



/* 31-MAR-06: New stuff for transmission, path radiance, spherical albedo interpolation */

int tr_interpolation(int refout, dims_struct ***outputs)
/*
!C*****************************************************************************
!Description:
	Routine which is the main() routine's interface for the transmission, 
	path radiance and spherical albedo's interpolation.

!Input Parameters:
        none   

!Input/Output Parameters:
        refout
	outputs
	
!Output Parameters:
        none
	
!Returns:
        0 success
	1 file problem

!Revision History:

	Original version March 2006, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:
!References and Credits
!Design Notes

	
!END

****************************************************************************/
{
int32 sds,dims[2],start[2], sdindex;
int16 *data[16];
uint8 *qa;
uint16 *cm;
int16 *ldata[16];
uint8 *lqa;
int16 *ndata[16];
uint8 *nqa;
char location[40]="aerosols.c, tr_interpolation\0";
char *ref_product_file[30];
int n_1kmlines_file[30], n_lines, ln_lines, nn_lines;
int32 all_sd[30];
int i, j, startfile, endfile;
extern char sdsname[ALLBANDS-BANDPROCESS+1][50];
char name[40];
int multiple_write(int i, int startfile, int endfile, int32 *all_sd, int *n_1kmlines_file, int SDSINDEX, 
                  int16 *data[16], int16 *ldata[16], int16 *ndata[16]);
int multiple_read(int i, int startfile, int endfile, int32 *all_sd, int *n_1kmlines_file, int SDSINDEX, 
                  int16 *data[16], int16 *ldata[16], int16 *ndata[16]);
int tr_interpol_routine(int n_lines, int16 *data[16], uint8 *qa, 
                        int ln_lines, int16 *ldata[16], uint8 *lqa, 
                        int nn_lines, int16 *ndata[16], uint8 *nqa, uint16 *cm);

startfile = endfile = -1;
n_lines = ln_lines = nn_lines = 0;

/*return(0); / * 14-APR-06 : for now... */

for (i=0;i<refout;i++) {

   n_1kmlines_file[i] = -1;   /* 17-APR-06 */
   
   if ((outputs[0][i]->filename != NULL) &&
      ((outputs[0][i]->nblines > 2000) && 
       (outputs[0][i]->nblines < 2080)))  { 
   
      if (startfile == -1)startfile = i;
      if (endfile < i) endfile = i;
      
      ref_product_file[i] = (char *)malloc(strlen(outputs[0][i]->filename)+1);
      strcpy(ref_product_file[i], outputs[0][i]->filename);
      n_1kmlines_file[i] = outputs[0][i]->nblines; 
      
      if ( (all_sd[i] = SDstart(ref_product_file[i], DFACC_RDWR)) == -1) {
	  sprintf(msg, "'%s' ", ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
	  return(1);
         }
      
      }
      else {
         sprintf(msg, "Possible error: file '%s' has unanticipated number of lines (%d), cannot interpolate TR/PR/SA arrays in that file", 
	     outputs[0][i]->filename, outputs[0][i]->nblines);
         modsmf(MODIS_E_FILE_ACCESS, msg, location);        
         }
   }   

if (startfile == -1 || endfile == -1) {
    sprintf(msg, "Programming error, line %d; cannot interpolate TR/PR/SA arrays", __LINE__);
    modsmf(MODIS_E_FILE_ACCESS, msg, location);
    return(2);
   }

for (i=startfile;i<=endfile;i++) { 

   if (n_1kmlines_file[i] == -1) continue;   /* 17-APR-06 */

   /* Initialize */
   n_lines = ln_lines = nn_lines = 0;
   for(j=0;j<16;j++) {
      data[j] = (int16 *)NULL;
      ldata[j] = (int16 *)NULL;
      ndata[j] = (int16 *)NULL;
      }
    qa = (uint8 *)NULL;
    cm = (uint16 *)NULL;
    lqa = (uint8 *)NULL;
    nqa = (uint8 *)NULL;

   /* Get memory for the current granule */
   qa = (uint8 *)AllocMem(n_1kmlines_file[i]*NBSPOTS_1KM*sizeof(uint8), location);   
   cm = (uint16 *)AllocMem(n_1kmlines_file[i]*NBSPOTS_1KM*sizeof(uint16), location);   
   for(j=0;j<16;j++) {
      data[j] = (int16 *)AllocMem(n_1kmlines_file[i]*NBSPOTS_1KM*sizeof(int16), location); 
      }
      
   if ((i > startfile) && (n_1kmlines_file[i-1] > 0)) {
      /* Get memory for the preceeding granule */
      lqa = (uint8 *)AllocMem(n_1kmlines_file[i-1]*NBSPOTS_1KM*sizeof(uint8), location);   
      for(j=0;j<16;j++) {
         ldata[j] = (int16 *)AllocMem(n_1kmlines_file[i-1]*NBSPOTS_1KM*sizeof(int16), location); 
         }
      }
   if ((i != endfile) && (n_1kmlines_file[i+1] > 0)) {
      /* Get memory for the next granule */
      nqa = (uint8 *)AllocMem(n_1kmlines_file[i+1]*NBSPOTS_1KM*sizeof(uint8), location);   
      for(j=0;j<16;j++) {
         ndata[j] = (int16 *)AllocMem(n_1kmlines_file[i+1]*NBSPOTS_1KM*sizeof(int16), location); 
         }
      }

   start[0] = start[1] = 0;
   dims[1] = NBSPOTS_1KM;

   /* 31-MAY-06: Read the cm */  /* at one time the cm was actually called "1km Atmospheric Optical Depth Band CM";
                                     but we reversed the names... */
   sdindex = SDnametoindex(all_sd[i], "1km Atmospheric Optical Depth Band QA");
   sds = SDselect(all_sd[i], sdindex);
   if ((sdindex == -1) || (sds == -1)) {
        sprintf(msg,  "SDS '1km Atmospheric Optical Depth Band QA' not found in file %s ", 
	      ref_product_file[i]);
	 modsmf(MODIS_E_FILE_ACCESS, msg, location);
         for (j=startfile;j<=endfile;j++) SDend(all_sd[j]);
	 return(1);
       }
      
   dims[0] = n_1kmlines_file[i];
   SDreaddata(sds, start, NULL, dims, (VOID *)cm);
   SDendaccess(sds);
   
   /* Read the qa's */
   sdindex = SDnametoindex(all_sd[i], "1km Atmospheric Optical Depth Band CM");
   sds = SDselect(all_sd[i], sdindex);
   if ((sdindex == -1) || (sds == -1)) {
        sprintf(msg,  "SDS '1km Atmospheric Optical Depth Band CM' not found in file %s ", 
	      ref_product_file[i]);
	 modsmf(MODIS_E_FILE_ACCESS, msg, location);
         for (j=startfile;j<=endfile;j++) SDend(all_sd[j]);
	 return(1);
       }
      
   dims[0] = n_1kmlines_file[i];
   SDreaddata(sds, start, NULL, dims, (VOID *)qa);
   SDendaccess(sds);
   n_lines = n_1kmlines_file[i];

   if ((i > startfile) && (n_1kmlines_file[i-1] > 0)) {
      sdindex = SDnametoindex(all_sd[i-1], "1km Atmospheric Optical Depth Band CM");
      sds = SDselect(all_sd[i-1], sdindex);
      if ((sdindex == -1) || (sds == -1)) {
           sprintf(msg,  "SDS '1km Atmospheric Optical Depth Band CM' not found in file %s ", 
	      ref_product_file[i-1]);
	    modsmf(MODIS_E_FILE_ACCESS, msg, location);
            for (j=startfile;j<=endfile;j++) SDend(all_sd[j]);
	    return(1);
          }
      dims[0] = n_1kmlines_file[i-1];
      SDreaddata(sds, start, NULL, dims, (VOID *)lqa);
      SDendaccess(sds);
      ln_lines = n_1kmlines_file[i-1];
      }
   if ((i != endfile) && (n_1kmlines_file[i+1] > 0)) {
      sdindex = SDnametoindex(all_sd[i+1], "1km Atmospheric Optical Depth Band CM");
      sds = SDselect(all_sd[i+1], sdindex);
      if ((sdindex == -1) || (sds == -1)) {
           sprintf(msg,  "SDS '1km Atmospheric Optical Depth Band CM' not found in file %s ", 
	      ref_product_file[i+1]);
	    modsmf(MODIS_E_FILE_ACCESS, msg, location);
            for (j=startfile;j<=endfile;j++) SDend(all_sd[j]);
	    return(1);
          }
      dims[0] = n_1kmlines_file[i+1];
      SDreaddata(sds, start, NULL, dims, (VOID *)nqa);
      SDendaccess(sds);
      nn_lines = n_1kmlines_file[i+1];
      }

    
   /* Process transmission */
   sprintf(msg, "interpolating transmission, file %s\0", ref_product_file[i]); 
   modsmf(MODIS_W_INFORMATION, msg, location);

   multiple_read(i, startfile, endfile, all_sd, n_1kmlines_file, TRBAND1, data, ldata, ndata);
   tr_interpol_routine(n_lines, data, qa, ln_lines, ldata, lqa, nn_lines, ndata, nqa, NULL);
   multiple_write(i, startfile, endfile, all_sd, n_1kmlines_file, TRBAND1, data, ldata, ndata);
     
   /* Process path radiance */       
   sprintf(msg, "interpolating path radiance, file %s\0", ref_product_file[i]); 
   modsmf(MODIS_W_INFORMATION, msg, location);

   multiple_read(i, startfile, endfile, all_sd, n_1kmlines_file, PRBAND1, data, ldata, ndata);
   tr_interpol_routine(n_lines, data, qa, ln_lines, ldata, lqa, nn_lines, ndata, nqa, NULL);
   multiple_write(i, startfile, endfile, all_sd, n_1kmlines_file, PRBAND1, data, ldata, ndata);
     
   /* Process spherical albedo */   
   sprintf(msg, "interpolating spherical albedo, file %s\0", ref_product_file[i]); 
   modsmf(MODIS_W_INFORMATION, msg, location);

   multiple_read(i, startfile, endfile, all_sd, n_1kmlines_file, SABAND1, data, ldata, ndata);
   tr_interpol_routine(n_lines, data, qa, ln_lines, ldata, lqa, nn_lines, ndata, nqa, cm);
   multiple_write(i, startfile, endfile, all_sd, n_1kmlines_file, SABAND1, data, ldata, ndata);
 

   /* 31-MAY-06: NOW WRITE THE CM */
   sdindex = SDnametoindex(all_sd[i], "1km Atmospheric Optical Depth Band QA");
   sds = SDselect(all_sd[i], sdindex);
   if ((sdindex == -1) || (sds == -1)) {
        sprintf(msg,  "SDS '1km Atmospheric Optical Depth Band QA' not found in file %s ", 
	      ref_product_file[i]);
	 modsmf(MODIS_E_FILE_ACCESS, msg, location);
       }
      
   dims[0] = n_1kmlines_file[i];
   SDwritedata(sds, start, NULL, dims, (VOID *)cm);
   SDendaccess(sds);


   /* Free memory */
   
   free(cm);
   free(qa);
   if (i > startfile) free(lqa);
   if (i != endfile) free(nqa);
   for(j=0;j<16;j++) {
      free(data[j]);
      if (i > startfile) free(ldata[j]);
      if (i != endfile)  free(ndata[j]);
     }
        
   }   /*  for (i=startfile;i<=endfile;i++) */

/* Oh yes, by the way... */
for (i=startfile;i<=endfile;i++) SDend(all_sd[i]);
   

return(0);
}

int multiple_read(int i, int startfile, int endfile, int32 *all_sd, int *n_1kmlines_file, int SDSINDEX, 
                   int16 *data[16], int16 *ldata[16], int16 *ndata[16])
{
int32 j, sds, start[2], dims[2], sdindex;
char name[50];
extern char sdsname[ALLBANDS-BANDPROCESS+1][50];
char location[40]="aerosols.c, multiple_read\0";
int32 max_cache, cache_flag;
int set_chunking = 0;
int32 ret, lun, dex;   /* 14-APR-06 */
long total_n_pixels;
PGSt_IO_Gen_FileHandle *fd;

#ifdef TOTEMP  /* 14-APR-06 */
   dex = -1;
   switch (SDSINDEX) {
      case TRBAND1: dex = TRANS; break;
      case PRBAND1: dex = PATHR; break;
      case SABAND1: dex = SPHER; break;
      case TGBAND1: dex = GASTR; break;
      default: break;
      }
   if( dex == -1 ) {
       sprintf(msg, "determining parameter");
       modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
      }
      
   total_n_pixels = n_1kmlines_file[i]*NBSPOTS_1KM;   
   lun = (PGSt_PC_Logical)(TEMPLUN + (i*4) + dex);
      
   if( (ret = PGS_IO_Gen_Temp_Open( PGSd_IO_Gen_NoEndurance, lun, PGSd_IO_Gen_Read, &fd ) )!= PGS_S_SUCCESS ) {
       sprintf(msg, "opening temp file, LUN %d", lun);
       modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
      }
   fseek(fd, 0L, SEEK_SET);
   for(j=0;j<16;j++) {
       fread(data[j], 2, total_n_pixels, fd);
      }
      
   if( ( ret = PGS_IO_Gen_Close( fd )) != PGS_S_SUCCESS ) {
       sprintf(msg, "closing temp file, LUN %d", lun);
       modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
      }
      
   if ((i > startfile) && (n_1kmlines_file[i-1] > 0)) {
      total_n_pixels = n_1kmlines_file[i-1]*NBSPOTS_1KM;   
      lun = (PGSt_PC_Logical)(TEMPLUN + ((i-1)*4) + dex);
      
      if( (ret = PGS_IO_Gen_Temp_Open( PGSd_IO_Gen_NoEndurance, lun, PGSd_IO_Gen_Read, &fd ) )!= PGS_S_SUCCESS ) {
          sprintf(msg, "opening temp file, LUN %d", lun);
          modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
         }
      fseek(fd, 0L, SEEK_SET);
      for(j=0;j<16;j++) {
          fread(ldata[j], 2, total_n_pixels, fd);
         }   
      
      if( ( ret = PGS_IO_Gen_Close( fd )) != PGS_S_SUCCESS ) {
          sprintf(msg, "closing temp file, LUN %d", lun);
          modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
         }
      }
   if ((i != endfile) && (n_1kmlines_file[i+1] > 0)) {
      total_n_pixels = n_1kmlines_file[i+1]*NBSPOTS_1KM;   
      lun = (PGSt_PC_Logical)(TEMPLUN + ((i+1)*4) + dex);
      
      if( (ret = PGS_IO_Gen_Temp_Open( PGSd_IO_Gen_NoEndurance, lun, PGSd_IO_Gen_Read, &fd ) )!= PGS_S_SUCCESS ) {
          sprintf(msg, "opening temp file, LUN %d", lun);
          modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
         }
      fseek(fd, 0L, SEEK_SET);
      for(j=0;j<16;j++) {
          fread(ndata[j], 2, total_n_pixels, fd);
         }   
      
      if( ( ret = PGS_IO_Gen_Close( fd )) != PGS_S_SUCCESS ) {
          sprintf(msg, "closing temp file, LUN %d", lun);
          modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
         }
      }
   
#else
max_cache = 8;
cache_flag = 0;

   for(j=0;j<16;j++) {
      sprintf(name, "1km %s\0", sdsname[(j+SDSINDEX)-BAND20ALB]);
      sdindex = SDnametoindex(all_sd[i], name);
      sds = SDselect(all_sd[i], sdindex);
      if (set_chunking) SDsetchunkcache(sds, max_cache, cache_flag);
      if ((sdindex == -1) || (sds == -1)) {
           sprintf(msg,  "SDS '%s' not found in file ", name);
	    modsmf(MODIS_E_FILE_ACCESS, msg, location);
	    return(1);
          }
      
      dims[0] = n_1kmlines_file[i];
      start[0] = start[1] = 0;
      dims[1] = NBSPOTS_1KM;
      SDreaddata(sds, start, NULL, dims, (VOID *)data[j]);
      SDendaccess(sds);

      if ((i > startfile) && (n_1kmlines_file[i-1] > 0)) {
         sdindex = SDnametoindex(all_sd[i-1], name);
         sds = SDselect(all_sd[i-1], sdindex);
         if (set_chunking)SDsetchunkcache(sds, max_cache, cache_flag);
         dims[0] = n_1kmlines_file[i-1];
         SDreaddata(sds, start, NULL, dims, (VOID *)ldata[j]);
         SDendaccess(sds);
         }
      if ((i != endfile) && (n_1kmlines_file[i+1] > 0)) {
         sdindex = SDnametoindex(all_sd[i+1], name);
         sds = SDselect(all_sd[i+1], sdindex);
         if (set_chunking)SDsetchunkcache(sds, max_cache, cache_flag);
         dims[0] = n_1kmlines_file[i+1];
         SDreaddata(sds, start, NULL, dims, (VOID *)ndata[j]);
         SDendaccess(sds);
         }
     }
#endif    
return(0);     
}
		    
int multiple_write(int i, int startfile, int endfile, int32 *all_sd, int *n_1kmlines_file, int SDSINDEX, 
                    int16 *data[16], int16 *ldata[16], int16 *ndata[16])
{
int32 j, sds, start[2], dims[2], sdindex;
char name[50];
extern char sdsname[ALLBANDS-BANDPROCESS+1][50];
HDF_CHUNK_DEF cdef;
char location[40]="aerosols.c, multiple_write\0";
int set_chunking = 0;
int32 ret, lun, dex;   /* 14-APR-06 */
long total_n_pixels;
PGSt_IO_Gen_FileHandle *fd;

#ifdef TOTEMP  /* 14-APR-06 */
   dex = -1;
   switch (SDSINDEX) {
      case TRBAND1: dex = TRANS; break;
      case PRBAND1: dex = PATHR; break;
      case SABAND1: dex = SPHER; break;
      case TGBAND1: dex = GASTR; break;
      default: break;
      }
   if( dex == -1 ) {
       sprintf(msg, "determining parameter");
       modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
      }
      
   total_n_pixels = n_1kmlines_file[i]*NBSPOTS_1KM;   
   lun = (PGSt_PC_Logical)(TEMPLUN + (i*4) + dex);
   
   /* Get rid of the file! */
   PGS_IO_Gen_Temp_Delete( lun );

   if( (ret = PGS_IO_Gen_Temp_Open( PGSd_IO_Gen_NoEndurance, lun, PGSd_IO_Gen_Write, &fd ) )!= PGS_S_SUCCESS ) {
       sprintf(msg, "opening temp file, LUN %d", lun);
       modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
      }
   fseek(fd, 0L, SEEK_SET);
   for(j=0;j<16;j++) {
       fwrite(data[j], 2, total_n_pixels, fd);
      }
      
   if( ( ret = PGS_IO_Gen_Close( fd )) != PGS_S_SUCCESS ) {
       sprintf(msg, "closing temp file, LUN %d", lun);
       modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
      }
      
#else
   for(j=0;j<16;j++) {
      sprintf(name, "1km %s\0", sdsname[(j+SDSINDEX)-BAND20ALB]);
      sdindex = SDnametoindex(all_sd[i], name);
      sds = SDselect(all_sd[i], sdindex);
      cdef.chunk_lengths[0] = cdef.comp.chunk_lengths[0] = n_1kmlines_file[i];
      cdef.chunk_lengths[1] = cdef.comp.chunk_lengths[1] = NBSPOTS_1KM;
      cdef.comp.comp_type = COMP_CODE_DEFLATE;
      cdef.comp.cinfo.deflate.level = 8;

      if ((sdindex == -1) || (sds == -1)) {
           sprintf(msg,  "SDS '%s' not found in file ", name);
	    modsmf(MODIS_E_FILE_ACCESS, msg, location);
	    return(1);
          }
      
      /*  This gives me an automatic core dump on patmos. */
      if (set_chunking) SDsetchunk(sds, cdef, (HDF_CHUNK | HDF_COMP));
      dims[0] = n_1kmlines_file[i];
      start[0] = start[1] = 0;
      dims[1] = NBSPOTS_1KM;
      SDwritedata(sds, start, NULL, dims, (VOID *)data[j]);
      SDendaccess(sds);

     }
#endif     
return(0);     
}
		    

int tr_interpol_routine(int n_lines, int16 *data[16], uint8 *qa, 
                        int ln_lines, int16 *ldata[16], uint8 *lqa, 
                        int nn_lines, int16 *ndata[16], uint8 *nqa, uint16 *cm)
/*
!C*****************************************************************************
!Description:
	Yet another interface for the transmission, 
	path radiance and spherical albedo's interpolation.

!Input Parameters:
        none   

!Input/Output Parameters:
        n_lines      The number of lines in data and qa
	data         The vectorized data being processed (TR|PR|SA)
	qa           The associated QA
        ln_lines     SAME FOR THE PRECEDING GRANULE
	ldata        
	lqa         
        nn_lines     SAME FOR THE FOLLOWING GRANULE
	ndata        
	nqa         
	cm           contents of "1km Atmospheric Optical Depth Band QA" SDS, to be updated.
	             Only for the qa/data granule, not for preceding/following...        
	
!Output Parameters:
        none
	
!Returns:
         0 success
        -1 error

!Revision History:

	Original version March 2006, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:
!References and Credits
!Design Notes

	Makes the preceding, current and next granules' data/qa into 
	swathlettes; does interpolation; then cuts out the right data 
	again;
	
!END

****************************************************************************/
{
#define NNN 50
int i, j, k, DS, TIERMAX, total_pixels, do_qa, n_interp, row, col, n_cols;
int16 **avg_data;
long l, in_dex, start, realstart;
uint8 *avg_qa;
uint8 *failmask;
uint8 *dummy;
int16 *localdata[16];
uint8 *localqa;
int ds_lines, ds_cols, total_lines;
int16 **TRdownsample(int16 **data, int *n_row, int *n_col, int factor);
void TRupsample(int16 **data, uint8 *fail, int n_lines, int n_cols, int16 **avg_data, 
              int factor);
uint8 *downsampleqa(uint8 *qa, int *n_row, int *n_col, int factor);
void TR_interpolate(uint8 *qa, int16 **data, uint8 *fail, int n_row, int n_col, int TIERMAX);
char location[256] = "aerosols.c, tr_interpol_routine\0";
void update_failmask(uint8 *failmask, uint8 *dsfailmask, int n_lines, int n_cols, int factor);
uint8 *TBImask, qaval;  /* New, 31-MAY-06 */

DS=10;          /* 10 -> 15, 22-JAN-02; 15 -> 10, 30-OCT-02. */
TIERMAX=15;     /* 20 -> 24, 22-JAN-02; 24 -> 15, 30-OCT-02. */

n_cols = NBSPOTS_1KM;

/*printf("%d %d %d\n", n_lines, nn_lines, ln_lines);*/

if (n_lines == 0) return(-1);

total_lines = n_lines;
if (nn_lines > 0) total_lines += NNN;
if (ln_lines > 0) total_lines += NNN;

total_pixels = total_lines*n_cols;  

localqa = (uint8 *)AllocMem(total_lines*NBSPOTS_1KM*sizeof(uint8), location);   
for(j=0;j<16;j++) {
   localdata[j] = (int16 *)AllocMem(total_lines*NBSPOTS_1KM*sizeof(int16), location); 
   }

/* New, 31-MAY-06 */
TBImask = (uint8 *)CAllocMem(n_lines*NBSPOTS_1KM,sizeof(uint8));   
for (l=0L;l<(n_lines*NBSPOTS_1KM);l++) {
     qaval = qa[l];
     if ((qaval != 0)&&(qaval != 6)&&(qaval != 15)&&(qaval != 4)&&(qaval != 9)&&(qaval != 19)) 
	 TBImask[l] = 1;   /* to-be-interpolated */
    }

/* Make a swathlette 
 ************************/
realstart = start = 0L;
if (total_lines == n_lines) {   /* one granule being processed; do essentially nothing */
   for (l=0L;l<n_lines*n_cols;l++) {
      localqa[l] = qa[l];
      for(j=0;j<16;j++) localdata[j][l] = data[j][l];
      }
   }
else {

   if (ln_lines > 0) {  /* Take their last fifty */
       l = 0;
       for (i=ln_lines-NNN;i<ln_lines;i++) {
          in_dex = i*n_cols;
	  for (k=0;k<n_cols;k++) localqa[l++] = lqa[in_dex+k];
          }
       start = l;
       realstart = l;
       for(j=0;j<16;j++) {
          l = 0;
          for (i=ln_lines-NNN;i<ln_lines;i++) {
             in_dex = i*n_cols;
	     for (k=0;k<n_cols;k++) {
	        localdata[j][l++] = ldata[j][in_dex+k];
		}
             }
         }
      }  /* if (ln_lines > 0) */
      
    for (l=0L;l<n_lines*n_cols;l++) {
         localqa[start+l] = qa[l];
	 for(j=0;j<16;j++) localdata[j][start+l] = data[j][l];
         }
    start += n_lines*n_cols;	
     
    if (nn_lines > 0) {  /* Take their first fifty */
       l = 0;
       for (i=0;i<NNN;i++) {
          in_dex = i*n_cols;
	  for (k=0;k<n_cols;k++,l++) localqa[start+l] = nqa[in_dex+k];
          }
       for(j=0;j<16;j++) {
          l = 0;
          for (i=0;i<NNN;i++) {
             in_dex = i*n_cols;
	     for (k=0;k<n_cols;k++,l++) {
	        localdata[j][start+l] = ndata[j][in_dex+k];
		}
             }
         }
      }  /* if (nn_lines > 0) */
      
      
    }
   

  /* Make first pass of interpolation over data 
   ***********************************************/
  avg_qa = NULL;
  
  if (  (failmask = (uint8 *)calloc(total_pixels,sizeof(uint8))) == NULL)  {
      sprintf(msg,  " cannot allocate %d uint8's (%d by %d) ", 
	         total_pixels, total_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(data);
      return(4);
          }

  TR_interpolate(localqa, localdata, failmask, total_lines, n_cols, TIERMAX);
  
  /* 30-OCT-02: Do a second pass at high resolution, but lower tiermax value */
  if (  (dummy = (uint8 *)calloc(total_pixels,sizeof(uint8))) == NULL)  {
      sprintf(msg,  " cannot allocate %d uint8's (%d by %d) ", 
	         total_pixels, total_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(data);
      return(4);
          }
  TR_interpolate(failmask, localdata, dummy, total_lines, n_cols, TIERMAX/2);
  memcpy(failmask, dummy, total_pixels*sizeof(uint8));
  free(dummy);

  /* Average the data */
  ds_lines = total_lines;
  ds_cols  = n_cols;
  avg_data = TRdownsample(localdata, &ds_lines, &ds_cols, DS);

  /* Average the fail-mask */
  ds_lines = total_lines;
  ds_cols  = n_cols;
  avg_qa = downsampleqa(failmask, &ds_lines, &ds_cols, DS);

  /* Make second pass of interpolation over averaged data */
  if (  (dummy = (uint8 *)malloc(ds_lines*ds_cols*sizeof(uint8))) == NULL)  {
      sprintf(msg,  " cannot allocate %d uint8's (%d by %d) ", 
	         ds_lines*ds_cols, ds_lines, ds_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(data);
      return(4);
          }
  TR_interpolate(avg_qa, avg_data, dummy, ds_lines, ds_cols, TIERMAX*2); 

  /* new, 31-MAY-06 */
  update_failmask(failmask, dummy, total_lines, n_cols, DS);

  free(dummy);

  /* Fill in holes in full-resolution data from averaged, interpolated data */
  TRupsample(localdata, failmask, total_lines, n_cols, avg_data, DS);

  for(i=0;i<16;i++) {
     if (avg_data[i] != NULL) free(avg_data[i]);
     }
  if (avg_qa != NULL) free(avg_qa);
  
  if (cm != NULL) {
     /* 31-MAY-06: adjust the failmask to fit just the present granule... */
     if (realstart > 0) {
        for (l=0L;l<n_lines*n_cols;l++) {
           failmask[l] = failmask[realstart+l];
               }
           }
	   
     /* 31-MAY-06: Use the TBImask and the final failmask to determine which 
        cm pixels need to have their climatology bit unset. */
     for (l=0L;l<n_lines*n_cols;l++) {
     
        /*if (l == (1141*n_cols + 535)) {
	    printf("%d %d %d\n", cm[l], TBImask[l], failmask[l]);
	   }*/
     
        if ((TBImask[l] == 1)   /* if it was to be interpolated in the first place... */
               &&               /* and... */
	    (failmask[l] == 0)) /* it hasn't failed interpolation... */
	    cm[l] = cm[l] & ~QACLIMATOLOGY;  /* unset */
        }
     }
      
  free(failmask);
  free(TBImask);


/* Now snip out the data/qa arrays from localdata, localqa 
 ***********************************************************/
for (l=0L;l<n_lines*n_cols;l++) {
   qa[l] = localqa[realstart+l];
   for(j=0;j<16;j++) data[j][l] = localdata[j][realstart+l];
         }

/* and free localdata, localqa 
 *********************************/
free(localqa);
for(j=0;j<16;j++) free(localdata[j]);
   
   
return(0);
}


void update_failmask(uint8 *failmask, uint8 *dsfailmask, int n_lines, int n_cols, int factor)
{
int ib, i,ii,j,jj,k,newrow,newcol;
long l;

newrow = (int)(((float)n_lines)/((float)factor) + 0.5F);
newcol = (int)(((float)n_cols)/((float)factor) + 0.5F);

l = 0L;
for (i=0;i<n_lines;i++) {
   ii = i/factor;
   if (ii >= newrow) continue;
   k = ii*newcol;
   for (j=0;j<n_cols;j++,l++) {
      jj = j/factor;
      if (jj >= newcol) continue;
      
      if ( (failmask[l] == 1) && (dsfailmask[k+jj] != 1))
         failmask[l] = 0;
	 
   }
}
}


void TR_interpolate(uint8 *qa, int16 **data, uint8 *fail, int n_row, int n_col, int TIERMAX)
{
/*
!C******************************************************************

!Description: This function interpolates vectorized transmission, 
              path radiance and spherical albedo data.

!Input Parameters:
qa:	              Key to the interpolation step. 
data:		      Data to be interpolated.
fail:                 A failed-mask, which indicates which pixels have failed to interpolate.
row:                  Rows and
col:                     columns of data/qa.
TIERMAX:              How many tiers of interpolation to try (at the most)

!Output Parameters:
none; 

!Returns:
nothing


!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov)


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
     This routine is only meant to interpolate aerosols generated with routine 
     aerosol_retrieval(), in file atmcor.c.  1-km versions of the retrieved aerosols
     are written to MOD09 files during the aerosol-retrieval part of PGE11 (the first 
     pass).  These are to be read from the MOD09 files (in tr_interpolation())
     and interpolated.  
     
     Values to be interpolated are data values which do not have values 0 or 6 OR 15 OR 4 in 
     their corresponding qa.  
     Values used in the interpolation are those which have either a 0 or a 6 in the
     qa.
     
!END****************************************************************************
*/
#define TIERMX 60
#define TMSQ   3721   /* (TIERMAX+1)*(TIERMAX+1) */
int TMSQ1;
int ib, i, j, ii, jj, d, count, tier;
float n;
int iin_col;
int upindex,dnindex,rtindex,lfindex;
float32 f32[16];
long l,k;
long UL[TMSQ],UR[TMSQ],LL[TMSQ],LR[TMSQ];
long UP[TIERMX+1],DN[TIERMX+1],LF[TIERMX+1],RT[TIERMX+1];
float wUL[TMSQ],wUR[TMSQ],wLL[TMSQ],wLR[TMSQ];
float wUP[TIERMX+1],wDN[TIERMX+1],wLF[TIERMX+1],wRT[TIERMX+1], weight;
int corner_got[4], corn, side, side_got[4];
short failed, lastfailed;
int getout;
long FUR, FLR;
int crs = 0;
uint8 qaval;

if (n_col < 1000) crs = 1;

if (TIERMAX > TIERMX) return;

TMSQ1 = (TIERMAX+1)*(TIERMAX+1);

/****************/
              l = -1L;
              for (i=0;i<n_row;i++) {
                 for (j=0;j<n_col;j++) {
                     l++;
		     lastfailed=failed;
                     failed = 0;
		     
                     /* forget corners, edges */
                     if ((i==0)||(j==0)||(j == (n_col-1))||(i == (n_row-1))) {
		        fail[l] = qa[l];  /* 21-MAR-06 */
		        continue;
			}
                      
	     /*if ((i == 1895) && (j == 899)) printf("%d\n", qa[l]);*/

                     /* skip interpolation of good data (qa == 0 or 6) and (as of 08-NOV-02) over clouds 
		        and (as of 26-MAR-03) over desert (qa == 15) *  /
		        and (as of 09-FEB-05) over adj to cloud (qa == 19) */
	             /* 08-AUG-03: skip also over high solar zenith regions (qa == 9).  Yes, I know 
		        such regions need to be filled in, but they very seldom occur in isolation -- 
			no islands of high SZ surrounded by low SZ -- so they are very difficult to 
			fill in. */
		     qaval = qa[l];
                     if ((qaval != 0)&&(qaval != 6)&&(qaval != 15)&&(qaval != 4)&&(qaval != 9)&&(qaval != 19)) {
		        failed = 1;

                        for (ii=0;ii<TMSQ1;ii++) {    
                           UL[ii] = -1;
                           UR[ii] = -1;
                           LL[ii] = -1;
                           LR[ii] = -1;
                           wUL[ii] = 0.0F;
                           wUR[ii] = 0.0F;
                           wLL[ii] = 0.0F;
                           wLR[ii] = 0.0F;
                        }
                        for (ii=0;ii<(TIERMAX+1);ii++) {    
                           UP[ii] = -1;
                           DN[ii] = -1;
                           LF[ii] = -1;
                           RT[ii] = -1;
                           wUP[ii] = 0.0F;
                           wDN[ii] = 0.0F;
                           wLF[ii] = 0.0F;
                           wRT[ii] = 0.0F;
                        }

                        for (tier=1;tier<TIERMAX;tier++) {    /* tier distance */

                           for(ib=0;ib<16;ib++) f32[ib] = 0.0F;
                           n = 0.0F;
                           for(jj=0;jj<4;jj++) {
			       corner_got[jj] = 0;
			       side_got[jj] = 0;
			       }

                           /* set up indices for corners, sides */
                           d = 0;
                           for (ii=1;ii<tier+1;ii++) {
			      iin_col = ii*n_col;
                              for (jj=1;jj<tier+1;jj++) {
                                  
                                  /* forget edges */
                                  if( ((i-ii) < 0) || ((i+ii) >= n_row) ||
                                      ((j-jj) < 0) || ((j+jj) >= n_col) ) continue;
 
				 weight = 1.0F;
                                 if (crs)
				    weight /= ((float)(ii+jj+2));
				 
                                 UL[d] = l - (iin_col) - (jj);
                                 UR[d] = l - (iin_col) + (jj);
                                 LL[d] = l + (iin_col) - (jj);
                                 LR[d] = l + (iin_col) + (jj);
                                 wUL[d] = weight;
                                 wUR[d] = weight;
                                 wLL[d] = weight;
                                 wLR[d] = weight;
                                 d++;
                              }
                           }
                           upindex = 0;
                           dnindex = 0;
                           lfindex = 0;
                           rtindex = 0;
                           for (ii=1;ii<tier+1;ii++) {
			       iin_col = ii*n_col;
			       weight = 1.0F;
                               if (crs)
				   weight /= (float)(ii+1);
				   
                               if ( (i-ii) > -1 )    {
			           UP[upindex] = l - (iin_col); 
			           wUP[upindex++] = weight; 
				   }
                               if ( (i+ii) < n_row ) {
			           DN[dnindex] = l + (iin_col); 
			           wDN[dnindex++] = weight; 
				   }
                               if ( (j-ii) > -1 )    {
			           LF[lfindex] = l - ii;
			           wLF[lfindex++] = weight;
				   } 
                               if ( (j+ii) < n_col ) {
			           RT[rtindex] = l + ii;
			           wRT[rtindex++] = weight;
				   } 
                           }                          
                           
                           /* sum up all non-interpolated data until all corners have something */
                           /* first, add the sides */
                           for (ii=0;ii<upindex;ii++) {
			       qaval = qa[UP[ii]];
			       if ((qaval == 0)||(qaval == 6)) {
                                   for(ib=0;ib<16;ib++)f32[ib] += (float32)data[ib][UP[ii]]*wUP[ii];
				   side_got[0]++;
                                   n+=wUP[ii];
                                    }
				    }
                           for (ii=0;ii<dnindex;ii++) {
			       qaval = qa[DN[ii]];
			       if ((qaval == 0)||(qaval == 6)) {
                                   for(ib=0;ib<16;ib++)f32[ib] += (float32)data[ib][DN[ii]]*wDN[ii];
				   side_got[1]++;
                                   n+=wDN[ii];
                                    }
				    }
                           for (ii=0;ii<lfindex;ii++) {
			       qaval = qa[LF[ii]];
			       if ((qaval == 0)||(qaval == 6)) {
				   side_got[2]++;
                                   for(ib=0;ib<16;ib++)f32[ib] += (float32)data[ib][LF[ii]]*wLF[ii];
                                   n+=wLF[ii];
                                    }
				    }
                           for (ii=0;ii<rtindex;ii++) {
			       qaval = qa[RT[ii]];
 			       if ((qaval == 0)||(qaval == 6)) {
                                   for(ib=0;ib<16;ib++)f32[ib] += (float32)data[ib][RT[ii]]*wRT[ii];
				   side_got[3]++;
                                   n+=wRT[ii];
                                    }
				    }
/*if ((i==28)&&(j==15))
   printf("tier %d: UP side %d, DN side %d, LF side %d, RT side %d, %f %f\n", 
      tier,side_got[0], side_got[1], side_got[2], side_got[3], f32/n, n );
*/
                           for (ii=0;ii<tier*tier;ii++) {   
                               if (UL[ii] != -1) {
			          qaval = qa[UL[ii]];
			          if ((qaval == 0)||(qaval == 6)) {
                                      corner_got[0]++;
                                      for(ib=0;ib<16;ib++)f32[ib] += (float32)data[ib][UL[ii]]*wUL[ii];
                                      n+=wUL[ii];
                                       }
				    }
                               if (UR[ii] != -1) {
			          qaval = qa[UR[ii]];
			          if ((qaval == 0)||(qaval == 6)) {
                                      corner_got[1]++;
                                      for(ib=0;ib<16;ib++)f32[ib] += (float32)data[ib][UR[ii]]*wUR[ii];
                                      n+=wUR[ii];
                                       }
				    }
                               if (LL[ii] != -1) {
			          qaval = qa[LL[ii]];
			          if ((qaval == 0)||(qaval == 6)) {
                                      corner_got[2]++;
                                      for(ib=0;ib<16;ib++)f32[ib] += (float32)data[ib][LL[ii]]*wLL[ii];
                                      n+=wLL[ii];
                                       }
				    }
                               if (LR[ii] != -1) {
			          qaval = qa[LR[ii]];
			          if ((qaval == 0)||(qaval == 6)) {
                                      corner_got[3]++;
                                      for(ib=0;ib<16;ib++)f32[ib] += (float32)data[ib][LR[ii]]*wLR[ii];
                                      n+=wLR[ii];
                                       }
				    }
                           }
                           
/*if ((i==28)&&(j==15))
   printf("tier %d: UL corner %d, UR corner %d, LL corner %d, LR corner %d, %f %f\n", 
      tier,corner_got[0], corner_got[1], corner_got[2], corner_got[3], f32/n, n );
 */     
                           /* figure out if each corner has been sampled */
			   if (crs) {
                             corn = 0;
			     side = 0;
                           }
			   else {
                             corn = 1;
			     side = 1;
			   }
			   
                           for (ii=0;ii<4;ii++) {
			      if (crs) {
			         corn += (corner_got[ii]!=0);
			         side += (side_got[ii]!=0);
			      }
			      else {
			         corn *= corner_got[ii];
			         side *= side_got[ii];
			      }
                           }
			   if (crs) {
                              if ( ((corn > 1)||(side > 1)) &&
			            (n > 0.50F)) {   
                                 for(ib=0;ib<16;ib++)data[ib][l] = (uint16)(f32[ib]/n);
			         /*qa[l] = 6;*/
			         failed = 0;
                                 break;  /* to get us out of the "for (tier=1...) loop */
                              }
                           }
			   else {
                              if ((corn != 0)||(side != 0)) {   /* and if so... */
                                 for(ib=0;ib<16;ib++)data[ib][l] = (uint16)(f32[ib]/n);
			         /*qa[l] = 6;*/
			         failed = 0;
                                 break;  /* to get us out of the "for (tier=1...) loop */
                              }
                           }
                        } /* for (tier=1;tier<TIERMAX;tier++) */
			
			
                     }  /* if (qaval == ...) */
		     
		     fail[l] = (uint8)failed; 
		     
                  }  /* for (j=0;j<n_col;j++) */
              }  /* for (i=0;i<n_row;i++) */


}



int16 **TRdownsample(int16 **data, int *n_row, int *n_col, int factor)
{
uint8 *n_avg;
int16 **asdf;
int16 **ds_data;
int32 sum;
int16 avg, oldavg;
int ib, i,ii,j,jj,k,newrow,newcol,factor_squared;
int8 n, orign;
float SD;
long l;

newrow = (int)(((float)*n_row)/((float)factor) + 0.5F);
newcol = (int)(((float)*n_col)/((float)factor) + 0.5F);

asdf = (int16 **)malloc(newrow*newcol*sizeof(int16 *));
factor_squared = factor*factor + 5;
for(i=0;i<newrow*newcol;i++) 
    asdf[i] = calloc(factor_squared,sizeof(int16));

ds_data = (int16 **)malloc(16*sizeof(int16*));
for(i=0;i<16;i++) ds_data[i] = (int16 *)calloc(newrow*newcol,sizeof(int16));

/* collect the data */
for(ib=0;ib<16;ib++) {
n_avg = (uint8 *)calloc(newrow*newcol,sizeof(uint8));
l = 0L;
for (i=0;i<*n_row;i++) {
   ii = i/factor;
   if (ii >= newrow) continue;
   k = ii*newcol;
   for (j=0;j<*n_col;j++,l++) {
      jj = j/factor;
      if (jj >= newcol) continue;
      asdf[k+jj][n_avg[k+jj]] = data[ib][l];    
      n_avg[k+jj]++;
   }
}

/* get avg, SD of data */
for (i=0;i<newrow*newcol;i++) {
   avg = 0;
   SD = 0.0F;
   if (n_avg[i] > 0) {
         /* get avg */
	 sum = (int32)0;
	 orign = 0;
	 for(ii=0;ii<n_avg[i];ii++) {
	    if (asdf[i][ii] > 0) {
	        sum += (int32)asdf[i][ii];
		orign++;
		}
	    }
         if (orign > 0)
	    avg = (int16)(sum/(int32)orign);
	 else  
	    /* avg is still zero!  Just set it to anything!  */
	    avg = (int16)(int32)asdf[i][0];

         /* get sigma */
	 if (orign > 1) {
	    sum = (int32)0;
	    for(ii=0;ii<n_avg[i];ii++) {
	        if (asdf[i][ii] > 0) {
	            sum += ((int32)asdf[i][ii] - avg)*((int32)asdf[i][ii] - avg);
		    }
		}
	    sum /= (int32)(orign - 1);
 	    SD = sqrtf((float)sum);
	    }
	 
	 /* recompute avg, rejecting values beyond SD from old avg */
	 oldavg = avg;
	 sum = (int32)0;
	 avg += (int16)SD;
	 n = 0;
	 for(ii=0;ii<n_avg[i];ii++) {
	     if ((asdf[i][ii] < avg)&&(asdf[i][ii] > 0)) {
	         sum += (int32)asdf[i][ii];
		 n++;
		 }
	     }	     
	 if (n > 0)
            avg = (int16)(sum/(int32)n);
	 else 
	    avg = (int16)oldavg;   /* This had been set to ZERO!! Can't do that!!! 
	                              This is where a lot of artifacts had been coming from (31-MAR-06) */
        } 

   ds_data[ib][i] = avg;
   if (ib==15)free(asdf[i]);
   }
   free(n_avg);
   
      }  /* for(ib=0;ib<16;ib++) */
free(asdf);
*n_row = newrow;
*n_col = newcol;
return(ds_data);
}


void TRupsample(int16 **data, uint8 *fail, int n_lines, int n_cols, int16 **avg_data, 
              int factor)
{
int ib, i,ii,j,jj,k,newrow,newcol;
long l;

newrow = (int)(((float)n_lines)/((float)factor) + 0.5F);
newcol = (int)(((float)n_cols)/((float)factor) + 0.5F);

l = 0L;
for (i=0;i<n_lines;i++) {
   ii = i/factor;
   if (ii >= newrow) continue;
   k = ii*newcol;
   for (j=0;j<n_cols;j++,l++) {
      jj = j/factor;
      if (jj >= newcol) continue;
      
      if ( (fail[l] == 1) && (avg_data[0][k+jj] != 0))
         for(ib=0;ib<16;ib++) data[ib][l] = avg_data[ib][k+jj];
	 
   }
}
}

/* END 31-MAR-06: New stuff for transmission, path radiance, spherical albedo interpolation */


int aot_interpolation(int refout, dims_struct ***outputs /*char *ref_product_file[]*/)
{
/*
!C******************************************************************

!Description: This function sets up retrieved aerosols for interpolation.

!Input Parameters:
refout:	              the maximum number of files with aerosols to be interpolated
ref_product_file:     a list of files with the aerosols.  Some of these strings may be
                      NULL-valued; these are to be skipped (obviously).


!Output Parameters:
none; returns 0 for success, nonzero for error.


!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov)


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
     This routine is only meant to interpolate aerosols generated with routine 
     aerosol_retrieval(), in file atmcor.c.  1-km versions of the retrieved aerosols
     are written to MOD09 files during the aerosol-retrieval part of PGE11 (the first 
     pass).  These are to be read from the MOD09 files (listed in ref_product_files)
     and interpolated.  
     
     This routine looks for an SDS with a "qa" in its title, and interpolates the rest
     of the SDSs according to the values found in the "qa" SDS.
     
     The actual interpolation happens in routine AOT_interpolate(), below.
     
     This interpolation is done on a full swath: i. e., all granules are read in 
     sequence into a swath, which is then interpolated.

!END****************************************************************************
*/
int i,j,ii, got_qa, n_sds, n_lines, n_cols, ds_lines, ds_cols;
int32 sd, sds, cmsds;
int32 start[2], dims[2];
int32 n_sets1, n_gattr1, rank1, dims1[2], number_type1, nattr1;
int32 each_files_lines[30], max_lines;
char sds_to_interpolate[10][256];
char msg[256];
char name[256];
char QAname[256];
char location[40]="aerosols.c, aot_interpolation\0";
int16 *data;
int16 *avg_data;
uint8 *avg_qa;
uint8 *qa;
uint8 *failmask;
uint8 *dummy;
int16 *data1;
uint8 *qa1;
int DS, TIERMAX;
char *ref_product_file[30];
void update_cm(uint16 *cm, int16 *olddata, int16 *data1, int32 each_files_lines, int32 n_cols);

uint16 *cm;
int16 *olddata;

uint8 *aotmodel, *aotmodel1, *copy_aotmodel;   /* 01-NOV-05 */
char MODELname[256];
int16 **subdata;
uint8 **subqa;
uint8 model, maxmodels, qaval, models[MODELS], n_each_model[MODELS+1], we_have_this_model[MODELS+1];
int got_model, total_pixels, do_qa, n_interp, row, col, max_n, max_mod;
int16 values[MODELS];


/* new, 22-JAN-02 */
DS=10;          /* 10 -> 15, 22-JAN-02; 15 -> 10, 30-OCT-02. */
TIERMAX=15;     /* 20 -> 24, 22-JAN-02; 24 -> 15, 30-OCT-02. */

/* Get names of SDSs to look for from first file */
got_qa = 0;
got_model = 0;   /* 01-NOV-05 */
n_sds = 0;

for (i=0;i<refout;i++) {

   if (outputs[0][i]->filename != NULL) {
      ref_product_file[i] = (char *)malloc(strlen(outputs[0][i]->filename)+1);
      strcpy(ref_product_file[i], outputs[0][i]->filename);
      }
   else ref_product_file[i] = (char *)NULL;
   
   }

for (i=0;i<refout;i++) {

   if (ref_product_file[i] != NULL) {

       if ( (sd = SDstart(ref_product_file[i], DFACC_RDONLY)) == -1) {
	  sprintf(msg, "'%s' (aerosol optical thickness)", ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
	  return(1);
       }
      SDfileinfo(sd, &n_sets1, &n_gattr1);
      if (n_sets1 == 0) {
	  sprintf(msg,  "file %s (aerosol optical thickness) is unreadable ", ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
          SDend(sd);
	  return(1);
       }
      
      for (j=0;j<n_sets1;j++) {
          sds = SDselect(sd, j);
          SDgetinfo(sds, name, &rank1, dims1, &number_type1, &nattr1);
	  SDendaccess(sds);
	  /*for (ii=0;ii<strlen(name);ii++)name[ii] = toupper(name[ii]);*/
	  if (strstr(name, "1km Atmospheric Optical Depth Band CM"))  got_qa = 1;   /* 5.0.2 */
	  if (strstr(name, "Model")) got_model = 1;   /* 01-NOV-05 */
      }  /* for (j=0;j<n_sets1;j++) */


      if ((!got_qa)||(!got_model)) {
	  sprintf(msg,  "file %s (aerosol optical thickness) doesn't have QA or Model SDS", ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
          SDend(sd);
	  return(1);
       }
      else {
        for (j=0;j<n_sets1;j++) {
          sds = SDselect(sd, j);
          SDgetinfo(sds, name, &rank1, dims1, &number_type1, &nattr1);
	  SDendaccess(sds);
	  /*for (ii=0;ii<strlen(name);ii++)name[ii] = toupper(name[ii]);*/
	  if (strstr(name, "1km Atmospheric Optical Depth Band CM"))    /* 5.0.2 */
	      strcpy(QAname, name);

	  else if (strstr(name, "Model"))    /* 01-NOV-05 */
	      strcpy(MODELname, name);

          else if((strstr(name, "1km Atmospheric Optical Depth Band 3"))|| 
	          (strstr(name, "1km Atmospheric Optical Depth Band 8")))   /* do NOT interpolate "Band 1" */
	      strcpy(sds_to_interpolate[n_sds++], name);

	}
       }
       
      if (n_sds == 0) {
         sprintf(msg,  "file %s (aerosol optical thickness) is unreadable ", ref_product_file[i]);
         modsmf(MODIS_E_FILE_ACCESS, msg, location);
         SDend(sd);
         return(1);
      }
      SDend(sd);
      
      break;
   }
      }  /* for (i=0;i<refout;i++) */

/* Let's assume all AOT SDSs to be interpolated are in contiguous files, no gaps. */
/* Let's also assume all AOT SDSs in each file are of the same dimensions.  
   Get the total number of lines to be processed.  Also make sure each file has
   all the required SDSs.  */
n_cols = 0;
max_lines = 0;
for (i=0;i<refout;i++) {
   each_files_lines[i] = 0;

   if (ref_product_file[i] != NULL) {
       if ( (sd = SDstart(ref_product_file[i], DFACC_RDONLY)) == -1) {
	  sprintf(msg, "'%s' (aerosol optical thickness) ", ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
	  return(2);
       }
      SDfileinfo(sd, &n_sets1, &n_gattr1);
      if (n_sets1 == 0) {
	  sprintf(msg,  "file %s (aerosol optical thickness) is unreadable ", ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
          SDend(sd);
	  return(2);
       }

      if ( (sds = SDselect(sd, SDnametoindex(sd, MODELname))) == -1) {
	  sprintf(msg,  "SDS %s not found in file %s (aerosol optical thickness) ", 
	      MODELname, ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
          SDend(sd);
	  return(2);
       }
     
      if ( (sds = SDselect(sd, SDnametoindex(sd, QAname))) == -1) {
	  sprintf(msg,  "SDS %s not found in file %s (aerosol optical thickness) ", 
	      QAname, ref_product_file[i]);
	  modsmf(MODIS_E_FILE_ACCESS, msg, location);
          SDend(sd);
	  return(2);
       }
     
       SDgetinfo(sds, name, &rank1, dims1, &number_type1, &nattr1);
       SDendaccess(sds);
       each_files_lines[i] = dims1[0];
       if (each_files_lines[i] > max_lines) max_lines = each_files_lines[i];
       n_cols = dims1[1];

       for (j=0;j<n_sds;j++) {
         if ( (sds = SDselect(sd, SDnametoindex(sd, sds_to_interpolate[j]))) == -1) {
	     sprintf(msg,  "SDS %s not found in file %s (aerosol optical thickness) ", 
	         sds_to_interpolate[j], ref_product_file[i]);
	     modsmf(MODIS_E_FILE_ACCESS, msg, location);
             SDend(sd);
	     return(2);
          }
	 else
             SDendaccess(sds);
       }
       
      SDend(sd);
      
   }  /* if (ref_product_file[i] != NULL)  */
} /* for (i=0;i<refout;i++) */

n_lines = 0;
for (i=0;i<refout;i++)
   n_lines += each_files_lines[i];

/* Allocate memory! */
data       = (int16 *)NULL;  /* These must be initialized */
qa         = (uint8 *)NULL;
aotmodel   = (uint8 *)NULL;
failmask   = (uint8 *)NULL;
/*next line added by C. Davidson on 11/15/05 per direction from Jim Ray*/  /* the sole v5.2.1 change */
copy_aotmodel = (uint8 *)NULL;
 
total_pixels = n_lines*n_cols;
if (n_lines > 0) {
   if (  (data = (int16 *)malloc(total_pixels*sizeof(int16))) == NULL)  {
      sprintf(msg,  " cannot allocate %d int16's (%d by %d) ", 
	         total_pixels, n_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      return(4);
          }
   if (  (qa = (uint8 *)malloc(total_pixels*sizeof(uint8))) == NULL)  {
      sprintf(msg,  " cannot allocate %d uint8's (%d by %d) ", 
	         total_pixels, n_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(data);
      return(4);
          }	  
   /* Added 01-NOV-05 */
   if (  (aotmodel = (uint8 *)malloc(total_pixels*sizeof(uint8))) == NULL)  {
      sprintf(msg,  " cannot allocate %d uint8's (%d by %d) ", 
	         total_pixels, n_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(data);
      return(4);
          }
   if (  (copy_aotmodel = (uint8 *)malloc(total_pixels*sizeof(uint8))) == NULL)  {
      sprintf(msg,  " cannot allocate %d uint8's (%d by %d) ",
                 total_pixels, n_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(data);
      return(4);
          }
	  
   /* Added 26-JUN-01 */
   if (  (failmask = (uint8 *)malloc(total_pixels*sizeof(uint8))) == NULL)  {
      sprintf(msg,  " cannot allocate %d uint8's (%d by %d) ", 
	         total_pixels, n_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(data);
      return(4);
          }
}

data1 = (int16 *)NULL;
qa1   = (uint8 *)NULL;
aotmodel1   = (uint8 *)NULL;
olddata = (int16 *)NULL;
cm   = (uint16 *)NULL;
if (max_lines > 0) {
   if (  (data1 = (int16 *)malloc(max_lines*n_cols*sizeof(int16))) == NULL)  {
      sprintf(msg,  " cannot allocate %d int16's (%d by %d) ", 
	         max_lines*n_cols, max_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      return(4);
          }
   if (  (qa1 = (uint8 *)malloc(max_lines*n_cols*sizeof(uint8))) == NULL)  {
      sprintf(msg,  " cannot allocate %d uint8's (%d by %d) ", 
	         max_lines*n_cols, max_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(data1);
      return(4);
          }
   if (  (aotmodel1 = (uint8 *)malloc(max_lines*n_cols*sizeof(uint8))) == NULL)  {
      sprintf(msg,  " cannot allocate %d uint8's (%d by %d) ", 
	         max_lines*n_cols, max_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(data1);
      return(4);
          }
   /* new, 25-APR-05 */
   if (  (olddata = (int16 *)malloc(max_lines*n_cols*sizeof(int16))) == NULL)  {
      sprintf(msg,  " cannot allocate %d int16's (%d by %d) ", 
	         max_lines*n_cols, max_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      return(4);
          }
   if (  (cm = (uint16 *)malloc(max_lines*n_cols*sizeof(uint16))) == NULL)  {
      sprintf(msg,  " cannot allocate %d uint16's (%d by %d) ", 
	         max_lines*n_cols, max_lines, n_cols);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(data1);
      return(4);
          }
     }
avg_data = (int16 *)NULL;
avg_qa   = (uint8 *)NULL;
start[0] = start[1] = 0;

/* For each SDS to be processed, read, interpolate, write */
for (j=0;j<n_sds;j++) {
  n_lines = 0;
  for (i=0;i<refout;i++) {
      if (ref_product_file[i] != NULL) {

          if ( (sd = SDstart(ref_product_file[i], DFACC_RDONLY)) == -1) {
	     sprintf(msg, "'%s' (aerosol optical thickness) ", ref_product_file[i]);
	     modsmf(MODIS_E_FILE_ACCESS, msg, location);
	     return(2);
          }
          sds = SDselect(sd, SDnametoindex(sd, sds_to_interpolate[j]));
          HDF_Read_2D(sds, 0, each_files_lines[i], n_cols, (VOID *)data1);
	  SDendaccess(sds);
	  
	  if (j==0) {
             sds = SDselect(sd, SDnametoindex(sd, QAname));
             HDF_Read_2D(sds, 0, each_files_lines[i], n_cols, (VOID *)qa1);
	     SDendaccess(sds);
	     
	     /* Read the model ONCE in the beginning ... */
             sds = SDselect(sd, SDnametoindex(sd, MODELname));
             HDF_Read_2D(sds, 0, each_files_lines[i], n_cols, (VOID *)aotmodel1);
	     
	     SDendaccess(sds);
	  }
	  
          SDend(sd);
	  memcpy(&data[n_lines*n_cols], data1, each_files_lines[i]*n_cols*sizeof(int16));
	  if (j==0){
	      memcpy(&qa[n_lines*n_cols], qa1, each_files_lines[i]*n_cols*sizeof(uint8));
	      
	      /* Assemble the model ONCE in the beginning ... */
	      memcpy(&aotmodel[n_lines*n_cols], aotmodel1, each_files_lines[i]*n_cols*sizeof(uint8));
	      }
          n_lines += each_files_lines[i];

      }  /* if (ref_product_file[i] != NULL)  */
  } /* for (i=0;i<refout;i++) */

  /* 01-NOV-05: code for splitting AOT into separate subbands, according to the model */
  if (j == 0) {
     for (i=0;i<total_pixels;i++) {
        copy_aotmodel[i] = aotmodel[i];
	}
     }
  else if (j == 1) {
     for (i=0;i<total_pixels;i++) {
        aotmodel[i] = copy_aotmodel[i];
        if (aotmodel[i] > 0) aotmodel[i] = 1;
	}
     }
  subdata = (int16 **)malloc(MODELS*sizeof(int16 *));
  subqa = (uint8 **)malloc(MODELS*sizeof(uint8 *));
  for(i=0;i<MODELS;i++) {

   if (  (subdata[i] = (int16 *)malloc(total_pixels*sizeof(int16))) == NULL)  {
      sprintf(msg,  " cannot allocate %d int16's ", total_pixels);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(subdata);
      return(4);
          }
   if (  (subqa[i] = (uint8 *)malloc(total_pixels*sizeof(uint8))) == NULL)  {
      sprintf(msg,  " cannot allocate %d uint8's ", total_pixels);
      modsmf(MODIS_E_MEM_FAILURE, msg, location);
      free(subqa);
      return(4);
          }

   for(ii=0;ii<total_pixels;ii++) {
      subdata[i][ii] = AOT1_FILL;
      subqa[i][ii] = 1;
      }
   }  

/*printf("X %d\n", aotmodel[0]);*/
    
  maxmodels = 0;  
  for(ii=0;ii<total_pixels;ii++) {
     model = aotmodel[ii];
     if ((model < MODELS) && (model > maxmodels))maxmodels = model;
     if ((model < MODELS) && (model > 0)) {
        subdata[model-1][ii] = data[ii];
        subqa[model-1][ii] = qa[ii];
        }
     /* But we absolutely need to hold on to cloud information */
     /* and high-solar-zenith and desert information (17-NOV-05) */
     if ((qa[ii] == 4)||(qa[ii] == 19)||(qa[ii] == 9)||(qa[ii] == 15)) {
          for(i=0;i<MODELS;i++) subqa[i][ii] = qa[ii];
          }
     }    


for(ii=0;ii<maxmodels;ii++) {
  /*printf("Interpolating %s, model %d\n", sds_to_interpolate[j], ii+1); fflush(stdout);*/

  sprintf(msg, "interpolating %s, model %d\0", sds_to_interpolate[j], ii+1); 
  modsmf(MODIS_W_INFORMATION, msg, location);   /* This isn't actually a warning, but I'm hesitant to update the seed file */
  
  /* Make first pass of interpolation over data */
  memset(failmask, '\0', n_lines*n_cols*sizeof(uint8));
  AOT_interpolate(subqa[ii], subdata[ii], failmask, n_lines, n_cols, TIERMAX);
 
  /* 30-OCT-02: Do a second pass at high resolution, but lower tiermax value */
  if (  (dummy = (uint8 *)calloc(n_lines*n_cols,sizeof(uint8))) == NULL)  {
     sprintf(msg,  " cannot allocate %d uint8's ", n_lines*n_cols);
     modsmf(MODIS_E_MEM_FAILURE, msg, location);
     return(4);
         }

  AOT_interpolate(failmask, subdata[ii], dummy, n_lines, n_cols, TIERMAX/2);

  memcpy(failmask, dummy, n_lines*n_cols*sizeof(uint8));

  free(dummy);
 
   /* Average the data */
  ds_lines = n_lines;
  ds_cols  = n_cols;
  avg_data = downsample(subdata[ii], &ds_lines, &ds_cols, DS);

  /* Average the fail-mask */
  ds_lines = n_lines;
  ds_cols  = n_cols;
  avg_qa = downsampleqa(failmask, &ds_lines, &ds_cols, DS);

  /* Make second pass of interpolation over averaged data */
  dummy = (uint8 *)AllocMem(ds_lines*ds_cols*sizeof(uint8), location);
  AOT_interpolate(avg_qa, avg_data, dummy, ds_lines, ds_cols, TIERMAX*2); 
  free(dummy);

  /* Fill in holes in full-resolution data from averaged, interpolated data */
  upsample(subdata[ii], failmask, n_lines, n_cols, avg_data, DS);

  if (avg_data != NULL) free(avg_data);
  if (avg_qa != NULL) free(avg_qa);

}
  /* 01-NOV-05: code for reconstituting the AOT */
for(ii=0;ii<total_pixels;ii++) {
   n_interp = 0;
   
   if (aotmodel[ii] == 0) {   /* then this will have only interpolated data: I think */
      for (i=0;i<maxmodels;i++) {
         values[i] = 0;
         models[i] = 0;
         }
      for (i=0;i<MODELS+1;i++) we_have_this_model[i] = 0; 
      
      /* Then we MAY have an interpolated value somewhere in data[whatever] */
      for (i=0;i<maxmodels;i++) {
         if (subdata[i][ii] != AOT1_FILL) {
	    values[n_interp] = subdata[i][ii];
	    models[n_interp] = i+1;
	    we_have_this_model[i+1] = 1;
            n_interp++;
	    }
	 }

       if (n_interp > 0) { 
          if (n_interp == 1) { /* take it */
             data[ii] = values[0];
             aotmodel[ii] = models[0];
	     }
          else { 
	     /* Check each surrounding pixel for its model */
	     row = j/n_cols;
	     col = j%n_cols;
	     for (i=0;i<MODELS+1;i++) n_each_model[i] = 0;
	     
	     if ( col > 0 ) n_each_model[aotmodel[j-1]]++;
	     if ( col < n_cols ) n_each_model[aotmodel[j+1]]++;
	     if ( row > 0 ) {
	        if ( col > 0 ) n_each_model[aotmodel[j-n_cols-1]]++;
	        n_each_model[aotmodel[j-n_cols]]++;
	        if ( col < n_cols ) n_each_model[aotmodel[j-n_cols+1]]++;
	        }
	     if ( row < n_lines ) {
	        if ( col > 0 ) n_each_model[aotmodel[j+n_cols-1]]++;
	        n_each_model[aotmodel[j+n_cols]]++;
	        if ( col < n_cols ) n_each_model[aotmodel[j+n_cols+1]]++;
	        }
		
             /* Set to ZERO the n_each_model[i] values that are irrelevant to the 
	      * pixel in question  (sure, most of the surrounding pixels might be
	      * from model 1, but if we have to decide between model 2 and model 3, 
	      * forget that the max surrounding it is model 1...)
	      ************************************************************************/
	     for (i=1;i<MODELS+1;i++) {
                if (we_have_this_model[i] == 0) n_each_model[i] = 0;
	        }

             /* Find the plurality (forget zero) */
	     max_n = max_mod = -1;
	     /*printf("\n ");*/
	     for (i=1;i<MODELS+1;i++) {
	        /*printf(" %d - %d; ", i, n_each_model[i]);*/
	        if ((int16)n_each_model[i] > max_n) {
		   max_n = (int16)n_each_model[i];
		   max_mod = i;
		   }
		}
		
	     /*printf("\n ");*/
             /* Find which model array value matches max_mod */
             for (i=0;i<n_interp;i++) {
	       /*printf(" %d (M%d)", values[i], models[i]);*/
               if ((int16)models[i] == max_mod) {
                  data[ii] = values[i];
                  aotmodel[ii] = models[i];
		  max_mod = 99;
	            }
	         }
	      if (max_mod != 99) {  /* Then we failed: just take anything */
                 data[ii] = values[0];
                 aotmodel[ii] = models[0];
	        }
	       /*printf("\nTaken: %d (M%d)", data[ii], aotmodel[ii]);*/
 	     }
         }

      }
   }
   for(i=0;i<MODELS;i++) {
      if (subdata[i] != (int16 *)NULL) free(subdata[i]);
      if (subqa[i] != (uint8 *)NULL) free(subqa[i]);
      }   
   if (subdata != (int16 **)NULL) free(subdata);
   if (subqa != (uint8 **)NULL) free(subqa);
   
  /* end, 01-NOV-05 code for reconstituting the AOT */
    
  n_lines = 0;
  for (i=0;i<refout;i++) {
      if (ref_product_file[i] != NULL) {
      
          if (j==0) memcpy(aotmodel1, &aotmodel[n_lines*n_cols], each_files_lines[i]*n_cols*sizeof(uint8));
	  memcpy(data1, &data[n_lines*n_cols], each_files_lines[i]*n_cols*sizeof(int16));
          n_lines += each_files_lines[i];

          if ( (sd = SDstart(ref_product_file[i], DFACC_RDWR)) == -1) {
	     sprintf(msg, "'%s' (aerosol optical thickness) ", ref_product_file[i]);
	     modsmf(MODIS_E_FILE_ACCESS, msg, location);
	     return(2);
          }
	  
          sds = SDselect(sd, SDnametoindex(sd, sds_to_interpolate[j]));
 
          if (j==0) {  /* 25-APR-05: if j==0, update the CM, also */
             cmsds = SDselect(sd, SDnametoindex(sd, "1km Atmospheric Optical Depth Band QA"));
	     if (cmsds != -1) {
                HDF_Read_2D(cmsds, 0, each_files_lines[i], n_cols, (VOID *)cm);
		
		/* Update the CM by comparing interpolated data with uninterpolated data */
                HDF_Read_2D(sds, 0, each_files_lines[i], n_cols, (VOID *)olddata);
		update_cm(cm, olddata, data1, each_files_lines[i], n_cols);
                HDF_Write_2D(cmsds, 0, each_files_lines[i], n_cols, (VOID *)cm);
		SDendaccess(cmsds);
		
	         }
	      }
	      
          HDF_Write_2D(sds, 0, each_files_lines[i], n_cols, (VOID *)data1);
	  SDendaccess(sds);

          if (j==0) {  /* 01-NOV-05: if j==0, update the model, also */
             sds = SDselect(sd, SDnametoindex(sd, MODELname));
             HDF_Write_2D(sds, 0, each_files_lines[i], n_cols, (VOID *)aotmodel1);
	     SDendaccess(sds);
	      }
	  
          SDend(sd);
	  
      }  /* if (ref_product_file[i] != NULL)  */
  } /* for (i=0;i<refout;i++) */
  
} /* for (j=0;j<n_sds;j++) */

for (i=0;i<refout;i++) 
   if (ref_product_file[i] != NULL) 
      free(ref_product_file[i]);

if (aotmodel1 != NULL) free(aotmodel1);
if (aotmodel != NULL) free(aotmodel);
if (copy_aotmodel != NULL) free(copy_aotmodel);
   
if (data1 != NULL) free(data1);
if (qa1 != NULL) free(qa1);
if (data != NULL) free(data);
if (qa != NULL) free(qa);
if (olddata != NULL) free(olddata);
if (cm != NULL) free(cm);
if (failmask != NULL) free(failmask);
return(0);
}

void update_cm(uint16 *cm, int16 *olddata, int16 *data1, int32 each_files_lines, int32 n_cols)
{
int i,j;
long l = 0L;

for(i=0;i<each_files_lines;i++) {
   for(j=0;j<n_cols;j++) {
       if ((olddata[l] == AOT1_FILL) && (data1[l] != AOT1_FILL)) cm[l] = cm[l] & ~QACLIMATOLOGY;
       l++;
       }
   }
   
}


int16 *downsample(int16 *data, int *n_row, int *n_col, int factor)
{
/*
!C******************************************************************

!Description: This function averages the interpolated data from 
               routine AOT_interpolate() (below).

!Input Parameters:
qa:	              data. 
n_row:                Rows and
n_col:                   columns of data.
factor:               Factor by which data is to be downsampled.

!Output Parameters:
none; 

!Returns:
the averaged data

!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), June, 2001; updated, 30-OCT-02.


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
     This routine is only meant to average data interpolated in AOT_interpolate.
     New and improved, 30-OCT-02: routine now gets avg and SD of all non-zero data,
     and recomputes avg, rejecting values above (old avg+SD).
     
!END****************************************************************************
*/
uint8 *n_avg;
int16 **asdf;
int16 *ds_data;
int32 sum;
int16 avg;
int i,ii,j,jj,k,newrow,newcol,factor_squared;
int8 n, orign;
float SD;
long l;

newrow = (int)(((float)*n_row)/((float)factor) + 0.5F);
newcol = (int)(((float)*n_col)/((float)factor) + 0.5F);

asdf = (int16 **)malloc(newrow*newcol*sizeof(int16 *));
factor_squared = factor*factor + 5;
for(i=0;i<newrow*newcol;i++) 
    asdf[i] = calloc(factor_squared,sizeof(int16));

ds_data = (int16 *)calloc(newrow*newcol,sizeof(int16));
n_avg = (uint8 *)calloc(newrow*newcol,sizeof(uint8));

/* collect the data */
l = 0L;
for (i=0;i<*n_row;i++) {
   ii = i/factor;
   if (ii >= newrow) continue;
   k = ii*newcol;
   for (j=0;j<*n_col;j++,l++) {
      jj = j/factor;
      if (jj >= newcol) continue;
      asdf[k+jj][n_avg[k+jj]] = data[l];    
      n_avg[k+jj]++;
   }
}

/* get avg, SD of data */
for (i=0;i<newrow*newcol;i++) {
   avg = 0;
   SD = 0.0F;
   if (n_avg[i] > 0) {
         /* get avg */
	 sum = (int32)0;
	 orign = 0;
	 for(ii=0;ii<n_avg[i];ii++) {
	    if (asdf[i][ii] > 0) {
	        sum += (int32)asdf[i][ii];
		orign++;
		}
	    }
         if (orign > 0)
	    avg = (int16)(sum/(int32)orign);

         /* get sigma */
	 if (orign > 1) {
	    sum = (int32)0;
	    for(ii=0;ii<n_avg[i];ii++) {
	        if (asdf[i][ii] > 0) {
	            sum += ((int32)asdf[i][ii] - avg)*((int32)asdf[i][ii] - avg);
		    }
		}
	    sum /= (int32)(orign - 1);
 	    SD = sqrtf((float)sum);
	    }
	 
	 /* recompute avg, rejecting values beyond SD from old avg */
	 sum = (int32)0;
	 avg += (int16)SD;
	 n = 0;
	 for(ii=0;ii<n_avg[i];ii++) {
	     if ((asdf[i][ii] < avg)&&(asdf[i][ii] > 0)) {
	         sum += (int32)asdf[i][ii];
		 n++;
		 }
	     }	     
	 if (n > 0)
            avg = (int16)(sum/(int32)n);
	 else 
	    avg = (int16)0;
        } 

   ds_data[i] = avg;
   free(asdf[i]);
   }
      
free(asdf);
free(n_avg);
*n_row = newrow;
*n_col = newcol;
return(ds_data);
}

int16 *olddownsample(int16 *data, int *n_row, int *n_col, int factor)
{
/*
!C******************************************************************

!Description: This function averages the interpolated data from 
               routine AOT_interpolate() (below).  
	       (Obsolete, no longer used).

!Input Parameters:
qa:	              data. 
n_row:                Rows and
n_col:                   columns of data.
factor:               Factor by which data is to be downsampled.

!Output Parameters:
none; 

!Returns:
the averaged data

!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), June, 2001


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
     This routine is only meant to average data interpolated in AOT_interpolate.
     
!END****************************************************************************
*/
uint8 *n_avg;
int16 *ds_data;
int32 *summed_data;
int i,ii,j,jj,k,newrow,newcol;
long l;

newrow = (int)(((float)*n_row)/((float)factor) + 0.5F);
newcol = (int)(((float)*n_col)/((float)factor) + 0.5F);

summed_data = (int32 *)CAllocMem(newrow*newcol,sizeof(int32));
ds_data = (int16 *)CAllocMem(newrow*newcol,sizeof(int16));
n_avg = (uint8 *)CAllocMem(newrow*newcol,sizeof(uint8));

l = 0L;
for (i=0;i<*n_row;i++) {
   ii = i/factor;
   if (ii >= newrow) continue;
   k = ii*newcol;
   for (j=0;j<*n_col;j++,l++) {
      jj = j/factor;
      if (jj >= newcol) continue;
      summed_data[k+jj] += (int32)data[l];
      n_avg[k+jj]++;
   }
}
for (i=0;i<newrow*newcol;i++) {
   if (n_avg[i] != 0) 
         summed_data[i] /= (int32)n_avg[i];
   ds_data[i] = (int16) summed_data[i];
   }
      
free(n_avg);
free(summed_data);
*n_row = newrow;
*n_col = newcol;
return(ds_data);
}

void upsample(int16 *data, uint8 *fail, int n_lines, int n_cols, int16 *avg_data, 
              int factor)
/*
!C******************************************************************

!Description: This function upsamples the data by plugging values from the
               averaged, interpolated avg_data into the full resolution data 
	       where necessary.

!Input Parameters:
data:	              full-resolution data. 
fail:                 fail mask
n_lines:              Rows and
n_cols:                  columns of full-rez data/qa.
avg_data:             averaged, interpolated data.
factor:               Factor by which avg_data was downsampled.

!Output Parameters:
none; 

!Returns:
nothing

!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), June, 2001


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
     
!END****************************************************************************
*/
{
int i,ii,j,jj,k,newrow,newcol;
long l;

newrow = (int)(((float)n_lines)/((float)factor) + 0.5F);
newcol = (int)(((float)n_cols)/((float)factor) + 0.5F);

l = 0L;
for (i=0;i<n_lines;i++) {
   ii = i/factor;
   if (ii >= newrow) continue;
   k = ii*newcol;
   for (j=0;j<n_cols;j++,l++) {
      jj = j/factor;
      if (jj >= newcol) continue;
      
      if ( /* (data[l] == 0) && */ (fail[l] == 1) && (avg_data[k+jj] != 0))
         data[l] = avg_data[k+jj];
	 
   }
}
}

uint8 *downsampleqa(uint8 *qa, int *n_row, int *n_col, int factor)
{
/*
!C******************************************************************

!Description: This function downsamples the failed-mask generated in
               routine AOT_interpolate() (below).

!Input Parameters:
qa:	              Failed mask. 
n_row:                Rows and
n_col:                   columns of data/qa.
factor:               Factor by which mask is to be downsampled.

!Output Parameters:
none; 

!Returns:
the downsampled failed-mask

!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), June, 2001


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
     This routine is only meant to downsample the failed-mask generated in AOT_interpolate.
     The failed-mask indicates which pixels have failed to interpolate.
     
!END****************************************************************************
*/
uint8 *n;
uint8 **asdf;
int i,ii,j,jj,k,newrow,newcol,factor_squared,replace;
long l;
char location[40]="aerosols.c, downsampleqa\0";

newrow = (int)(((float)*n_row)/((float)factor) + 0.5F);
newcol = (int)(((float)*n_col)/((float)factor) + 0.5F);

asdf = (uint8 **)AllocMem(newrow*newcol*sizeof(uint8 *), location);
factor_squared = factor*factor;
for(i=0;i<newrow*newcol;i++) 
    asdf[i] = CAllocMem(factor_squared,sizeof(uint8));

n = (uint8 *)CAllocMem(newrow*newcol,sizeof(uint8));

l = 0L;
for (i=0;i<*n_row;i++) {
   ii = i/factor;
   if (ii >= newrow) continue;
   k = ii*newcol;
   for (j=0;j<*n_col;j++,l++) {
      jj = j/factor;
      if (jj >= newcol) continue;
      asdf[k+jj][n[k+jj]] = qa[l];
      n[k+jj]++;
   }
}

/* if any qa of the column is a 1, the column as a whole is 1 */
for(i=0;i<newrow*newcol;i++) {
    replace = 0;
    for(j=0;j<n[i];j++) 
        if (asdf[i][j] == 1) 
	    replace = 1;
	
    n[i] = replace;   /* n is now the downsampled failmask */
    free(asdf[i]);
}
free(asdf);

*n_row = newrow;
*n_col = newcol;
return(n);
}


void AOT_interpolate(uint8 *qa, int16 *data, uint8 *fail, int n_row, int n_col, int TIERMAX)
{
/*
!C******************************************************************

!Description: This function interpolates retrieved aerosols.

!Input Parameters:
qa:	              Key to the interpolation step. 
data:		      Data to be interpolated.
fail:                 A failed-mask, which indicates which pixels have failed to interpolate.
row:                  Rows and
col:                     columns of data/qa.
TIERMAX:              How many tiers of interpolation to try (at the most)

!Output Parameters:
none; 

!Returns:
nothing


!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov)


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
     This routine is only meant to interpolate aerosols generated with routine 
     aerosol_retrieval(), in file atmcor.c.  1-km versions of the retrieved aerosols
     are written to MOD09 files during the aerosol-retrieval part of PGE11 (the first 
     pass).  These are to be read from the MOD09 files (listed in argvs)
     and interpolated.  
     
     Values to be interpolated are data values which do not have values 0 or 6 OR 15 OR 4 in 
     their corresponding qa.  
     Values used in the interpolation are those which have either a 0 or a 6 in the
     qa.
     
!END****************************************************************************
*/
#define TIERMX 60
#define TMSQ   3721   /* (TIERMAX+1)*(TIERMAX+1) */
int TMSQ1;
int i, j, ii, jj, d, count, tier;
float n;
int iin_col;
int upindex,dnindex,rtindex,lfindex;
float32 f32;
long l,k;
long UL[TMSQ],UR[TMSQ],LL[TMSQ],LR[TMSQ];
long UP[TIERMX+1],DN[TIERMX+1],LF[TIERMX+1],RT[TIERMX+1];
float wUL[TMSQ],wUR[TMSQ],wLL[TMSQ],wLR[TMSQ];
float wUP[TIERMX+1],wDN[TIERMX+1],wLF[TIERMX+1],wRT[TIERMX+1], weight;
int corner_got[4], corn, side, side_got[4];
short failed, lastfailed;
int getout;
long FUR, FLR;
int crs = 0;
uint8 qaval;

if (n_col < 1000) crs = 1;

if (TIERMAX > TIERMX) return;

TMSQ1 = (TIERMAX+1)*(TIERMAX+1);

/****************/
              l = -1L;
              for (i=0;i<n_row;i++) {
                 for (j=0;j<n_col;j++) {
                     l++;
		     lastfailed=failed;
                     failed = 0;
		     
                     /* forget corners, edges */
                     if ((i==0)||(j==0)||(j == (n_col-1))||(i == (n_row-1))) continue;
                      
	     /*if ((i == 1895) && (j == 899)) printf("%d\n", qa[l]);*/

                     /* skip interpolation of good data (qa == 0 or 6) and (as of 08-NOV-02) over clouds 
		        and (as of 26-MAR-03) over desert (qa == 15) *  /
		        and (as of 09-FEB-05) over adj to cloud (qa == 19) */
	             /* 08-AUG-03: skip also over high solar zenith regions (qa == 9).  Yes, I know 
		        such regions need to be filled in, but they very seldom occur in isolation -- 
			no islands of high SZ surrounded by low SZ -- so they are very difficult to 
			fill in. */
		     qaval = qa[l];
                     if ((qaval != 0)&&(qaval != 6)&&(qaval != 15)&&(qaval != 4)&&(qaval != 9)&&(qaval != 19)) {
		        failed = 1;

                        for (ii=0;ii<TMSQ1;ii++) {    
                           UL[ii] = -1;
                           UR[ii] = -1;
                           LL[ii] = -1;
                           LR[ii] = -1;
                           wUL[ii] = 0.0F;
                           wUR[ii] = 0.0F;
                           wLL[ii] = 0.0F;
                           wLR[ii] = 0.0F;
                        }
                        for (ii=0;ii<(TIERMAX+1);ii++) {    
                           UP[ii] = -1;
                           DN[ii] = -1;
                           LF[ii] = -1;
                           RT[ii] = -1;
                           wUP[ii] = 0.0F;
                           wDN[ii] = 0.0F;
                           wLF[ii] = 0.0F;
                           wRT[ii] = 0.0F;
                        }

                        for (tier=1;tier<TIERMAX;tier++) {    /* tier distance */

                           f32 = 0.0F;
                           n = 0.0F;
                           for(jj=0;jj<4;jj++) {
			       corner_got[jj] = 0;
			       side_got[jj] = 0;
			       }

                           /* set up indices for corners, sides */
                           d = 0;
                           for (ii=1;ii<tier+1;ii++) {
			      iin_col = ii*n_col;
                              for (jj=1;jj<tier+1;jj++) {
                                  
                                  /* forget edges */
                                  if( ((i-ii) < 0) || ((i+ii) >= n_row) ||
                                      ((j-jj) < 0) || ((j+jj) >= n_col) ) continue;
 
				 weight = 1.0F;
                                 if (crs)
				    weight /= ((float)(ii+jj+2));
				 
                                 UL[d] = l - (iin_col) - (jj);
                                 UR[d] = l - (iin_col) + (jj);
                                 LL[d] = l + (iin_col) - (jj);
                                 LR[d] = l + (iin_col) + (jj);
                                 wUL[d] = weight;
                                 wUR[d] = weight;
                                 wLL[d] = weight;
                                 wLR[d] = weight;
                                 d++;
                              }
                           }
                           upindex = 0;
                           dnindex = 0;
                           lfindex = 0;
                           rtindex = 0;
                           for (ii=1;ii<tier+1;ii++) {
			       iin_col = ii*n_col;
			       weight = 1.0F;
                               if (crs)
				   weight /= (float)(ii+1);
				   
                               if ( (i-ii) > -1 )    {
			           UP[upindex] = l - (iin_col); 
			           wUP[upindex++] = weight; 
				   }
                               if ( (i+ii) < n_row ) {
			           DN[dnindex] = l + (iin_col); 
			           wDN[dnindex++] = weight; 
				   }
                               if ( (j-ii) > -1 )    {
			           LF[lfindex] = l - ii;
			           wLF[lfindex++] = weight;
				   } 
                               if ( (j+ii) < n_col ) {
			           RT[rtindex] = l + ii;
			           wRT[rtindex++] = weight;
				   } 
                           }                          
                           
                           /* sum up all non-interpolated data until all corners have something */
                           /* first, add the sides */
                           for (ii=0;ii<upindex;ii++) {
			       qaval = qa[UP[ii]];
			       if ((qaval == 0)||(qaval == 6)) {
                                   f32 += (float32)data[UP[ii]]*wUP[ii];
				   side_got[0]++;
                                   n+=wUP[ii];
                                    }
				    }
                           for (ii=0;ii<dnindex;ii++) {
			       qaval = qa[DN[ii]];
			       if ((qaval == 0)||(qaval == 6)) {
                                   f32 += (float32)data[DN[ii]]*wDN[ii];
				   side_got[1]++;
                                   n+=wDN[ii];
                                    }
				    }
                           for (ii=0;ii<lfindex;ii++) {
			       qaval = qa[LF[ii]];
			       if ((qaval == 0)||(qaval == 6)) {
				   side_got[2]++;
                                   f32 += (float32)data[LF[ii]]*wLF[ii];
                                   n+=wLF[ii];
                                    }
				    }
                           for (ii=0;ii<rtindex;ii++) {
			       qaval = qa[RT[ii]];
 			       if ((qaval == 0)||(qaval == 6)) {
                                   f32 += (float32)data[RT[ii]]*wRT[ii];
				   side_got[3]++;
                                   n+=wRT[ii];
                                    }
				    }

                           for (ii=0;ii<tier*tier;ii++) {   
                               if (UL[ii] != -1) {
			          qaval = qa[UL[ii]];
			          if ((qaval == 0)||(qaval == 6)) {
                                      corner_got[0]++;
                                      f32 += (float32)data[UL[ii]]*wUL[ii];
                                      n+=wUL[ii];
                                       }
				    }
                               if (UR[ii] != -1) {
			          qaval = qa[UR[ii]];
			          if ((qaval == 0)||(qaval == 6)) {
                                      corner_got[1]++;
                                      f32 += (float32)data[UR[ii]]*wUR[ii];
                                      n+=wUR[ii];
                                       }
				    }
                               if (LL[ii] != -1) {
			          qaval = qa[LL[ii]];
			          if ((qaval == 0)||(qaval == 6)) {
                                      corner_got[2]++;
                                      f32 += (float32)data[LL[ii]]*wLL[ii];
                                      n+=wLL[ii];
                                       }
				    }
                               if (LR[ii] != -1) {
			          qaval = qa[LR[ii]];
			          if ((qaval == 0)||(qaval == 6)) {
                                      corner_got[3]++;
                                      f32 += (float32)data[LR[ii]]*wLR[ii];
                                      n+=wLR[ii];
                                       }
				    }
                           }
                           
/*if ((i==1002)&&(j==996))
   printf("tier %d: UL corner %d, UR corner %d, LL corner %d, LR corner %d\n", 
      tier,corner_got[0], corner_got[1], corner_got[2], corner_got[3] );
   */   
                           /* figure out if each corner has been sampled */
			   if (crs) {
                             corn = 0;
			     side = 0;
                           }
			   else {
                             corn = 1;
			     side = 1;
			   }
			   
                           for (ii=0;ii<4;ii++) {
			      if (crs) {
			         corn += (corner_got[ii]!=0);
			         side += (side_got[ii]!=0);
			      }
			      else {
			         corn *= corner_got[ii];
			         side *= side_got[ii];
			      }
                           }
			   if (crs) {
                              if ( ((corn > 1)||(side > 1)) &&
			            (n > 0.50F)) {   
                                 data[l] = (uint16)(f32/n);
			         /*qa[l] = 6;*/
			         failed = 0;
                                 break;  /* to get us out of the "for (tier=1...) loop */
                              }
                           }
			   else {
                              if ((corn != 0)||(side != 0)) {   /* and if so... */
                                 data[l] = (uint16)(f32/n);
			         /*qa[l] = 6;*/
			         failed = 0;
                                 break;  /* to get us out of the "for (tier=1...) loop */
                              }
                           }
                        } /* for (tier=1;tier<TIERMAX;tier++) */
			
			
                     }  /* if (qaval == ...) */
		     
		     fail[l] = (uint8)failed; 
		     
                  }  /* for (j=0;j<n_col;j++) */
              }  /* for (i=0;i<n_row;i++) */


/***************/
}


void geo_sat(float *lat_buffer, float *lon_buffer, int total_lines, int in_line, int in_pix,
             float outlat, float outlon, int *out_line, int *out_pix, int16 *its)
/*
!C*****************************************************************************
!Description:
	Finds line and pixel (out_line, out_pixel) of latitude and longitude in lat_buffer,
	   lon_buffer that corresponds to outlat, outlon.  Searches starting at in_line, in_pix,
	   assuming in_line/in_pix are close to the desired lat/lon.

!Input Parameters:
        lat_buffer
	lon_buffer
	total_lines
	in_line
	in_pix
	outlat
	outlon

!Output Parameters:
        out_line
	out_pixel
	
!Returns:
        nothing

!Revision History:

	Original version March, 2002, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:

!References and Credits

!Design Notes

	
!END***************************************************************************
*/
{
int index, rowincr, colincr, iterations;
int row, col;
float lat0, lon0;
int finished, ret;
char location[30]="aerosols.c, geo_sat\0";
char msg[256];
int not_finished( int total_lines, float *lat_buffer, float *lon_buffer, int row, int col, 
                  float latval, float lonval, int *rowincr, int *colincr);

/* sanity check */
index = in_line*NBSPOTS_1KM + in_pix;
lat0 = lat_buffer[index];
lon0 = lon_buffer[index];

if ((lat0 > LATLIMIT)||(lat0 < -LATLIMIT)||(lon0 > 180.1F)||(lon0 < -180.1F)) {
   *out_line = -1;
   *out_pix = -1;
   return;
  }
  
row = in_line;
col = in_pix;
rowincr = 0;
colincr = 0;
finished = 0;
iterations = 0;
while ( !finished ) {

   iterations++;
   ret = not_finished( total_lines, lat_buffer, lon_buffer, row, col, 
                      outlat, outlon, &rowincr, &colincr); 
   row += rowincr;
   col += colincr;
   rowincr = 0;
   colincr = 0;
   if (!ret) finished = 1;
   if (iterations > 100) {
       sprintf(msg, "Function aborted at row=%d, col=%d", row, col);
       modsmf((PGSt_SMF_code)MODIS_E_FUNCTION_ERROR, msg, location);
       finished = 1;   /* to prevent hang-ups */
       }

  } 
  
*out_line = row;
*out_pix = col;
*its = iterations;
return;

}

int not_finished( int total_lines, float *lat_buffer, float *lon_buffer, int row, int col, 
                  float latval, float lonval, int *rowincr, int *colincr)
/*
!C*****************************************************************************
!Description:
	Checks latval/lonval against lat/lon of the pixels around row/col; 
	determines shortest distance; returns in rowincr/colincr the values to add 
	to row/col to get to the shortest distance (return nonzero) --
	or not (return 0, you are there).
	
!Input Parameters:
        total_lines      the total number of lines in the swath
	lat_buffer       whole-swath latitude buffer
	lon_buffer       whole-swath longitude buffer
	row              which row to start looking at in swath
	col              which column to start looking at in swath
	latval           which latitude to look for in swath
	lonval           which longitude to look for in swath

!Output Parameters:
        rowincr          value to add to row in order to get to latval
	colincr          value to add to col in order to get to lonval
	
!Returns:
        returns 0 if finished, 1 if NOT finished 

!Revision History:

	Original version March, 2002, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:

!References and Credits

!Design Notes

	
!END***************************************************************************
*/
{
/* This step is to save some multiplications and divisions */
#define DEGTORAD     0.017453293      /* PI/180.0 */
#define DEGTORADERAD 111.3194908      /* PI/180.0*6378.137 */
int index, iindex, indx, incr;
int i,j;
float lat, lon, lat0, lon0, target_distance, dist, mindist;

index = row*NBSPOTS_1KM + col;
lat0 = lat_buffer[index];
if ((lat0 > LATLIMIT)||(lat0 < -LATLIMIT)) {  /* forget it */
    *rowincr = 0;
    *colincr = 0;
    return(0);
   }  
lon0 = lon_buffer[index];
mindist = 100000.0F;

lat = (latval-lat0)*DEGTORADERAD;
lon = (lonval-lon0)*DEGTORAD*(cosf((latval+lat0)*DEGTORAD/2.0F))*ERAD;

target_distance = lat*lat + lon*lon;

if ((col < 70)||(col > (NBSPOTS_1KM-70))) incr = 5;
else if ((col < 100)||(col > (NBSPOTS_1KM-100))) incr = 4;
else if ((col < 200)||(col > (NBSPOTS_1KM-200))) incr = 3;
else if ((col < 400)||(col > (NBSPOTS_1KM-400))) incr = 2;
else incr = 1;

/*printf("%f %f TD = %f\n", lat, lon,target_distance);*/

for(i=(row-incr);i<=(row+incr);i++) {
   if ((i < 0)||(i >= total_lines)) continue;
   indx = i*NBSPOTS_1KM;
   for(j=(col-incr);j<=(col+incr);j++) {
       if ((j < 0)||(j >= NBSPOTS_1KM)) continue;
       iindex = indx + j;
       if (index == iindex) continue;
       
       
       if ((lat_buffer[iindex] > LATLIMIT)||(lat_buffer[iindex] < -LATLIMIT)) continue; 
     

       lat = (latval-lat_buffer[iindex])*DEGTORADERAD;
       lon = (lonval-lon_buffer[iindex])*DEGTORAD*(cosf((latval+lat_buffer[iindex])*DEGTORAD/2.0F))*ERAD;
         
       dist = lat*lat + lon*lon;
/*printf("%f %f D = %f\n",lat_buffer[iindex] ,lon_buffer[iindex],dist);*/
       if (dist < mindist) {
           mindist = dist;
	   *rowincr = -(row - i);
	   *colincr = -(col - j);
          }
   
       }  /* for(j=col-1;j<=col+1;j++)) */
   }  /* for(i=row-1;i<=row+1;i++) */

if (target_distance < mindist) {  /* you are there */
    *rowincr = 0;
    *colincr = 0;
    return(0);
   }  
else {
    return(1);
}

}


int float32_fread_wrapper(FILE *fd, float32 *pntr, int length, char *calling_routine) 
/*
!C*****************************************************************************
!Description:
	Does an fread on the file pointed to by fd, assuming it contains floating-point data.
	Puts the read data into pntr.  Swaps the data if box is LINUX.
	
!Input Parameters:

!Output Parameters:
	
!Returns:
     0   success
    -1   fread failure
    
!Revision History:

	Original version March, 2002, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:

!References and Credits

!Design Notes

	
!END***************************************************************************
*/
{
#ifdef LINUX
unsigned int iskp, icnt, iptr, itmp;
#endif
char location[256];

sprintf(location, "{called from %s} aerosols.c, float32_fread_wrapper\0", calling_routine);

if (fread(pntr, length*sizeof(float32), 1, fd) != 1) {
    modsmf(MODIS_E_FILE_READ_ERROR, "  ", location);
    return -1;
    }

#ifdef LINUX
       /* Swap every 4 bytes in refl so that the resulting byte order is 4, 3, 2, 1. */
       iskp = sizeof(float32);
       for ( icnt = 0 ; icnt < length*iskp ; icnt = icnt + iskp ) { 
           iptr = (unsigned int) pntr + icnt;
           itmp = bswap_32 ( * (int32*) iptr );
           * (int32*) iptr = itmp;
       }
#endif
return 0;
}

int int32_fread_wrapper(FILE *fd, int32 *pntr, int length, char *calling_routine) 
/*
!C*****************************************************************************
!Description:
	Does an fread on the file pointed to by fd, assuming it contains int32 data.
	Puts the read data into pntr.  Swaps the data if box is LINUX.
	
!Input Parameters:

!Output Parameters:
	
!Returns:
     0   success
    -1   fread failure
    
!Revision History:

	Original version March, 2002, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:

!References and Credits

!Design Notes

	
!END***************************************************************************
*/
{
#ifdef LINUX
unsigned int iskp, icnt, iptr, itmp;
#endif
char location[256];

sprintf(location, "{called from %s} aerosols.c, int32_fread_wrapper\0", calling_routine);

if (fread(pntr, length*sizeof(int32), 1, fd) != 1) {
    modsmf(MODIS_E_FILE_READ_ERROR, "   ", location);
    return -1;
    }

#ifdef LINUX
       /* Swap every 4 bytes in refl so that the resulting byte order is 4, 3, 2, 1. */
       iskp = sizeof(int32);
       for ( icnt = 0 ; icnt < length*iskp ; icnt = icnt + iskp ) { 
           iptr = (unsigned int) pntr + icnt;
           itmp = bswap_32 ( * (int32*) iptr );
           * (int32*) iptr = itmp;
       }
#endif
return 0;
}

int short_fread_wrapper(FILE *fd, short *pntr, int length, int n_lines, char *calling_routine) 
/*
!C*****************************************************************************
!Description:
	Does an fread on the file pointed to by fd, assuming it contains short data.
	Puts the read data into pntr.  Swaps the data if box is LINUX.
	
!Input Parameters:

!Output Parameters:
	
!Returns:
     0   success
    -1   fread failure
    
!Revision History:

	Original version March, 2002, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:

!References and Credits

!Design Notes

	
!END***************************************************************************
*/
{
#ifdef LINUX
unsigned int iskp, icnt, iptr, itmp;
#endif
char location[256];

sprintf(location, "{called from %s} aerosols.c, short_fread_wrapper\0", calling_routine);

if (fread(pntr, length*sizeof(short), n_lines, fd) != n_lines) {
    modsmf(MODIS_E_FILE_READ_ERROR, "   ", location);
    return -1;
    }

#ifdef LINUX
       /* Swap every 2 bytes in pntr so that the resulting byte order is 2, 1. */
       iskp = sizeof(short);
       for ( icnt = 0 ; icnt < length*n_lines*iskp ; icnt = icnt + iskp ) { 
           iptr = (unsigned int) pntr + icnt;
           itmp = bswap_16 ( * (int16*) iptr );
           * (int16*) iptr = itmp;
       }
#endif
return 0;
}

int init_aotr_rejection(int lun)
{
/*
!C*****************************************************************************
!Description:
	Initializes AOT retrieval rejection mask; reads a file pointed to by LUN
	    which contains compressed mask.

!Input Parameters:
        LUN

!Output Parameters:
        None (global variables are updated)
    	
!Returns:
   1   MODIS_E_FILE_ACCESS   file access errors
   2   MODIS_E_HDF_ERROR     SDS access errors
   3   MODIS_E_MEM_FAILURE   memory errors
   4   MODIS_E_GET_METADATA  metadata error
   5   MODIS_E_GETREF_FAIL   PGS_PC error
   0   OK 

!Revision History:

	Original version March, 2003, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:
!References and Credits
!Design Notes

     Writes data to global variable aotr_struct, a aerosol_retr_rej_t-type structure
     that holds aerosol retrieval rejection data.

!END***************************************************************************
*/
int32 sdid, sdss, rank, number_type, n_gattr, n_attr, n_sets, dims[5], start[1], rez_attr;
char name[256], filename[256], msg[256];
char location[256] = "aerosols.c, init_aotr_rejection\0";
int ver = 1;

/* initialize the global aerosol_retr_rej_t structure */
aotr_struct.aotr_offsets = (uint32 *)NULL;
aotr_struct.aotr_lengths = (uint32 *)NULL;
aotr_struct.n_aotr_offsets = 0;
aotr_struct.aotr_resolution = 0.0F;

/* Check file */
if(PGS_PC_GetReference(lun, &ver, filename)!=PGS_S_SUCCESS){
   sprintf(msg, "LUN %d, version 1 (AOT retrieval rejection mask)", (int32)lun);
   modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, msg, location);
   return(5);
    }
if ( (sdid=SDstart(filename, DFACC_RDONLY)) == -1 ) {
   sprintf(msg, "cannot run SDstart on %s, check the file", filename);
   modsmf(MODIS_E_FILE_ACCESS, msg, location);
   return(1);
   }
if ( SDfileinfo(sdid, &n_sets, &n_gattr) == -1 ) {
   sprintf(msg, "cannot run SDfileinfo on %s, check the file", filename);
   modsmf(MODIS_E_FILE_ACCESS, msg, location);
   return(1);
   }
if (( n_sets != 2 )||( n_gattr != 2 )) {
   sprintf(msg, "something wrong with file %s", filename);
   modsmf(MODIS_E_FILE_ACCESS, msg, location);
   return(1);
   }
   
/* Check for "offsets" SDS */
if ( (sdss=SDselect(sdid, 0)) == -1 ) {
   sprintf(msg, "cannot run SDselect on first SDS of %s, check the file", filename);
   modsmf(MODIS_E_HDF_ERROR, msg, location);
   return(2);
   }
if ( SDgetinfo(sdss, name, &rank, dims, &number_type, &n_attr) == -1 ) {
   sprintf(msg, "cannot run SDgetinfo on %s, check the file", filename);
   modsmf(MODIS_E_HDF_ERROR, msg, location);
   return(2);
   }
if (( number_type != DFNT_UINT32 )||( rank != 1 )||(strcmp(name, "offsets\0"))||(dims[0] <= 0)) {
   sprintf(msg, "something wrong with file %s", filename);
   modsmf(MODIS_E_HDF_ERROR, msg, location);
   return(2);
   }
   
/* get memory */
aotr_struct.aotr_offsets = (uint32 *)malloc(dims[0]*sizeof(uint32));
if (aotr_struct.aotr_offsets == NULL) {
    sprintf(msg, "cannot allocate %d uint32's", dims[0]);
    modsmf(MODIS_E_MEM_FAILURE, msg, location);
    return(3);
    }
aotr_struct.aotr_lengths = (uint32 *)malloc(dims[0]*sizeof(uint32));
if (aotr_struct.aotr_lengths == NULL) {
    sprintf(msg, "cannot allocate %d uint32's", dims[0]);
    modsmf(MODIS_E_MEM_FAILURE, msg, location);
    return(3);
    }
    
/* read offsets */
start[0] = 0;
if ( SDreaddata(sdss, start, NULL, dims, aotr_struct.aotr_offsets) == -1 ) {
   sprintf(msg, "cannot run SDreaddata on %s, check the file", filename);
   modsmf(MODIS_E_FILE_ACCESS, msg, location);
   return(2);
   }
SDendaccess(sdss);
aotr_struct.n_aotr_offsets = dims[0];

/* Check for "lengths" SDS */
if ( (sdss=SDselect(sdid, 1)) == -1 ) {
   sprintf(msg, "cannot run SDselect on second SDS of %s, check the file", filename);
   modsmf(MODIS_E_FILE_ACCESS, msg, location);
   return(2);
   }
if ( SDgetinfo(sdss, name, &rank, dims, &number_type, &n_attr) == -1 ) {
   sprintf(msg, "cannot run SDgetinfo on %s, check the file", filename);
   modsmf(MODIS_E_HDF_ERROR, msg, location);
   return(2);
   }
if (( number_type != DFNT_UINT32 )||( rank != 1 )||(strcmp(name, "lengths\0"))||(dims[0] <= 0)) {
   sprintf(msg, "something wrong with file %s", filename);
   modsmf(MODIS_E_HDF_ERROR, msg, location);
   return(2);
   }

/* read lengths */
start[0] = 0;
if ( SDreaddata(sdss, start, NULL, dims, aotr_struct.aotr_lengths) == -1 ) {
   sprintf(msg, "cannot run SDreaddata on %s, check the file", filename);
   modsmf(MODIS_E_FILE_ACCESS, msg, location);
   return(2);
   }
SDendaccess(sdss);

/* get resolution  */
if ( (rez_attr = SDfindattr(sdid, "PIXPERDEG\0")) == -1 ){
   sprintf(msg, "cannot find resolution attribute in %s, check the file", filename);
   modsmf(MODIS_E_GET_METADATA, msg, location);
   return(4);
   }
if ( SDreadattr(sdid, rez_attr, &(aotr_struct.aotr_resolution)) == -1 ) {
   sprintf(msg, "cannot run SDreadattr on %s, check the file", filename);
   modsmf(MODIS_E_GET_METADATA, msg, location);
   return(4);
   }
SDend(sdid);
return(0);

}

int aotr_rejection(float lat, float lon)
/*
!C*****************************************************************************
!Description:
	Compares lat/lon versus data in the aotr_offsets and aotr_lengths, 
	    returning 1 to REJECT that lat/lon combination, 0 to ACCEPT that lat/lon.

!Input Parameters:
        lat
	lon

!Output Parameters:
	none
	
!Returns:
   1   reject
   0   accept

!Revision History:

	Original version March, 2003, Jim Ray, jim@kratmos.gsfc.nasa.gov

!Team-unique Header:
!References and Credits
!Design Notes

    Uses data from global variable aotr_struct, a aerosol_retr_rej_t-type structure
     that holds aerosol retrieval rejection data.
 	
!END***************************************************************************
*/
{
/* returns 1 to reject, 0 to accept */
int i; 
int row, col, hit;
long target;

/* Put lat, lon into global offsets from +90 lat, -180 longitude, at the appropriate resolution */
row = (int)((90.00 - lat)*aotr_struct.aotr_resolution);
col = (int)((lon - (-180.00))*aotr_struct.aotr_resolution);

/* get target offset */
target = (long)(360.0F*aotr_struct.aotr_resolution);
target *= (long)row;
target += (long)col;

/* check each offset/length combination for the target */
hit = 0;
for(i=0;i<aotr_struct.n_aotr_offsets;i++){
    if ((aotr_struct.aotr_offsets[i] <= target)&&(target <= (aotr_struct.aotr_offsets[i]+aotr_struct.aotr_lengths[i]))) hit=1;
    }
return(hit);
}


PGSt_SMF_status rmpath(char *buffer)
/*
!C****************************************************************************

Routine: rmpath()

!Description: This module removes any local path information from a 
             path/filename string.

!Input Parameters:
  char *buffer  Character buffer that contains the path/filename string

!Output Parameters:
  char *buffer  Character buffer that contains the filename string

Returns:
  PGSt_SMF_status
    on success:    MODIS_S_SUCCESS
    on error:      MODIS_E_NULL_STRING

!Revision History:
  Revision 1.0  1996/03/15
  Paul Fisher/SDST
  Original Development

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Paul S. Fisher
      MODIS Science Data Support Team   
      Research and Data Systems Corporation
      SAIC/GSC MODIS Support Office
      7501 Forbes Blvd
      Seabrook, MD 20706  
      fisher@ltpmail.gsfc.nasa.gov     

Design Notes:
  Error handling is performed through the modules smf.c and the
  toolkit SMF seed file MODIS_39501.t.

  Upon error, this function will return to the calling module to
  allow exection to continue.

  The input string should not be used for file Epro_access through the SDP
  toolkit after calls to this module, as the proper path information
  contained within will have been removed.
  
!END**************************************************************************/
{

  int32 length, start=-1;  /* initialize start to before beginning of string, 
			    length is end of string */
  register int32 i;       /* counter */

  /* check to make sure buffer is not null or empty*/
  if(!buffer||buffer[0]=='\0'){
    modsmf(MODIS_E_NULL_STRING, "buffer", "aerosols.c, rmpath");
    return (MODIS_E_NULL_STRING);
  }

  /* get string length */
  length=(int32)strlen(buffer);

  /* calculate beginning of file name */
  for(i=(length-1);i>-1;i--)if (buffer[i]=='/'){
    start=i;
    break;
  }

  /* shift filename and null terminator to beginning of buffer 
     use (length +1) to always include the NULL terminator */
  if(start!=-1)for(i=(start+1);i<(length+1);i++)buffer[i-(start+1)]=buffer[i];

  return(MODIS_S_SUCCESS);

}
