/*
!C******************************************************************

!Description: 
		ancillary.c

	This header is used to satisfy some bizarre ECS/MODIS requirement.
	The reader is instructed to examine the function headers for any
	real content.

	Note that in order to make this header "non-empty", we are including
	a copy of the US ASCII table (octal, hexadecimal, and decimal formats).
	This table is very useful.


!Input Parameters:
!Output Parameters: 
!Revision History:
!Team-unique Header:
!References and Credits 
!Design Notes


DESCRIPTION
     The ASCII character set defines a 1-to-1 mapping of characters to 8-bit
     values:

     octal:
     |000 nul |001 soh |002 stx	|003 etx |004 eot |005 enq |006	ack |007 bel |
     |010 bs  |011 ht  |012 nl	|013 vt	 |014 np  |015 cr  |016	so  |017 si  |
     |020 dle |021 dc1 |022 dc2	|023 dc3 |024 dc4 |025 nak |026	syn |027 etb |
     |030 can |031 em  |032 sub	|033 esc |034 fs  |035 gs  |036	rs  |037 us  |
     |040 sp  |041 !   |042 "	|043 #	 |044 $	  |045 %   |046	&   |047 '   |
     |050 (   |051 )   |052 *	|053 +	 |054 ,	  |055 -   |056	.   |057 /   |
     |060 0   |061 1   |062 2	|063 3	 |064 4	  |065 5   |066	6   |067 7   |
     |070 8   |071 9   |072 :	|073 ;	 |074 <	  |075 =   |076	>   |077 ?   |
     |100 @   |101 A   |102 B	|103 C	 |104 D	  |105 E   |106	F   |107 G   |
     |110 H   |111 I   |112 J	|113 K	 |114 L	  |115 M   |116	N   |117 O   |
     |120 P   |121 Q   |122 R	|123 S	 |124 T	  |125 U   |126	V   |127 W   |
     |130 X   |131 Y   |132 Z	|133 [	 |134 \	  |135 ]   |136	^   |137 _   |
     |140 `   |141 a   |142 b	|143 c	 |144 d	  |145 e   |146	f   |147 g   |
     |150 h   |151 i   |152 j	|153 k	 |154 l	  |155 m   |156	n   |157 o   |
     |160 p   |161 q   |162 r	|163 s	 |164 t	  |165 u   |166	v   |167 w   |
     |170 x   |171 y   |172 z	|173 {	 |174 |	  |175 }   |176	~   |177 del |

     hexadecimal:
     | 00 nul |	01 soh | 02 stx	| 03 etx | 04 eot | 05 enq | 06	ack | 07 bel |
     | 08 bs  |	09 ht  | 0a nl	| 0b vt	 | 0c np  | 0d cr  | 0e	so  | 0f si  |
     | 10 dle |	11 dc1 | 12 dc2	| 13 dc3 | 14 dc4 | 15 nak | 16	syn | 17 etb |
     | 18 can |	19 em  | 1a sub	| 1b esc | 1c fs  | 1d gs  | 1e	rs  | 1f us  |
     | 20 sp  |	21 !   | 22 "	| 23 #	 | 24 $	  | 25 %   | 26	&   | 27 '   |
     | 28 (   |	29 )   | 2a *	| 2b +	 | 2c ,	  | 2d -   | 2e	.   | 2f /   |
     | 30 0   |	31 1   | 32 2	| 33 3	 | 34 4	  | 35 5   | 36	6   | 37 7   |
     | 38 8   |	39 9   | 3a :	| 3b ;	 | 3c <	  | 3d =   | 3e	>   | 3f ?   |
     | 40 @   |	41 A   | 42 B	| 43 C	 | 44 D	  | 45 E   | 46	F   | 47 G   |
     | 48 H   |	49 I   | 4a J	| 4b K	 | 4c L	  | 4d M   | 4e	N   | 4f O   |
     | 50 P   |	51 Q   | 52 R	| 53 S	 | 54 T	  | 55 U   | 56	V   | 57 W   |
     | 58 X   |	59 Y   | 5a Z	| 5b [	 | 5c \	  | 5d ]   | 5e	^   | 5f _   |
     | 60 `   |	61 a   | 62 b	| 63 c	 | 64 d	  | 65 e   | 66	f   | 67 g   |
     | 68 h   |	69 i   | 6a j	| 6b k	 | 6c l	  | 6d m   | 6e	n   | 6f o   |
     | 70 p   |	71 q   | 72 r	| 73 s	 | 74 t	  | 75 u   | 76	v   | 77 w   |
     | 78 x   |	79 y   | 7a z	| 7b {	 | 7c |	  | 7d }   | 7e	~   | 7f del |

     decimal:
     |	0 nul |	 1 soh |  2 stx	|  3 etx |  4 eot |  5 enq |  6	ack |  7 bel |
     |	8 bs  |	 9 ht  | 10 nl	| 11 vt	 | 12 np  | 13 cr  | 14	so  | 15 si  |
     | 16 dle |	17 dc1 | 18 dc2	| 19 dc3 | 20 dc4 | 21 nak | 22	syn | 23 etb |
     | 24 can |	25 em  | 26 sub	| 27 esc | 28 fs  | 29 gs  | 30	rs  | 31 us  |
     | 32 sp  |	33 !   | 34 "	| 35 #	 | 36 $	  | 37 %   | 38	&   | 39 '   |
     | 40 (   |	41 )   | 42 *	| 43 +	 | 44 ,	  | 45 -   | 46	.   | 47 /   |
     | 48 0   |	49 1   | 50 2	| 51 3	 | 52 4	  | 53 5   | 54	6   | 55 7   |
     | 56 8   |	57 9   | 58 :	| 59 ;	 | 60 <	  | 61 =   | 62	>   | 63 ?   |
     | 64 @   |	65 A   | 66 B	| 67 C	 | 68 D	  | 69 E   | 70	F   | 71 G   |
     | 72 H   |	73 I   | 74 J	| 75 K	 | 76 L	  | 77 M   | 78	N   | 79 O   |
     | 80 P   |	81 Q   | 82 R	| 83 S	 | 84 T	  | 85 U   | 86	V   | 87 W   |
     | 88 X   |	89 Y   | 90 Z	| 91 [	 | 92 \	  | 93 ]   | 94	^   | 95 _   |
     | 96 `   |	97 a   | 98 b	| 99 c	 |100 d	  |101 e   |102	f   |103 g   |
     |104 h   |105 i   |106 j	|107 k	 |108 l	  |109 m   |110	n   |111 o   |
     |112 p   |113 q   |114 r	|115 s	 |116 t	  |117 u   |118	v   |119 w   |
     |120 x   |121 y   |122 z	|123 {	 |124 |	  |125 }   |126	~   |127 del |



!END*************************************************************************/

/* NODAAC may need to be defined when running this code
in any non-DAAC processing mode.  This feature was added
by MODIS SDST to ensure compliance with ESDIS standards upon
delivering this software to the DAAC. */

#ifdef NODAAC

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* PGS seed files */

#include "PGS_MODIS_37092.h"

/* PGS I/O & SMF module */
#include "smfio.h"

#include "proc_protos.h"
#include "proc_struct.h"
#include "proc_define.h"
#include "hdf.h"               /* HDF header files */
#include "mfhdf.h"
#include "mdefs.h"

#define UNDEFINED 9.999e20     /* Has to be same as in grib.h */

int32 get_cart_linesamp(float32 lat, float32 lon, float32 latres, int32 samps, 
		       float32 lonres, int32 *iline,
		       int32 *isamp, float32 *rowfactor, float32 *colfactor)
/*
!C*****************************************************************************
!Description:  Given a latitute and longitude, this function determines the 
               line, sample, subline factor and subsample factor of the point 
	       in the DAO geographic grid.

!Input Parameters:
lat: latitude of point in question
lon: longitude of point in question
latres: latitude resolution of the input data
lonres: longitude resolution of the input data
samps: columns in the input data file

!Output Parameters:
iline: line number containing the point in question
isamp: sample number containing the point in question
rowfactor: vertical distance of the point in question from the 
           center of iline to the center of the next southern line.
colfactor: horizontal distance of the point in question from the 
           center of isamp to the center of the next eastern pixel.

!Input/Output Parameters:

!Returns:
	0	failure (lat and/or lon is out of range)
	1	success

!Revision History:
        Revision 1.1, August 1997, P. Fisher
	1.  Added absolute value call into rowfactor/colfactor calculation
            (bug fix).

	Original version 1997, P. Fisher (fisher@ltpmail.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Paul Fisher
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		fisher@ltpmail.gsfc.nasa.gov, (301) 286-2980

!Design Notes
!END***************************************************************************
*/
{

  float32 rowsub, colsub;
  float64 tmp;

  /* check inputs */
  if(lat>90.0F||lat<-90.0F||lon>180.0F||lon<-180.0F)return(0);

  *iline = (int32)(((90.0F - lat)/latres)+0.5F);
  *rowfactor = (float32)((90.0F - ((float32)(*iline)*latres)) - lat)/latres;
  if((*rowfactor)<0.0F){
    (*iline)--;
    *rowfactor=1.0F-fabsf((float)(*rowfactor));
  }

  /* special wrap around case */
  if(lon>(180.0F-(lonres*0.5F)))lon=-180.0F-(180.0F-lon);

  *isamp = (int32)(((lon +180.0F)/lonres) + 0.5F);
  tmp = (float64)(lon +(180.0F - ((*isamp)*lonres)))/lonres; /*29-OCT-01*/
  *colfactor = (float32)tmp;
  if((*colfactor)<0.0F){
    (*isamp)--;
    /* special wrap around case */
    if((*isamp)<0)*isamp=(samps-1);

    *colfactor=1.0F-fabsf((float)(*colfactor));
    }
  
  return(1);
}


float32 get_cart_interp_NONdao(int32 iline, int32 isamp, int32 samps, float32 rowfactor, 
		     float32 colfactor, t_ancillary *anc)
/*
!C*****************************************************************************
!Description:  bilinearly interpolates a point in question through its 
               isamp, iline, row and column subsample distances 
	       (supplied through a call to get_cart_linesamp()) for data 
	       in the cartographic grid.  In the current version of the 
	       MOD09 algorithm, this routine is only used to interpolate
	       the ozone climatological table.

!Input Parameters:
iline: line number containing the point in question
isamp: sample number containing the point in question
samps: number of columns in the input DAO data file
rowfactor: vertical distance of the point in question from the 
           center of iline to the center of the next southern line.
colfactor: horizontal distance of the point in question from the 
           center of isamp to the center of the next eastern pixel.
anc: ancillary data structure

!Output Parameters: none

!Input/Output Parameters: none

!Returns: interpolated data value

!Revision History:
	Original version 1997, P. Fisher (fisher@ltpmail.gsfc.nasa.gov).


!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Paul Fisher
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		fisher@ltpmail.gsfc.nasa.gov, (301) 286-2980

!Design Notes
!END****************************************************************************/
{
  float32 v1, v2, v3, v4;
  int32 i1, i2, i3, i4;
  float32 interp;

  /* obtain four corners */
  v1= 0.0;
  v2= 0.0;
  v3= 0.0;
  v4= 0.0;
  i1= iline * samps + isamp;
  i2= (iline + 1) * samps + isamp;
  i3= (iline + 1) * samps + isamp+1;
  i4= iline * samps + isamp+1;
  if (i1 >= 0) v1= anc->data[i1];
  if (i2 >= 0) v2= anc->data[i2];
  if (i3 >= 0) v3= anc->data[i3];
  if (i4 >= 0) v4= anc->data[i4];

/*  v1 = anc->data[iline * samps + isamp];
  v2 = anc->data[(iline + 1) * samps + isamp];
  v3 = anc->data[(iline + 1) * samps + isamp+1];
  v4 = anc->data[iline * samps + isamp+1];
 */ 
  interp=(1.0F-rowfactor)*(1.0F-colfactor)*v1 +  
    rowfactor*(1.0F-colfactor)*v2 +
    rowfactor*colfactor*v3 + 
    colfactor*(1.0F-rowfactor)*v4;

  return(interp);

}

float32 get_cart_interp_NONdao_spres(int32 iline, int32 isamp, int32 samps, int32 iline0, int32 icol0, 
				     float32 rowfactor, 
				     float32 colfactor, t_ancillary *anc)
/*
!C*****************************************************************************
!Description:  bilinearly interpolates a point in question through its 
               isamp, iline, row and column subsample distances 
	       (supplied through a call to get_cart_linesamp()) for data 
	       in the cartographic grid.  In the current version of the 
	       MOD09 algorithm, this routine is only used to interpolate
	       the ozone climatological table.

!Input Parameters:
iline: line number containing the point in question
isamp: sample number containing the point in question
samps: number of columns in the input DAO data file
rowfactor: vertical distance of the point in question from the 
           center of iline to the center of the next southern line.
colfactor: horizontal distance of the point in question from the 
           center of isamp to the center of the next eastern pixel.
anc: ancillary data structure

!Output Parameters: none

!Input/Output Parameters: none

!Returns: interpolated data value

!Revision History:
	Original version 1997, P. Fisher (fisher@ltpmail.gsfc.nasa.gov).


!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Paul Fisher
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		fisher@ltpmail.gsfc.nasa.gov, (301) 286-2980

!Design Notes
!END****************************************************************************/
{
  float32 v1, v2, v3, v4;
  int32 i1, i2, i3, i4;
  float32 interp;
  /* obtain four corners */

  v1= 0.0;
  v2= 0.0;
  v3= 0.0;
  v4= 0.0;
  i1= (iline-iline0) * samps + (isamp-icol0);
  i2= ((iline-iline0) + 1) * samps + (isamp-icol0);
  i3= ((iline-iline0) + 1) * samps + (isamp-icol0)+1;
  i4= (iline-iline0) * samps + (isamp-icol0)+1;

  if (isamp+1 >= samps) {  /* 29-OCT-01: wrap-around */
      i3= ((iline-iline0) + 1) * samps;
      i4= (iline-iline0) * samps;
  }

  if (i1 >= 0) v1= anc->data[i1];
  if (i2 >= 0) v2= anc->data[i2];
  if (i3 >= 0) v3= anc->data[i3];
  if (i4 >= 0) v4= anc->data[i4];

/*  v1 = anc->data[(iline-iline0) * samps + (isamp-icol0)];
  v2 = anc->data[((iline-iline0) + 1) * samps + (isamp-icol0)];
  v3 = anc->data[((iline-iline0) + 1) * samps + (isamp-icol0)+1];
  v4 = anc->data[(iline-iline0) * samps + (isamp-icol0)+1];
  */

  interp=(1.0F-rowfactor)*(1.0F-colfactor)*v1 +  
    rowfactor*(1.0F-colfactor)*v2 +
    rowfactor*colfactor*v3 + 
    colfactor*(1.0F-rowfactor)*v4;

  return(interp);

}


float32 get_cart_interp(int32 iline, int32 isamp, int32 samps, float32 rowfactor, 
		     float32 colfactor, d_ancillary *anc)
/*
!C*****************************************************************************
!Description:  bilinearly interpolates a point in question through its 
               isamp, iline, row and collumn subsample distances 
	       (supplied through a call to get_cart_linesamp()) for data 
	       in the DAO geographic grid.

!Input Parameters:
iline: line number containing the point in question
isamp: saple number containing the point in question
smaps: number of collumns in the input DAO data file
rowfactor: vertical distance of the point in question from the 
           center of iline to the center of the next southern line.
colfactor: horizontal distance of the point in question from the 
           center of isamp to the center of the next eastern pixel.
anc: ancillary data structure

!Output Parameters: none

!Input/Output Parameters: none

!Returns: interpolated data value

!Revision History:
	Original version 1997, P. Fisher (fisher@ltpmail.gsfc.nasa.gov).


!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Paul Fisher
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		fisher@ltpmail.gsfc.nasa.gov, (301) 286-2980

!Design Notes
!END****************************************************************************/
{
  float32 v1, v2, v3, v4;
  int32 i1, i2, i3, i4;
  float32 interp;
  
  /* obtain four corners */
  v1= 0.0;
  v2= 0.0;
  v3= 0.0;
  v4= 0.0;
  i1= iline * samps + isamp;
  i2= (iline + 1) * samps + isamp;
  i3= (iline + 1) * samps + isamp+1;
  i4= iline * samps + isamp+1;
  if (i1 >= 0) v1= anc->data[i1];
  if (i2 >= 0) v2= anc->data[i2];
  if (i3 >= 0) v3= anc->data[i3];
  if (i4 >= 0) v4= anc->data[i4];

/*  v1 = anc->data[iline * samps + isamp];
  v2 = anc->data[(iline + 1) * samps + isamp];
  v3 = anc->data[(iline + 1) * samps + isamp+1];
  v4 = anc->data[iline * samps + isamp+1];
 */
  interp=(1.0F-rowfactor)*(1.0F-colfactor)*v1 +  
    rowfactor*(1.0F-colfactor)*v2 +
    rowfactor*colfactor*v3 + 
    colfactor*(1.0F-rowfactor)*v4;

  return(interp);

}


float32 getaerosol(t_ancillary *anc,float32 lat,float32 lon)
/*
!C*****************************************************************************
!Description:  interpolates data on the POLDER grid through a bilinear
               interpolation, except at the poles where it performs
               a nearest neighbor resampling (i.e. interpolates data on
	       the POLDER grid, but does not extrapolate near the poles).

!Input Parameters:
anc: data structure containing ancillary data
lat: latitude of the point in question
lon: longitude of the point in question

!Output Parameters: none

!Input/Output Parameters: none

!Returns: interpolated value

!Revision History:
	Original version 1997, P. Fisher (fisher@ltpmail.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Paul Fisher
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		fisher@ltpmail.gsfc.nasa.gov, (301) 286-2980

!Design Notes
!END***************************************************************************
*/
{
  float32 fline, fsamp;
  float32 dline, dupsamp, ddownsamp;
  int32 iline, isamp, upline, downline;
  int32 leftupsamp, rightupsamp, leftdownsamp, rightdownsamp, nnsamp;
  int32 Ni;
  float32 upleftval, uprightval, downleftval, downrightval, nnval;
  float32 upval, downval, result;
  enum{NOTPOLE, POLE}pole=NOTPOLE;

  if(pol_getline(lat, &fline)) {
    modsmf(MODIS_E_INTERP_ERROR, "polder grid, L3 aerosol data will not be "
	   "used to correct this pixel.  Using default aerosol estimate "
	   "instead", "getaerosol(), ancillary.c");
    return(-1.0F);
  }

  iline=(int32)fline;
  dline=fline-iline;
  
  /* calculate upline and downline.  We're not going to interpolate across the poles,
   when lat is greater than 89.5 degrees N or S.*/
  if(iline==0 && dline<=0.5F){
    upline=0;
    pole=POLE;
  }else if(iline==(POLDERLINES-1) && dline>0.5F){
    upline=(POLDERLINES-1);
    pole=POLE;
  }else{
    if(dline>0.5F){
      upline=iline;
      downline=upline+1;
      dline=1.5F-dline;
    }
    else{
      upline=iline-1;
      downline=upline+1;
      dline=0.5F-dline;
    }
  }

  
  /* get samp for upline */
  if(pol_getsampforline(upline, lon, &fsamp, &Ni)) {
    modsmf(MODIS_E_INTERP_ERROR, "polder grid, L3 aerosol data will not be "
	   "used to correct this pixel.  Using default aerosol estimate "
	   "instead", "getaerosol(), ancillary.c");
    return(-1.0F);
  }

  /* nearest neighbor over the poles lat > 89.5 degrees north and south */
  if(pole==POLE){
    nnsamp=fsamp+.5;

    nnval=anc->data[((upline-anc->latcolmin)*(anc->loncolmax-anc->loncolmin+1))+(nnsamp-anc->loncolmin)];

    return(nnval);
  }


  /* calculate leftsamp and right samp for upline */
  isamp=(int32)fsamp;
  dupsamp=fsamp-isamp;

  if(dupsamp>0.5F){
    leftupsamp=isamp;
    dupsamp=1.5F-dupsamp;
  }else{
    leftupsamp=isamp-1;
    dupsamp=0.5F-dupsamp;
  }
  rightupsamp=leftupsamp+1;


  /* 180 correction */
  if (leftupsamp<0)leftupsamp=((Ni*2)-1);
  if(rightupsamp==(Ni*2))rightupsamp=0;

  /* get samp for downlineline */
  if(upline!=downline){
    if(pol_getsampforline(downline, lon, &fsamp, &Ni)) {
    modsmf(MODIS_E_INTERP_ERROR, "polder grid, L3 aerosol data will not be "
	   "used to correct this pixel.  Using default aerosol estimate "
	   "instead", "getaerosol(), ancillary.c");
    return(-1.0F);
  }
    
    /* calculate leftsamp and right samp for downlineline*/
    isamp=(int32)fsamp;
    ddownsamp=fsamp-isamp;
    
    if(ddownsamp>0.5F){
      leftdownsamp=isamp;
      ddownsamp=1.5F-ddownsamp;
    }else{
      leftdownsamp=isamp-1;
      ddownsamp=0.5F-ddownsamp;
    }
    rightdownsamp=leftdownsamp+1;
    
    /* 180 correction */
    if (leftdownsamp<0)leftdownsamp=((Ni*2)-1);
    if(rightdownsamp==(Ni*2))rightdownsamp=0;
  }

  /* get values to use for interpolation */
  upleftval=anc->data[((upline-anc->latcolmin)*(anc->loncolmax-anc->loncolmin+1))+(leftupsamp-anc->loncolmin)];
  uprightval=anc->data[((upline-anc->latcolmin)*(anc->loncolmax-anc->loncolmin+1))+(rightupsamp-anc->loncolmin)];
  if(upline!=downline){
    downleftval=anc->data[((downline-anc->latcolmin)*(anc->loncolmax-anc->loncolmin+1))+(leftdownsamp-anc->loncolmin)];
    downrightval=anc->data[((downline-anc->latcolmin)*(anc->loncolmax-anc->loncolmin+1))+(rightdownsamp-anc->loncolmin)];
  }
  

  /* interpolate within each row */
  upval=upleftval*dupsamp + uprightval*(1.0F-dupsamp);
  if(upline!=downline)downval=downleftval*ddownsamp + downrightval*(1.0F-ddownsamp);

  /* interpolate across the rows */
  if(upline!=downline)result=upval*dline+downval*(1.0F-dline);
  else result=upval;

  return(result);


}




int32 writeancheader(PGSt_IO_Gen_FileHandle *fd,t_ancillary *ancillary,char *msg)
{
/*
!C******************************************************************

!Description: function used to Epro_write an ancillary type header file,
given a ancillary struture and an optional message.

!Input Parameters:
fd: file address
msg: message for the header
ancillary: the address of the ancillary structure.

!Output Parameters:

writeancheader: status (0 ok, -1 bad)


!Revision History:
$LOG: ancillary.c,v1.0$

Original version
E. VERMOTE (eric@kratmos.gsfc.nasa.gov)


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits
!Design Notes
!END****************************************************************************
*/
   char header[256];
   int32 i;

   sprintf(header,"KRAPS_ANCILLARY_DATA_%s",msg);
   sprintf(&header[strlen(header)],
           "_LATMIN=%07.2f_LATMAX=%07.2f_DELTALAT=%09.6f",
           ancillary->latmin,ancillary->latmax,ancillary->deltalat);
   sprintf(&header[strlen(header)],
           "_LONMIN=%07.2f_LONMAX=%07.2f_DELTALON=%09.6f",
           ancillary->lonmin,ancillary->lonmax,ancillary->deltalon);

   sprintf(&header[strlen(header)],"_LINES=%04d_COLS=%04d",
           ancillary->nblines,ancillary->nbcols);
   for (i = (int32) strlen(header);i < 255; i++)
    header[i]='*';
   header[255]='\0';
   if (fwrite(header,255,1,fd)!=1)
    return -1;
   else
    return 0;
   }


int32 getancpars(PGSt_IO_Gen_FileHandle *fd,t_ancillary *ancillary)
{
/*
!C******************************************************************

!Description: function used to Epro_read an ancillary type header file,
into an ancillary struture.

!Input Parameters:
fd: file address


!Output Parameters:
ancillary:  the ancillary structure. (updated if sucessfull)
getancpars: status (0 ok,-1,1 bad)


!Revision History:
$LOG: ancillary.c,v1.0$

Original version
E. VERMOTE (eric@kratmos.gsfc.nasa.gov)


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758

!References and Credits

!Design Notes

!END****************************************************************************
*/
   char header[256],tempo[20];


   if (fd==NULL)
      return -1;
   fseek(fd,0L,SEEK_SET);
   fread(header,255,1,fd);
   header[255]='\0';
   if (!strncmp(header,"MODIS_BETA3_ANCILLARY_DATA_",27)) {
      strncpy(tempo,(strstr(header,"LATMIN=")+7),7);
      tempo[7]='\0';
      ancillary->latmin=(float32)atof(tempo);
      strncpy(tempo,(strstr(header,"LATMAX=")+7),7);
      tempo[7]='\0';
      ancillary->latmax=(float32)atof(tempo);
      strncpy(tempo,(strstr(header,"DELTALAT=")+9),9);
      tempo[9]='\0';
      ancillary->deltalat=(float32)atof(tempo);
      strncpy(tempo,(strstr(header,"LONMIN=")+7),7);
      tempo[7]='\0';
      ancillary->lonmin=(float32)atof(tempo);
      strncpy(tempo,(strstr(header,"LONMAX=")+7),7);
      tempo[7]='\0';
      ancillary->lonmax=(float32)atof(tempo);
      strncpy(tempo,(strstr(header,"DELTALON=")+9),9);
      tempo[9]='\0';
      ancillary->deltalon=(float32)atof(tempo);
      strncpy(tempo,(strstr(header,"LINES=")+6),6);
      tempo[6]='\0';
      ancillary->nblines=atoi(tempo);
      strncpy(tempo,(strstr(header,"COLS=")+5),5);
      tempo[5]='\0';
      ancillary->nbcols=atoi(tempo);
      return 0;
      }
   else {
      fseek(fd,0L,SEEK_SET);
      return 1;
      }
   }



int32 getnewancpars(PGSt_IO_Gen_FileHandle *fd,char *signature, int month,
		    t_ancillary *ancillary)
{
/*          
!C******************************************************************

!Description: function used to read ancillary data files header 
into an ancillary struture.

!Input Parameters:
fd: file address
signature: ancillary data type
month: month of granule

!Output Parameters:
ancillary:  the ancillary structure. (updated if sucessful)
getancpars: status 0 on failure, 1 on success


!Revision History:
Revision 1.1, P. Fisher, SSAI/NASA's GSFC Code 923
1.  Updated to remove references to dao data.
2.  Updated to handle Tom's ozone climatology file.

Original version 
E. VERMOTE (eric@kratmos.gsfc.nasa.gov)


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771
		   
DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232 
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758
	      
	      
!References and Credits 

!Design Notes
      
!END****************************************************************************
*/
   char header[256],tempo[40],*ptrchar;
   int tmpint,contreading,origin;
   int tmpyr,tmpmo,tmpday;

   if (fd==NULL)
      return 0;
   fseek(fd,0L,SEEK_SET);
   fread(header,255,1,fd);
   header[255]='\0';


   /* 0=UNKNOWN 1=TOMSO3, 2=OORT, 3=ETOPOSP */
   if  (!strcmp(signature,"TOMSO3"))
     origin=1; 
   else if (!strcmp(signature,"OORTWV"))
     origin=2;
   else if (!strcmp(signature,"ETOPOSP"))
     origin=3;
   else
     origin=0;
   
   if(origin==1)strcpy(tempo,"MODIS_V2.0_ANCILLARY_DATA_");
   else strcpy(tempo,"MODIS_BETA3_ANCILLARY_DATA_");
   
  if (!strncmp(header,tempo,strlen(tempo))) {
   
    strcat(tempo,signature);
    
    ptrchar=strstr(header,"LATMIN=");
    if (ptrchar!=NULL) {
      strncpy(tempo,(ptrchar+7),7);
      tempo[7]='\0';
      ancillary->latmin=(float)atof(tempo);
    }
    ptrchar=strstr(header,"LATMAX=");
    if (ptrchar!=NULL) {
      strncpy(tempo,(ptrchar+7),7);
      tempo[7]='\0';
      ancillary->latmax=(float)atof(tempo);
    }
    ptrchar=strstr(header,"DELTALAT=");
    if (ptrchar!=NULL) {
      strncpy(tempo,(ptrchar+9),9);
      tempo[9]='\0';
      ancillary->deltalat=(float)atof(tempo);
    }
    ptrchar=strstr(header,"LONMIN=");
    if (ptrchar!=NULL) {
      strncpy(tempo,(ptrchar+7),7);
      tempo[7]='\0';
      ancillary->lonmin=(float)atof(tempo);
    }
    ptrchar=strstr(header,"LONMAX=");
    if (ptrchar!=NULL) {
      strncpy(tempo,(ptrchar+7),7);
      tempo[7]='\0';
      ancillary->lonmax=(float)atof(tempo);
    }
    ptrchar=strstr(header,"DELTALON=");
    if (ptrchar!=NULL) {
      strncpy(tempo,(strstr(header,"DELTALON=")+9),9);
      tempo[9]='\0';
      ancillary->deltalon=(float)atof(tempo);
    }
    ptrchar=strstr(header,"LINES=");
    if (ptrchar!=NULL) {
      strncpy(tempo,(ptrchar+6),6);
      tempo[6]='\0';
      ancillary->nblines=atoi(tempo);
    }
    ptrchar=strstr(header,"COLS=");
    if (ptrchar!=NULL) {
      strncpy(tempo,(ptrchar+5),5);
      tempo[5]='\0';
      ancillary->nbcols=atoi(tempo);
    }
    ptrchar=strstr(header,"SLOPE=");
    if (ptrchar!=NULL) {
      strncpy(tempo,(ptrchar+6),9);
      tempo[9]='\0';
      ancillary->slope=(float)atof(tempo);
    }
    ptrchar=strstr(header,"OFFSET=");
    if (ptrchar!=NULL) {
      strncpy(tempo,(ptrchar+7),9);
      tempo[9]='\0';
      ancillary->offset=(float)atof(tempo);
    }

    if (origin==2) { /* OORT : go to correct month */
      if (fseek(fd,(month-1)*(ancillary->nblines*ancillary->nbcols*2),SEEK_CUR))
	return 0;
    }
    return 1;
    
    
   }else{
     fseek(fd,0L,SEEK_SET);
     return 0;
   }
}


int32 pol_getline(float32 lat, float32 *i)
/*
!C*****************************************************************************
!Description:
	Given a latitude, return a line number from the polder grid.

!Input Parameters: lat, i

!Output Parameters:

!Input/Output Parameters:

!Returns:

!Revision History: - V2.0

!Team-unique Header:

    This software developed by the MODIS Land Science Team for 
    the National Aeronautics and Space Administration, Goddard 
    Space Flight Center, under contract NAS5-96062

!References and Credits

!Design Notes
!END***************************************************************************
*/
{

   /* special case lat */
   if(lat<=-90.0F)lat=-89.999F;

   *i=((90.0F-lat)*((float32)POLDERLINES)/180.0F);

   if ((*i<0)||(*i>=POLDERLINES)) return 1;
   return 0;

}


int32 pol_getsampforline(int32 i, float32 lon, float32 *j, int32 *Ni)
/*
!C*****************************************************************************
!Description:
	Given a line in the polder grid and a longitude, return the sample
	number and number of samples as parameters.

!Input Parameters: i, lon, j, Ni

!Output Parameters:

!Input/Output Parameters:

!Returns:

!Revision History: - V2.0

!Team-unique Header:

    This software developed by the MODIS Land Science Team for 
    the National Aeronautics and Space Administration, Goddard 
    Space Flight Center, under contract NAS5-96062

!References and Credits

!Design Notes
!END***************************************************************************
*/
{

   /* special case lon */
   if (lon >= 180.0F) lon -= 360.0F;

   *Ni=(int32)((1080.0F*cosf(PI/2.0F-PI*(i+1-0.5F)/((float32)POLDERLINES)))+0.5F);
   *j=(lon*(*Ni)/180.0F+((float32)POLDERLINES)); 
   if ((*j<(POLDERLINES-(*Ni)))||(*j>=POLDERLINES+(*Ni))) return 1;
   return 0;
}


int32 pol_geotogrid(float32 lat,float32 lon,float32 *i,float32 *j, int32 *Ni)
/*
!C*****************************************************************************
!Description:
	Convert specified geographic position into grid POLDER coordinate.

!Input Parameters:
	lat		latitude (degrees)
	lon		longitude (degrees)

!Output Parameters:
	i		grid row
	j		grid column

	Returns 1 if point is not contained in grid, 0 otherwise.

!Revision History:
        Revision 1.2, Paul Fisher (SSAI/NASA's GSFC Code 923)
	1. Altered calling sequence so that i and j are float32s. 
	   This allows the retrieval of sub-pixel info.
	2. Returns Ni as a parameter.

        Revision 1.1, Paul Fisher (SSAI/NASA's GSFC Code 923)
	1.  Altered computation to exactly match POLDER specifications.

	Original version, Nazmi El Saleous (nazmi@kratmos.gsfc.nasa.gov).

!Team-unique Header:
        References and Credits
        
!Design Notes
1. Ni is 1/2 the number of valid samples per line.

!END***************************************************************************
*/
{

   /* special case lat */
   if(lat<=-90.0F)lat=-89.999F;

   /* special case lon */
   if (lon >= 180.0F) lon -= 360.0F;

   *i=(int32)((90.0F-lat)*((float32)POLDERLINES)/180.0F);
   *Ni=(int32)((1080.0F*cosf(PI/2.0F-PI*(*i+1-0.5F)/((float32)POLDERLINES)))+0.5F);
   *j=(int32)(lon*(*Ni)/180.0F+((float32)POLDERLINES)); 
   if ((*i<0)||(*i>=POLDERLINES)||(*j<(POLDERLINES-(*Ni)))||(*j>=POLDERLINES+(*Ni))) return 1;
   return 0;
}


PGSt_SMF_status read_dao_plane(PGSt_PC_Logical file_lun, 
			       emeta_t **metadata, char *daovar, d_ancillary *ancillary)
/*
!C*****************************************************************************
!Description: On the basis of time, decides which plane of data in
              a DAO file should be read, and reads in that plane of data.

!Input Parameters:
file_lun: DAO File logical unit number
version: version number of the file to be read
metadata: metadata structure
daovar: name of the SDS to be read from the file
ancillary: d_ancillary ancillary data structure

!Output Parameters:
data: pointer to hold the data read
!Input/Output Parameters: none

!Returns: on success: MODIS_S_SUCCESS 
          on failure: MODIS_E_HDF_FAILURE
                      MODIS_E_GETREF_FAIL
!Revision History:
        Revision 2.0, P. Fisher, SSAI/NASA's GSFC Code 923
	1.  Modified to accurately read in DAO files beginning at 21:00

        Revision 2.0, P. Fisher, SSAI/NASA's GSFC Code 923
	1.  Modified, to select the correct data plane out of two potential 
	input files (days).
	2.  Data buffered, so that the same ancillary data is not read 
        in twice in a row.

        Revision 1.1, P. Fisher, SSAI/NASA's GSFC Code 923
	1. Revised to accept the d_ancillary structure as input
	2. Revised to read the line and sample parameters directly
	   from the input file.

	Original version 1997, P. Fisher (fisher@ltpmail.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Paul Fisher
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		fisher@ltpmail.gsfc.nasa.gov, (301) 286-2980

!Design Notes
1.  This function assumes the DAO files are 24 hours, beginning at 21:00 and
    divided into 3 hour time slices.
!END***************************************************************************
*/
{
  int32 sdid;
  int32 index;
  int32 sdsid;
  int32 hour;
  char chour[PGSd_PC_VALUE_LENGTH_MAX];
  int32 timeplane;
  float32 dectime;
  int32 start[3]={0,0,0};
  int32 edge[3]={1,1};
  char sds_name[200];
  int32 rank;
  int32 data_type;
  int32 nattrs;
  int32 timestep;
  float32 *datap;
  char hour_st[3], *tp;

  PGSt_double julian[2];
  int jules;

  char msg[PGS_SMF_MAX_MSG_SIZE]; /* buffer to hold SMF message */
  char tf_name[PGSd_PC_FILE_PATH_MAX]={'\0'};	    
  int32 i;
  PGSt_integer ver, sver;
  char location[]="read_dao_plane(), ancillary.c";

  if(PGS_TD_UTCtoUTCjd(metadata[BUILDBUFFER]->rbdt, julian)!=PGS_S_SUCCESS){
    modsmf(MODIS_E_BAD_TIME_CONV, "Unable to convert date/time of granule to julian date.  "
	   "Unable to process DAO data.", location);
    return(MODIS_E_BAD_TIME_CONV);
  }
  jules=(int)julian[0];

  /* note that dao files start @ 2100 the day before -- usually, but not always */
  tp=strchr(metadata[BUILDBUFFER]->rbdt, 'T');
  tp++;

  strncpy(hour_st, tp, 2);
  hour_st[2]='\0';
  hour=atoi(hour_st);

/*  Forget this, we've already figured which file to use in "read_metadata()", in
    meta.c.  

  for(i=0;i<2;i++)if((ancillary->jrbd[i]==jules && hour<21) ||
		     (ancillary->jrbd[i]==jules-1 && hour>=21))break;

  if(i==2){
    sprintf(msg, "LUN %d", (int)file_lun);
    modsmf(MODIS_E_DAO_FAIL, msg, location);
    return(MODIS_E_DAO_FAIL);
  }

 Just set i equal to 0; we can probably make ancillary->jrbd[] and 
 ancillary->versions[] into scalars also...
 */
 
  i = 0;
  
  sver=ver=(PGSt_integer)ancillary->versions[i];
  if(PGS_PC_GetReference(file_lun, &ver, tf_name)!=PGS_S_SUCCESS){
    sprintf(msg, "LUN %d, version %d", (int)file_lun, (int)ancillary->versions[i]);
    modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, msg, location);
    return(MODIS_E_GETREF_FAIL);
  }

/*printf("read_dao_plane(): LUN %d, version %d, file '%s'\n", file_lun, ancillary->versions[i],
        tf_name);
 */
  
  if((sdid=SDstart(tf_name, DFACC_RDONLY))==FAIL) {
    sprintf(msg, "LUN %d, version %d", (int32)file_lun, (int)ancillary->versions[i]);    
    modsmf(MODIS_E_OPEN_HDF_FILE, msg, "read_dao_plane(), modis_srefl.c");
    return (MODIS_E_OPEN_HDF_FILE);
  }

  if((index=SDnametoindex(sdid, daovar))==FAIL){ 
    modsmf(MODIS_E_HDF_FAILURE, "SDnametoindex()", location);
    SDend(sdid);
    return (MODIS_E_HDF_FAILURE);
  }

  if((sdsid=SDselect(sdid, index))==FAIL){
    modsmf(MODIS_E_HDF_FAILURE, "SDselect()",  location);
    SDend(sdid);
    return (MODIS_E_HDF_FAILURE);
  }
    
  if(SDgetinfo(sdsid, sds_name, &rank, ancillary->dims, &data_type, &nattrs)==FAIL){
    modsmf(MODIS_E_HDF_FAILURE, "SDgetinfo()", location);
    SDendaccess(sdsid);
    SDend(sdid);
    return (MODIS_E_HDF_FAILURE);
  }

  if(SDgetfillvalue(sdsid, &(ancillary->fillvalue))==FAIL){
       modsmf(MODIS_E_HDF_FAILURE, "SDgetfillvalue(), using default DAO fill value", 
	      location);
       ancillary->fillvalue=9.9999999e+14F;
  }

  /* get latitude and longitude resolution */
  ancillary->lonres=360.0/ancillary->dims[2];
  ancillary->latres=180.0/(ancillary->dims[1]-1);
    
  
  timestep=24/ancillary->dims[0];
  /* Assume three hour slices */
  if(timestep!=3){
    sprintf(msg, "LUN %d.  There are more or less than 8 time slices in the file", (int)file_lun);
    modsmf(MODIS_E_DAO_FAIL, msg, location);
  }
  
  /* Which plane do we need... */
  timeplane=hour/timestep;
  
  /* shift is necessary IF the file starts at 2100 hours */
  if (ancillary->jrbd[i] == 21) timeplane++;
  
  if (timeplane>=ancillary->dims[0])timeplane=0;
    
  /* don't read again, if the correct data is already in the buffer */
  if(timeplane!=ancillary->lasttimeplane || sver!=ancillary->lastversion[0]){
    
    
    ancillary->lasttimeplane=timeplane;
    ancillary->lastversion[0]=sver;
    
    
    /* allocate data array -- for safety's sake, all DAO files may not have the same 
       dimensions -- who knows!!!*/
    ancillary->data=(float32 *)ReAllocMem((void *)ancillary->data, 
					  ancillary->dims[1]*ancillary->dims[2]*sizeof(float32));
    
    /* initialize and transfer to maintain portability */
    start[0]=(int32)timeplane;
    edge[2]=(int32)ancillary->dims[2];
    
    datap=ancillary->data;
    
    for(i=ancillary->dims[1]-1;i>-1;i--){
      
      start[1]=(int32)i;

      if(SDreaddata(sdsid, start, NULL, edge, datap)==FAIL){
	modsmf(MODIS_E_HDF_FAILURE, "SDreaddata()", 
	       "read_dao_plane(), modis_srefl.c");
	SDendaccess(sdsid);
	SDend(sdid);
	return (MODIS_E_HDF_FAILURE);
      }
      
      datap+=ancillary->dims[2];
    }
  }/* if(timeplane... */
  
  SDendaccess(sdsid);
  SDend(sdid);
  
  
  return(MODIS_S_SUCCESS);
  
}


PGSt_SMF_status read_plane(PGSt_PC_Logical *file_lun, 
			   emeta_t **metadata, d_ancillary *ancillary)
/*
!C*****************************************************************************
!Description: On the basis of preset tags in the ancillary structure, tries
              to read DAO or GRIB data.

!Input Parameters:
file_lun: DAO File logical unit number
metadata: metadata structure
ancillary: d_ancillary ancillary data structure

!Output Parameters:
none

!Input/Output Parameters: 
ancillary is updated.

!Returns: on success: MODIS_S_SUCCESS 
          on failure: MODIS_E_HDF_FAILURE
                      MODIS_E_GETREF_FAIL
!Revision History:

	Original version 1999, Jim Ray (jim@cratmos.gsfc.nasa.gov), derived
	                       from read_dao_plane() (above).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes

!END***************************************************************************
*/
{
  int32 timeplane, hour;
  char hour_st[3], *tp;
  PGSt_double julian[2];
  int jules;
  
  /* variables for reading from GRIB files */
  FILE *fd;
  char where[15];
  int nx,ny;
  int grib_ret;
  int will_read, will_read_interpolate;
 /* static float *tmpfltarray[6][2] = {{NULL,NULL},{NULL,NULL},{NULL,NULL},
                                     {NULL,NULL},{NULL,NULL},{NULL,NULL}};*/
  /* 27-APR-05 */				     
  static float *tmpfltarray[7][2] = {{NULL,NULL},{NULL,NULL},{NULL,NULL},
                                     {NULL,NULL},{NULL,NULL},{NULL,NULL},{NULL,NULL}};
  int tempfltindex;
  int read_grib_array(FILE *input, char *what, char *where, int *nx, int *ny, float **narray);

  char msg[PGS_SMF_MAX_MSG_SIZE]; /* buffer to hold SMF message */
  char tf_name[PGSd_PC_FILE_PATH_MAX]={'\0'};	    
  int32 i,j;
  PGSt_integer ver, sver;
  PGSt_integer index;
  char location[]="read_plane(), ancillary.c";

  /* figure out which file to read */
  if (!strcmp(ancillary->anci_source, "DAO\0")) {
  
     /* If it's DAO, get the right LUN and call read_dao_plane() */
     if (!strcmp(ancillary->tag, DAOOZONEVARIABLE)) index = DAOOZONE;
     else if (!strcmp(ancillary->tag, DAOWVVARIABLE)) index = DAOWV;
     else if (!strcmp(ancillary->tag, DAOSPRESVARIABLE)) index = DAOSPRES;
     else {
        sprintf(msg, "Unexpected ancillary data file tag: '%s'", ancillary->tag);
        modsmf(MODIS_F_CODING_ERROR, msg, location);
     }
     return (read_dao_plane(file_lun[index], metadata, ancillary->tag, ancillary));

                                                } /* if (!strcmp(ancillary->anci_source, "DA.. */
						
  else if ( !strcmp(ancillary->anci_source, "GRIB\0")   || !strcmp(ancillary->anci_source, "EPT\0")) {

     /* If it's GRIB, get the right LUN */
     if (!strcmp(ancillary->tag, OZONE_GRIBTAG)) {
        index = EPT_OZ;
        strcpy(where, OZONE_GRIBLEVEL);
	tempfltindex=0;
	}
     else if (!strcmp(ancillary->tag, GOZONE_GRIBTAG)) {
        index = GRIB_WV;     /* get from GDAS file */
        strcpy(where, OZONE_GRIBLEVEL);
	tempfltindex=1;
	}
     else if (!strcmp(ancillary->tag, WV_GRIBTAG)) {
        index = GRIB_WV;
        strcpy(where, WV_GRIBLEVEL);
	tempfltindex=2;
	}
     else if (!strcmp(ancillary->tag, SPRES_GRIBTAG)) {
        index = GRIB_SP;
        strcpy(where, SPRES_GRIBLEVEL);
	tempfltindex=3;
	}
     else if (!strcmp(ancillary->tag, TMP_GRIBTAG)) { /* 01-MAY-01 */
        index = GRIB_SP;  /* read from the same file you read SP from */
        strcpy(where, TMP_GRIBLEVEL);
	tempfltindex=4;
	}
     else if (!strcmp(ancillary->tag, UWD_GRIBTAG)) { /* 14-AUG-01 */
        index = GRIB_SP;  /* read from the same file you read SP from */
        strcpy(where, UWD_GRIBLEVEL);
	tempfltindex=5;
	}
     else if (!strcmp(ancillary->tag, VWD_GRIBTAG)) { /* 14-AUG-01 */
        index = GRIB_SP;  /* read from the same file you read SP from */
        strcpy(where, VWD_GRIBLEVEL);
	tempfltindex=6;
	}
     else {
        sprintf(msg, "Unexpected ancillary data file tag: '%s'", ancillary->tag);
        modsmf(MODIS_F_CODING_ERROR, msg, location);
    }
     
     
     will_read=will_read_interpolate=0;
     
     if (ancillary->numdays==0) {
         /* then you don't have to read two sets of data, just one... */
	 
	 if (ancillary->versions[0] != ancillary->lastversion[0]){  /* if it hasn't been read yet */
	 
	    will_read = 1;
            sver=ver=(PGSt_integer)ancillary->versions[0];
            if(PGS_PC_GetReference(file_lun[index], &ver, tf_name)!=PGS_S_SUCCESS){
               sprintf(msg, "LUN %d, version %d", (int)file_lun[index], (int)ancillary->versions[i]);
               modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, msg, location);
               return(MODIS_E_GETREF_FAIL);
            }
            nx = ny = 0;
            if ((fd=fopen(tf_name,"rb")) == NULL) {
               sprintf(msg, "LUN %d, version %d", (int32)file_lun[index], sver);
               modsmf(MODIS_E_OPEN_GEN_FILE, msg, location);
               return(MODIS_E_OPEN_GEN_FILE);
   	    }
	    grib_ret=read_grib_array(fd, ancillary->tag, where, &ny, &nx, &(ancillary->data));
            /*printf("%s\n*** return value = %d, read %s, %d by %d ***\n", 
	        location,grib_ret,ancillary->tag, nx, ny); */
             
	    fclose(fd);
            ancillary->lastversion[0]=sver;
	   	  
            /* for now, use MODIS_E_HDF_FAILURE */
            if (grib_ret) {
	       return (MODIS_E_HDF_FAILURE);
	       }
	 
            ancillary->dims[2] = (int32)nx;
            ancillary->dims[1] = (int32)ny;
            ancillary->dims[0] = (int32)1;  /* ?? */
 
            /* get latitude and longitude resolution */
            ancillary->lonres=360.0/ancillary->dims[2];
            ancillary->latres=180.0/(ancillary->dims[1]-1);
       
            /* Note: as far as I know, there are no fill values in GRIB files. 
	       Routine 'read_grib_array()' uses UNDEFINED for missing values;
	       use same thing for fill values (14-JUL-99, Jim Ray, SSAI).  */
            ancillary->fillvalue=UNDEFINED;
	 
	 }
     }  /* if (ancillary->numdays==0) */
     else {
	 if ((ancillary->versions[0] != ancillary->lastversion[0])||
             (ancillary->versions[1] != ancillary->lastversion[1])) {  /* if they haven't been read yet */

	     will_read_interpolate = 1;	     
	     for(i=0;i<=ancillary->numdays;i++) {  /* 03-MAY-01 */

                 sver=ver=(PGSt_integer)ancillary->versions[i];
                 if(PGS_PC_GetReference(file_lun[index], &ver, tf_name)!=PGS_S_SUCCESS){
                    sprintf(msg, "LUN %d, version %d", (int)file_lun[index], (int)ancillary->versions[i]);
                    modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, msg, location);
                    return(MODIS_E_GETREF_FAIL);
                 }
                 nx = ny = 0;
                 if ((fd=fopen(tf_name,"rb")) == NULL) {
                    sprintf(msg, "LUN %d, version %d", (int32)file_lun[index], sver);
                    modsmf(MODIS_E_OPEN_GEN_FILE, msg, location);
                    return(MODIS_E_OPEN_GEN_FILE);
   	         }
		 
 	         if (tempfltindex == 0)   /* 27-APR-05: you are reading OZONE, use EP-TOMS reader */
	             grib_ret=read_EPT_array(fd, &ny, &nx, &tmpfltarray[tempfltindex][i]);
		 else   
 	             grib_ret=read_grib_array(fd, ancillary->tag, where, &ny, &nx, &tmpfltarray[tempfltindex][i]);

/* printf("%s LUN = %d, return value = %d, read %s, %d by %d tempflindex = %d***\n", 
  	           location,file_lun[index],grib_ret,ancillary->tag, nx, ny, tempfltindex);*/
                
	         fclose(fd);
                 ancillary->lastversion[i]=sver;
		 
                 /* for now, use MODIS_E_HDF_FAILURE */
                 if (grib_ret) {
	            return (MODIS_E_HDF_FAILURE);
	            }
	 
                 ancillary->dims[2] = (int32)nx;
                 ancillary->dims[1] = (int32)ny;
                 ancillary->dims[0] = (int32)1;  /* ?? */
 
                 /* get latitude and longitude resolution */
                 ancillary->lonres=360.0/ancillary->dims[2];
                 ancillary->latres=180.0/(ancillary->dims[1]-1);
       
                 /* Note: as far as I know, there are no fill values in GRIB files. 
	            Routine 'read_grib_array()' uses UNDEFINED for missing values;
	            use same thing for fill values (14-JUL-99, Jim Ray, SSAI).  */
                 ancillary->fillvalue=UNDEFINED;
		 
	     }
	 }
	 
	 /* Apply weights to the data; freshly read or not */
         ancillary->data = (float *) ReAllocMem(ancillary->data, ancillary->dims[2]*ancillary->dims[1]*sizeof(float));	 


         if (tempfltindex == 0) {  /* EP-TOMS ozone */
	     /* 30-APR-04: Set the higher weight to 1.0, the lower to 0.0 */
	     if (ancillary->weights[0] > ancillary->weights[1]) {
	         ancillary->weights[0] = 1.0F;
	         ancillary->weights[1] = 0.0F;
	        }
	     else {
	         ancillary->weights[0] = 0.0F;
	         ancillary->weights[1] = 1.0F;
	        }
	     }

/*printf("%s: Applying weights %f and %f\n", ancillary->tag, ancillary->weights[0], ancillary->weights[1]);*/
 
	 for (j=0;j<ancillary->dims[2]*ancillary->dims[1];j++) {
	     ancillary->data[j] = tmpfltarray[tempfltindex][0][j]*ancillary->weights[0] +
	                          tmpfltarray[tempfltindex][1][j]*ancillary->weights[1];
	     }
	     
         /*if (tempfltindex==2){
	 {FILE *fd;
	 char filename[50];
	 struct timeval  first;
         struct timezone tzp;

         gettimeofday (&first, &tzp);
 	 sprintf(filename, "array_%d_0.img\0", first.tv_usec);
	 fd=fopen(filename, "wb");
	 fwrite(tmpfltarray[tempfltindex][0], ancillary->dims[2]*ancillary->dims[1]*sizeof(float),1,fd);
	 fclose(fd);

	 sprintf(filename, "array_%d_1.img\0", first.tv_usec);
	 fd=fopen(filename, "wb");
	 fwrite(tmpfltarray[tempfltindex][1], ancillary->dims[2]*ancillary->dims[1]*sizeof(float),1,fd);
	 fclose(fd);


	 sprintf(filename, "array_%d_%f_%f.img\0", first.tv_usec, ancillary->weights[0],
	               ancillary->weights[1]);
	 fd=fopen(filename, "wb");
	 fwrite(ancillary->data, ancillary->dims[2]*ancillary->dims[1]*sizeof(float),1,fd);
	 fclose(fd);
	 }
	 }*/
     }
     
     /* 26-APR-01: convert NCEP units into our units: 
	PWAT is in kg/m^2, convert to g/cm^2 (divide by 10.0)
	SPRES is in Pa, convert to mbar (divide by 100.0).
      */
     if (!strcmp(ancillary->tag, SPRES_GRIBTAG)) {
	   for (j=0;j<ancillary->dims[2]*ancillary->dims[1];j++) ancillary->data[j] /= 100.0F;
        }
     else if (!strcmp(ancillary->tag, WV_GRIBTAG)) {
	   for (j=0;j<ancillary->dims[2]*ancillary->dims[1];j++) ancillary->data[j] /= 10.0F;
        }
      

     return(MODIS_S_SUCCESS);
  }  /* else if (!strcmp(ancillary->anci_source, "GRIB\0")) */
  else {  /* ancillary->anci_source has garbarge or nothing; in all likelihood, 
             none of the ancillary data files for this type of data are available */
        return (MODIS_E_HDF_FAILURE);
       }
  
}

int fill_ozone(d_ancillary *gdasozone, d_ancillary *daoozone, t_ancillary anc_ozone)
/*
!C**********************************************************************
!Description:   
  Fills holes in the daoozone data (when it's read from EPTOMS files, it has holes).
 
!Input Parameters:
  d_ancillary *gdasozone      "GDAS" ozone, which should be TOVS ozone as read from 
                                 GDAS files (GDAS_0ZF) or grb files (OZ_DAILY)
  t_ancillary anc_ozone       climatological ozone, as read from toms_10yr.anc (LUN 
                                 209554)  (Not used.)

!Input/Output Parameters:
  d_ancillary *daoozone
  
!Output Parameters:
  none
  
Global Variables:
  none

!Revision History:

!Team-unique Header:
!References and Credits:

!Design Notes: 
   

!END********************************************************************
*/
{
long li, lj;
int i, j, ii;

   if (gdasozone->wasread == 1)  {
       if (daoozone->wasread == 1) {
          for(ii=0;ii<gdasozone->dims[1]*gdasozone->dims[2];ii++) {
               if (daoozone->data[ii] == 0.0F)  daoozone->data[ii] = gdasozone->data[ii];
              }
          }
       else {
          daoozone->dims[0] = (int32)1;
          daoozone->dims[1] = gdasozone->dims[1];
          daoozone->dims[2] = gdasozone->dims[2];
          daoozone->lonres = gdasozone->lonres;
          daoozone->latres = gdasozone->latres;
	  daoozone->data = (float *) ReAllocMem(daoozone->data, daoozone->dims[1]*daoozone->dims[2]*sizeof(float));	 
          for(ii=0;ii<gdasozone->dims[1]*gdasozone->dims[2];ii++) {
               daoozone->data[ii] = gdasozone->data[ii];
              }
          }
      }
return(0);
}
			
int read_EPT_array(FILE *oz_fp, int *nlat, int *nlon, float **foz_data) 
/*
!C**********************************************************************
!Description:   
  Reads data from EP-TOMS ASCII files.
 
!Input Parameters:
  FILE *oz_fp    -- File pointer (return value of fopen()).

!Output Parameters:
  int  *nlat     -- number of latitude points for TOMS data
  int  *nlon     -- number of longitude points for TOMS data
  float **foz_data

Global Variables:

!Revision History:
  1997/10/?? Neal Devine -- Initial Version
  
  Adapted from MOD_PROZN code, routine Read_TOMS_Ozone(), on 29-APR-04 by Jim Ray (SSAI).

!Team-unique Header:
  !References and Credits:
     Neal Devine (neal.devine@gsfc.nasa.gov)

     This software is developed by the MODIS Science Data Support
     Team for the National Aeronautics and Space Administration,
     Goddard Space Flight Center, under contract NAS5-32373.

     HDF portions developed at the National Center for Supercomputing
     Applications at the University of Illinois at Urbana-Champaign.

!Design Notes: 
   
PDL:
!END********************************************************************
*/
{
#define MAX_LINE  333
#define NLAT_MAX  181
#define NLON_MAX  360
  char msg[256];
  char location[50]="ancillary.c, read_EPT_array\0";
  int ret_pgs;
  char line[333],one_value[4];
  int i,j,ilat,ilon,ilon_in_line,yday,year;
  float32 a,b,c,d;
  long li,lj;
  int16 *data;

  /* Read first three lines of TOMS Ozone file
   *  line 1) day & year,
   *  line 2) # longitude bins.
   *  line 3) # latitude bins.
   */
  for ( i = 0 ; i < 3 ; i++ ) {
    if ( fgets(line, MAX_LINE, oz_fp) == NULL ) {
      sprintf(msg, "failed to read line %d of EP-TOMS Ozone file", i);
      modsmf(MODIS_E_GET_METADATA, msg, location);
      return( 1 );
    }
    switch (i) {
    case 0 : sscanf( line, "%*s %d %*s %*d, %d", &yday, &year ); break;
    case 1 : sscanf( line, "%*s %d", nlon ); break;
    case 2 : sscanf( line, "%*s %*s %d" , nlat ); break;
      /*    case 0 : sscanf( line, "Day: %d %*s %*d, %d", yday, year ); break;*/
      /*    case 1 : sscanf( line, "Longitudes: %d", nlon ); break;*/
      /*    case 2 : sscanf( line, "Latitudes : %d" , nlat ); break;*/
    }
  }

  /* Read the data which is written from south to north and west_to_east (most rapid).
   * The data is stored as 4char ascii with upto 25 values per line.
   * The output data is organized north to south (reversed from input).
   * The transformation of the data is from SeaWiFS code o3nrt.c (if i > 0...).
   */
  if ((data = (int16 *) malloc(sizeof(int16) * (*nlat) * (*nlon) )) == NULL) return(-1);
  for ( ilat = 0 ; ilat < *nlat ; ilat++ ) {
    for ( ilon = 0 ; ilon < *nlon ; ilon++ , ilon_in_line++) {
      if ( ilon % 25 == 0 ) {
	ilon_in_line = 0;
	if ( fgets(line, MAX_LINE, oz_fp) == NULL ) {
	  printf("ERROR: failed to read line with lat,ilon = %f,%d of TOMS Ozone file\n", 
		     (180./ *nlat)*(ilat+.5)-90., ilon);
	  return( 2 );
	}
      }
      strncpy( one_value, line+1+3*ilon_in_line, 3 );  /* necessary to handle "  0###" correctly. */
      if ( sscanf( one_value, "%3d", &i ) != 1 ) {
	printf("ERROR: failed to sscanf line with lat,ilon = %f,%d of TOMS Ozone file\n", 
		     (180./ *nlat)*(ilat+.5)-90., ilon);
	return( 4 );
      }
      if ( i > 0 ) data[ (*nlat-1-ilat)*(*nlon)+ilon ] = i;
      else         data[ (*nlat-1-ilat)*(*nlon)+ilon ] = 0;
    }
  }

  /* this converts it all to floats, and expands the column-wise dimension from
     288 (EP-TOMS default) to 360 (which matches GDAS, GRB, etc.) */
  if ((*foz_data = (float *) malloc(sizeof(float)*NLON_MAX*NLAT_MAX)) == NULL) return(-1);
            
  for(i=0;i<*nlat;i++) {
        li = i*(*nlon);
	lj = i*NLON_MAX;
        for(j=0;j<((*nlon)/4);j++) {
	    a = (float32)data[li + (j*4) + 0];
	    b = (float32)data[li + (j*4) + 1];
	    c = (float32)data[li + (j*4) + 2];
	    d = (float32)data[li + (j*4) + 3];
	    (*foz_data)[lj++] = a;
	    
            /* do NOT include zeros in this process */
	    if ((a != 0.0F) && (b != 0.0F)) (*foz_data)[lj++] = a*0.25F + b*0.75F;
	    else                            (*foz_data)[lj++] = 0.0F;
	    if ((c != 0.0F) && (b != 0.0F)) (*foz_data)[lj++] = b*0.50F + c*0.50F;
	    else                            (*foz_data)[lj++] = 0.0F;
	    if ((c != 0.0F) && (d != 0.0F)) (*foz_data)[lj++] = d*0.25F + c*0.75F;
	    else                            (*foz_data)[lj++] = 0.0F;

	    (*foz_data)[lj++] = d;
        }
     }
  
  free(data);
  *nlon = NLON_MAX;
  *nlat = NLAT_MAX;
  return( 0 );

}

		
float32 local_interpolate(int32 iline, int32 lines, int32 isamp, int32 samps, 
                          float32 rowfactor, float32 colfactor, float32 *data)
{
  float32 v1, v2, v3, v4;
  int32 i1, i2, i3, i4;
  float32 interp;
  /* obtain four corners */

  v1= 0.0;
  v2= 0.0;
  v3= 0.0;
  v4= 0.0;
  i1= iline * samps + isamp;
  i2= (iline + 1) * samps + isamp;
  i3= (iline + 1) * samps + isamp+1;
  i4= iline * samps + isamp+1;

  if (isamp+1 >= samps) {  /* 29-OCT-01 */
     i3= (iline + 1) * samps;
     i4= iline * samps; 
  }
  if (i1 >= 0) v1= data[i1];
  if (i2 >= 0) v2= data[i2];
  if (i3 >= 0) v3= data[i3];
  if (i4 >= 0) v4= data[i4];
  
  interp=(1.0F-rowfactor)*(1.0F-colfactor)*v1 +  
    rowfactor*(1.0F-colfactor)*v2 +
    rowfactor*colfactor*v3 + 
    colfactor*(1.0F-rowfactor)*v4;

  return(interp);

}


