/*
!C-INC************************************************************************

File: ids.h

!Description:  Header file for ids.c, containing names of SDSs, enums, and LUNs.

!Input Parameters: (none)

!Input/output Parameters: (none)

!Output Parameters: (none)

!Revision History:

!Team-unique Header:
 
   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits:
    MODIS Science Team Members:
      Chris Justice
      MODIS Land Science Team           University of Maryland
      justice@kratmos.gsfc.nasa.gov     NASA/GSFC Code 923
      phone: 301-286-7372               Greenbelt, MD 20771
 
      Eric F Vermote
      Univerisity of MD.
      NASA's GSFC Code 923
      Greenbelt, MD , 20771
      eric@kratmos.gsfc.nasa.gov
      301-286-6232

      Jim Ray
      Science Systems and Applications Inc.
      NASA's Goddard Space Flight Center Code 923
      Greenbelt MD, 20771
      jim@kratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes

!END***************************************************************************
*/

#define COARSEB1NAME "Coarse Resolution Surface Reflectance Band 1\0"
#define COARSEB2NAME "Coarse Resolution Surface Reflectance Band 2\0"
#define COARSEB3NAME "Coarse Resolution Surface Reflectance Band 3\0"
#define COARSEB4NAME "Coarse Resolution Surface Reflectance Band 4\0"
#define COARSEB5NAME "Coarse Resolution Surface Reflectance Band 5\0"
#define COARSEB6NAME "Coarse Resolution Surface Reflectance Band 6\0"
#define COARSEB7NAME "Coarse Resolution Surface Reflectance Band 7\0"
#define COARSEB8NAME "Coarse Resolution Surface Reflectance Band 8\0"
#define COARSET1NAME "Coarse Resolution AOT Band 1\0"
#define COARSET3NAME "Coarse Resolution AOT Band 3\0"
#define COARSEQANAME "Coarse Resolution QA\0"

static char CRSGRID_Name_Array[][60]={COARSEB1NAME,  COARSEB2NAME, COARSEB3NAME,  COARSEB4NAME,  
				      COARSEB5NAME,  COARSEB6NAME, COARSEB7NAME,  
				      COARSET1NAME,  
				      COARSET3NAME,  
				      "Coarse Resolution Solar Zenith Angle\0",
				      "Coarse Resolution View Zenith Angle\0",
				      "Coarse Resolution Relative Azimuth Angle\0",
				      "Coarse Resolution Ozone\0",     /* new, 23-APR-04 */
				      "Coarse Resolution Brightness Temperature Band 20\0",     /* new, 18-MAY-04  */
				      "Coarse Resolution Brightness Temperature Band 21\0",     /* new, 18-MAY-04  */
				      "Coarse Resolution Brightness Temperature Band 31\0",     /* new, 18-MAY-04  */
				      "Coarse Resolution Brightness Temperature Band 32\0",     /* new, 18-MAY-04  */
				      "Coarse Resolution Granule Time\0",    /* 04-AUG-05 */			      
				      COARSEQANAME,  
				      "Coarse Resolution Internal CM\0",	      
				      "Coarse Resolution Atmospheric Optical Depth QA\0",			      
				      "n pixels averaged\0"
				      };
static int32 CRSGRID_Data_Type[]={DFNT_INT16,   DFNT_INT16,   DFNT_INT16,   DFNT_INT16,   
                                  DFNT_INT16,   DFNT_INT16,   DFNT_INT16,  
				  DFNT_INT16,   DFNT_INT16,      
				  DFNT_INT16,   DFNT_INT16,   DFNT_INT16,   DFNT_UINT8, 
				  DFNT_UINT16,  DFNT_UINT16,  DFNT_UINT16,  DFNT_UINT16,  DFNT_INT16,      
				  DFNT_UINT32,  DFNT_UINT16,  DFNT_UINT8,   DFNT_UINT8};

enum{CRSGRDB1,     CRSGRDB2,     CRSGRDB3, 
     CRSGRDB4,     CRSGRDB5,     CRSGRDB6,     CRSGRDB7,    
     CRSGRDTAU1,   CRSGRDTAU3,  
     CRSGRDSZ,     CRSGRDVZ,     CRSGRDRA,     CRSGRDOZ,  
     CRSGRDB20,    CRSGRDB21,    CRSGRDB31,    CRSGRDB32,    CRSGRDGTM,  /* 04-AUG-05 */  
     CRSGRDQA,     CRSGRDICM,    CRSGRDAQA,    CRSGRDN,      CRSGRDSDSS};
     
/* FOR FUTURE REFERENCE: Please note that although SDSs can be added anywhere prior to 
   CRSGRDSDSS and make it in the final product, ONLY SDSs UP TO BUT NOT INCLUDING CRSGRDQA 
   are put through transfer_sds() -- so new SDSs have to be added prior to CRSGRDQA. */
				  
#define PIXPERDEG 20.0000F 

/* 26-APR-06: need this to reduce memory load of IDS algorithm on SGIs */
#ifdef LINUX
  #define DEGPERWIN 10.0000F
#else
  #define DEGPERWIN 1.0000F
#endif

#define UNITSREFLECTANCE "reflectance"
#define REFOFFSET 0.0F
#define UNITSDEGREES "degrees"
#define UNITSTIME "HHMM"     /* 04-AUG-05 */

#define NPOLEFNAMELUN     209010
#define NONPOLEFNAMELUN   209011
#define SPOLEFNAMELUN     209012

#define DIM_ZERO "Data Lines"
#define DIM_ONE "Data Samples Per Line"

#define MAKECMGLUN  209943

/* All these externs refer to global.h */
extern commands_t commands;
extern d_sds **sds_outputlist;
extern int currentlyopen, current_n_CMGwindows,n_ret;
extern int make_this_cmg[3];
extern char CMG_lgid_template[256];
extern float leftlon;
extern input_metadata **CMGmetadata;

enum{NTHPL, MIDDLE, STHPL, NCMGFILES}; 
#define ULLAT 90.0000F
#define LRLAT -90.0000F
#define ULLON -180.0000F
#define LRLON 180.0000F 
#define POLELIM 60.0000F
#define NONPOLEWIDTH 90.0000F
