#include "mfhdf.h"
#include "PGS_TD.h"
#include "PGS_PC.h"
#include "PGS_MET.h"
#include "modis_srefl.h"
#include "struct.h"

#define DAOOZONEVARIABLE "OZONE\0" 
#define DAOSPRESVARIABLE "PS\0"
#define DAOWVVARIABLE    "TPW\0"

/* 16-JUN-99: Grib definitions */
#define OZONE_GRIBTAG    "EPTO3\0"  
#define GOZONE_GRIBTAG   "TOZNE\0"  
/*#define SPRES_GRIBTAG    "PRES\0"*/
#define SPRES_GRIBTAG    "PRMSL\0"
#define WV_GRIBTAG       "PWAT\0"
#define TMP_GRIBTAG      "TMP\0"
#define UWD_GRIBTAG      "UGRD\0"
#define VWD_GRIBTAG      "VGRD\0"
#define OZONE_GRIBLEVEL  "atmos col\0"
/*#define SPRES_GRIBLEVEL  "sfc\0"  */
#define SPRES_GRIBLEVEL  "MSL\0"  
#define WV_GRIBLEVEL     "atmos col\0"
#define TMP_GRIBLEVEL    "2 m above gnd\0"
#define UWD_GRIBLEVEL    "10 m above gnd\0"
#define VWD_GRIBLEVEL    "10 m above gnd\0"

#define GRIB_WV_SP_TIMERES 6.0    /* hours */
#define GRIB_OZONE_TIMERES 24.0   /* hours */
#define EPT_OZONE_TIMERES 24.0    /* hours */

/* will use these values as LUNs directly, not as indices to arrays of LUNs */
#define DAOOZONE OZONEFILE
#define DAOSPRES SURFPRESSURE
#define DAOWV    ANCWVT
#define GRIB_WV  GRIB_ANCWVT
#define GRIB_OZ  GRIB_OZONEFILE
#define GRIB_SP  GRIB_SURFPRESSURE
#define EPT_OZ   EPT_OZONEFILE
#define ULLAT 90.0000F
#define LRLAT -90.0000F
#define ULLON -180.0000F
#define LRLON 180.0000F 
#define POLELIM 60.0000F
#define NONPOLEWIDTH 90.0000F
enum{NTHPL, MIDDLE, STHPL, NCMGFILES}; 

/* 16-JUN-99: generic definitions */
#define OZZONE "OZ\0"
#define GOZZONE "GOZ\0"
#define SURFPR "SP\0"
#define WATERV "WV\0"
#define TEMPR  "TP\0"
#define UWND   "UW\0"
#define VWND   "VW\0"

/*enum{SRL1B1K, SRL1B500, SRL1B250, SRGEOLOC, SRCLOUD, SROZONE, SRWV, LEVTWOFILES, 
       ANCAOT1, ANCSRA, ANCSREFFILE0, ANCSREFFILE1, ANCSREFFILE2, ANCSREFFILE3, ANCSREFFILE4,
       ANCSREFFILE5, ANCSREFFILE6, ANCSREFFILE7, ANCSREFFILE8, 
       ANCSREFFILE9, ANCSREFFILE10, ANCSREFFILE11, ANCSREFFILE12, 
       ANCSREFFILE13, ANCSREFFILE14, ANCSREFFILE15, 
       ANCTAFILE0, ANCTAFILE1, ANCTAFILE2, ANCTAFILE3, ANCTAFILE4,
       ANCTAFILE5, ANCTAFILE6, ANCTAFILE7, ANCTAFILE8, 
       ANCTAFILE9, ANCTAFILE10, ANCTAFILE11, ANCTAFILE12, 
       ANCTAFILE13, ANCTAFILE14, ANCTAFILE15, 
       ANCRHOBARFILE0, ANCRHOBARFILE1, ANCRHOBARFILE2, ANCRHOBARFILE3, ANCRHOBARFILE4,
       ANCRHOBARFILE5, ANCRHOBARFILE6,
       ANCREFLDISORDFILE0, ANCREFLDISORDFILE1, ANCREFLDISORDFILE2, ANCREFLDISORDFILE3, ANCREFLDISORDFILE4,
       ANCREFLDISORDFILE5, ANCREFLDISORDFILE6, ANCRHO2BARFILE,
       BOSTONBRDFFILE0, BOSTONBRDFFILE1, BOSTONBRDFFILE2, BOSTONBRDFFILE3, BOSTONBRDFFILE4,
       BOSTONBRDFFILE5, BOSTONBRDFFILE6, BOSTONBRDFFILE7, BOSTONBRDFFILE8, 
       BOSTONBRDFFILE9, BOSTONBRDFFILE10, BOSTONBRDFFILE11, BOSTONBRDFFILE12, 
       BOSTONBRDFFILE13, BOSTONBRDFFILE14, BOSTONBRDFFILE15, ANCBIOME, 
       ALLFILES};
*/
#ifndef MDEFS_H
#define MDEFS_H


/* ECS Metadata Definitions */

/* this must match up with the inputpointer and systemfilename
   NUM_VALS in the MCF */
#define MAX_INPUT_FILES 450

static char Mfield_name[127][PGSd_PC_VALUE_LENGTH_MAX]=
{
  "INPUTPOINTER",
  "SYSTEMFILENAME",
  "SHORTNAME",
  "RANGEBEGINNINGDATE",
  "RANGEENDINGDATE",
  "ORBITNUMBER.1",
  "DAYNIGHTFLAG",
  "NORTHBOUNDINGCOORDINATE",
  "SOUTHBOUNDINGCOORDINATE",
  "EASTBOUNDINGCOORDINATE",
  "WESTBOUNDINGCOORDINATE",
  "GRINGPOINTSEQUENCENO.1",
  "GRINGPOINTLATITUDE.1",
  "GRINGPOINTLONGITUDE.1",
  "EXCLUSIONGRINGFLAG.1",
  "INSTRUMENTSCANSPROCESSED",
  "REPROCESSINGACTUAL",
  "REPROCESSINGPLANNED",
  "SEAPROCESSED",
  "ADJACENCYCORRECTIONATTEMPTED",
  "ADDITIONALATTRIBUTENAME.1",
  "ADDITIONALATTRIBUTENAME.2",
  "ADDITIONALATTRIBUTENAME.3",
  "ADDITIONALATTRIBUTENAME.4",
  "ADDITIONALATTRIBUTENAME.5",
  "ADDITIONALATTRIBUTENAME.6",
  "ADDITIONALATTRIBUTENAME.7",
  "ADDITIONALATTRIBUTENAME.8",
  "ADDITIONALATTRIBUTENAME.9",
  "ADDITIONALATTRIBUTENAME.10",
  "ADDITIONALATTRIBUTENAME.11",
  "ADDITIONALATTRIBUTENAME.12",
  "ADDITIONALATTRIBUTENAME.13",
  "ADDITIONALATTRIBUTENAME.14",
  "ADDITIONALATTRIBUTENAME.15",
  "ADDITIONALATTRIBUTENAME.16",
  "ADDITIONALATTRIBUTENAME.17",
  "ADDITIONALATTRIBUTENAME.18",
  "ADDITIONALATTRIBUTENAME.19",
  "ADDITIONALATTRIBUTENAME.20",
  "ADDITIONALATTRIBUTENAME.21",
  "BRDFCORRECTIONATTEMPTED",
  "QUALITYCLASSPERCENTAGE250MBAND1",
  "QUALITYCLASSPERCENTAGE250MBAND2",
  "QUALITYCLASSPERCENTAGE500MBAND1",
  "QUALITYCLASSPERCENTAGE500MBAND2",
  "QUALITYCLASSPERCENTAGE500MBAND3",
  "QUALITYCLASSPERCENTAGE500MBAND4",
  "QUALITYCLASSPERCENTAGE500MBAND5",
  "QUALITYCLASSPERCENTAGE500MBAND6",
  "QUALITYCLASSPERCENTAGE500MBAND7",
  "QUALITYCLASSPERCENTAGE1KMBAND1",
  "QUALITYCLASSPERCENTAGE1KMBAND2",
  "QUALITYCLASSPERCENTAGE1KMBAND3",
  "QUALITYCLASSPERCENTAGE1KMBAND4",
  "QUALITYCLASSPERCENTAGE1KMBAND5",
  "QUALITYCLASSPERCENTAGE1KMBAND6",
  "QUALITYCLASSPERCENTAGE1KMBAND7",
  "PARAMETERVALUE.5",
  "PARAMETERVALUE.6",
  "PARAMETERVALUE.7",
  "PARAMETERVALUE.8",
  "PARAMETERVALUE.9",
  "PARAMETERVALUE.10",
  "PARAMETERVALUE.11",
  "PARAMETERVALUE.12",
  "PARAMETERVALUE.13",
  "PARAMETERVALUE.14",
  "PARAMETERVALUE.15",
  "PARAMETERVALUE.16",
  "PARAMETERVALUE.17",
  "PARAMETERVALUE.18",
  "PARAMETERVALUE.19",
  "PARAMETERVALUE.20",
  "PARAMETERVALUE.21",
  "PARAMETERVALUE.1",
  "PARAMETERVALUE.2",
  "PARAMETERVALUE.3",
  "PARAMETERVALUE.4",
  "QAPERCENTOUTOFBOUNDSDATA.1",
  "QAPERCENTMISSINGDATA.1",
  "LOCALGRANULEID",
  "ANCILLARYINPUTPOINTER.1", 
  "ANCILLARYSYSTEMFILENAME",
  "RANGEBEGINNINGTIME",
  "RANGEENDINGTIME",
  "RANGEBEGINNINGDATETIME",
  "RANGEENDINGDATETIME",
  "PARAMETERNAME.1",
  "PGEVERSION",
  "PARAMETERVALUE.1",
  "AUTOMATICQUALITYFLAG.1",
  "ADDITIONALATTRIBUTENAME.5",
  "PARAMETERVALUE.5",
  "PARAMETERVALUE.2",
  "PARAMETERVALUE.3",
  "SCISTATE",
  "SCIABNORM",
  "PERCENTNOTSEA",
  "OPERATIONMODE",
  "OPERATIONMODE.1",
  "PERCENTLAND",
  "ATMOSPHERICCORRECTIONATTEMPTED",
  "EQUATORCROSSINGDATE.1",
  "EQUATORCROSSINGLONGITUDE.1",
  "EQUATORCROSSINGTIME.1",
  "ADDITIONALATTRIBUTENAME.6",
  "PARAMETERVALUE.6",
  "NUMBEROFHIGHSOLARZENITHPIXELS1KM",
  "AUTOMATICQUALITYFLAGEXPLANATION.1",
  "ANCILLARYINPUTTYPE.1",
  "CLOUDPIXELS",
  "WATERPIXELS",
  "UNKNOWNPIXELS",
  "NOFIREPIXELS",
  "GLINTPIXELS",
  "CLOUDADJACENTFIREPIXELS",
  "WATERADJACENTFIREPIXELS",
  "AEROSOLQUANTITYCLMH",
  "ADDITIONALATTRIBUTENAME.22",
  "PARAMETERVALUE.22",
  "ADDITIONALATTRIBUTENAME.7",
  "PARAMETERVALUE.7",
  "INTERNALCLOUDMASK",
  "VERSIONID",
  "PROCESSINGENVIRONMENT",
  "LOCALVERSIONID"
};


/* metadata field name definitions */
enum{INPUTPOINTER 
       , SYSTEMFILENAME 
       , SHORTNAME 
       , RANGEBEGINNINGDATE 
       , RANGEENDINGDATE 
       , ORBITNUMBER 
       , DAYNIGHTFLAG 
       , NORTHBOUNDINGCOORDINATE 
       , SOUTHBOUNDINGCOORDINATE 
       , EASTBOUNDINGCOORDINATE 
       , WESTBOUNDINGCOORDINATE 
       , GRINGPOINTSEQUENCENO 
       , GRINGPOINTLATITUDE 
       , GRINGPOINTLONGITUDE 
       , EXCLUSIONGRINGFLAG 
       , INSTRUMENTSCANSPROCESSED 
       , REPROCESSINGACTUAL 
       , REPROCESSINGPLANNED 
       , SEAPROCESSED 
       , ADJACENCYCORRECTIONATTEMPTED 
       , ATTRIBUTENAMEQAPERCENTGOODQUALITY 
       , ATTRIBUTENAMEQAPERCENTOTHERQUALITY 
       , ATTRIBUTENAMEQAPERCENTNOTPRODUCEDCLOUD 
       , ATTRIBUTENAMEQAPERCENTNOTPRODUCEDOTHER 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT250MBAND1 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT250MBAND2 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND1 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND2 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND3 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND4 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND5 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND6 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND7 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND1 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND2 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND3 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND4 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND5 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND6 
       , ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND7 
       , ATTRIBUTENAMEGRANULENUMBER 
       , BRDFCORRECTIONATTEMPTED 
       , QUALITYCLASSPERCENTAGE250MBAND1 
       , QUALITYCLASSPERCENTAGE250MBAND2 
       , QUALITYCLASSPERCENTAGE500MBAND1 
       , QUALITYCLASSPERCENTAGE500MBAND2 
       , QUALITYCLASSPERCENTAGE500MBAND3 
       , QUALITYCLASSPERCENTAGE500MBAND4 
       , QUALITYCLASSPERCENTAGE500MBAND5 
       , QUALITYCLASSPERCENTAGE500MBAND6 
       , QUALITYCLASSPERCENTAGE500MBAND7 
       , QUALITYCLASSPERCENTAGE1KMBAND1 
       , QUALITYCLASSPERCENTAGE1KMBAND2 
       , QUALITYCLASSPERCENTAGE1KMBAND3 
       , QUALITYCLASSPERCENTAGE1KMBAND4 
       , QUALITYCLASSPERCENTAGE1KMBAND5 
       , QUALITYCLASSPERCENTAGE1KMBAND6 
       , QUALITYCLASSPERCENTAGE1KMBAND7 
       , QAPERCENTPOOROUTPUT250MBAND1 
       , QAPERCENTPOOROUTPUT250MBAND2 
       , QAPERCENTPOOROUTPUT500MBAND1 
       , QAPERCENTPOOROUTPUT500MBAND2 
       , QAPERCENTPOOROUTPUT500MBAND3 
       , QAPERCENTPOOROUTPUT500MBAND4 
       , QAPERCENTPOOROUTPUT500MBAND5 
       , QAPERCENTPOOROUTPUT500MBAND6
       , QAPERCENTPOOROUTPUT500MBAND7 
       , QAPERCENTPOOROUTPUT1KMBAND1 
       , QAPERCENTPOOROUTPUT1KMBAND2 
       , QAPERCENTPOOROUTPUT1KMBAND3 
       , QAPERCENTPOOROUTPUT1KMBAND4 
       , QAPERCENTPOOROUTPUT1KMBAND5 
       , QAPERCENTPOOROUTPUT1KMBAND6 
       , QAPERCENTPOOROUTPUT1KMBAND7 
       , GRANULENUMBER 
       , QAPERCENTGOODQUALITY 
       , QAPERCENTOTHER 
       , QAPERCENTNOTPRODUCEDCLOUD 
       , QAPERCENTNOTPRODUCEDOTHER 
       , QAPERCENTOUTOFBOUNDS 
       , QAPERCENTMISSING 
       , LOCALGRANULEID 
       , ANCILLARYINPUTPOINTER 
       , ANCILLARYSYSTEMFILENAME 
       , RANGEBEGINNINGTIME 
       , RANGEENDINGTIME 
       , RANGEBEGINNINGDATETIME 
       , RANGEENDINGDATETIME 
       , PARAMETERNAME 
       , PGEVERSION 
       , INPUTGRANULENUMBER 
       , AUTOMATICQUALITYFLAG 
       , ATTRIBUTENAMEFIREPIXELS 
       , FIREPIXELS
       , SCI_STATE 
       , SCI_ABNORM 
       , SCI_STATE_OUT 
       , SCI_ABNORM_OUT 
       , PERCENTNOTSEA 
       , OPERATIONMODE 
       , OPERATIONMODEOUT 
       , PERCENTLAND 
       , ATMOSPHERICCORRECTIONATTEMPTED 
       , EQUATORCROSSINGDATE 
       , EQUATORCROSSINGLONGITUDE 
       , EQUATORCROSSINGTIME 
       , ATTRIBUTENAMEGRANULENUMBERFIRE 
       , GRANULENUMBERFIRE 
       , NUMBEROFHIGHSOLARZENITHPIXELS
       , AUTOMATICQUALITYFLAGEXPLANATION
       , ANCILLARYINPUTTYPE
       , CLOUDPIXELS
       , NUMWATERPIXELS
       , NUMUNKNOWNPIXELS
       , NUMNOFIREPIXELS
       , NUMGLINTPIXELS
       , NUMCLOUDADJACENTFIREPIXELS
       , NUMWATERADJACENTFIREPIXELS
       , AEROSOLQUANTITY
       , ATTRIBUTENAMEPROCESSVERSION
       , PROCESSVERSION
       , ATTRIBUTENAMEPROCESSVERSIONFIRE
       , PROCESSVERSIONFIRE
       , USEDINTCLOUDMASK
       , VERSIONID
       , PROCESSINGENVIRONMENT
       , LOCALVERSIONID
};

/* metadata structure defaults */
#define PROCESSINGDATETIMEDEFAULT "PROCESSINGDATETIME not calculated"
#define RANGEBEGINNINGDATETIMEDEFAULT "RANGEBEGINNINGDATETIME not read from input file"
#define RANGEENDINGDATETIMEDEFAULT "RANGEENDINGDATETIME not read from input file"
#define ORBITNUMBERDEFAULT 0
#define DAYNIGHTFLAGDEFAULT "DAYNIGHTFLAG undefined"
#define COORDINATEDEFAULT 361.0F
#define GRINGPOINTSEQUENCENODEFAULT 0
#define EXCLUSIONGRINGFLAGDEFAULT "EXCLUSIONGRINGFLAGDEFAULT not read from input file"
#define INSTRUMENTSCANSPROCESSEDDEFAULT -1
#define REPROCESSINGACTUALDEFAULT "REPROCESSINGACTUAL metadata field was not set"
#define REPROCESSINGPLANNEDDEFAULT "REPROCESSINGPLANNED metadata field was not set"
#define SEAPROCESSEDDEFAULT "SEAPROCESSED metadata field was not set"
#define ADJACENCYCORRECTIONATTEMPTEDDEFAULT "ADJACENCYCORRECTIONATTEMPTED metadata field was not set"
#define ATMOSPHERICCORRECTIONATTEMPTEDDEFAULT "ATMOSPHERICCORRECTIONATTEMPTED metadata field was not set"
#define ATTRIBUTENAMEQAPERCENTGOODQUALITYDEFAULT "QAPERCENTGOODQUALITY"
#define ATTRIBUTENAMEQAPERCENTOTHERQUALITYDEFAULT "QAPERCENTOTHERQUALITY"
#define ATTRIBUTENAMEQAPERCENTNOTPRODUCEDCLOUDDEFAULT "QAPERCENTNOTPRODUCEDCLOUD"
#define ATTRIBUTENAMEQAPERCENTNOTPRODUCEDOTHERDEFAULT "QAPERCENTNOTPRODUCEDOTHER"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT250MBAND1DEFAULT "QAPERCENTPOOROUTPUT250MBAND1"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT250MBAND2DEFAULT "QAPERCENTPOOROUTPUT250MBAND2"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND1DEFAULT "QAPERCENTPOOROUTPUT500MBAND1"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND2DEFAULT "QAPERCENTPOOROUTPUT500MBAND2"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND3DEFAULT "QAPERCENTPOOROUTPUT500MBAND3"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND4DEFAULT "QAPERCENTPOOROUTPUT500MBAND4"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND5DEFAULT "QAPERCENTPOOROUTPUT500MBAND5"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND6DEFAULT "QAPERCENTPOOROUTPUT500MBAND6"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT500MBAND7DEFAULT "QAPERCENTPOOROUTPUT500MBAND7"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND1DEFAULT "QAPERCENTPOOROUTPUT1KMBAND1"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND2DEFAULT "QAPERCENTPOOROUTPUT1KMBAND2"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND3DEFAULT "QAPERCENTPOOROUTPUT1KMBAND3"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND4DEFAULT "QAPERCENTPOOROUTPUT1KMBAND4"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND5DEFAULT "QAPERCENTPOOROUTPUT1KMBAND5"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND6DEFAULT "QAPERCENTPOOROUTPUT1KMBAND6"
#define ATTRIBUTENAMEQAPERCENTPOOROUTPUT1KMBAND7DEFAULT "QAPERCENTPOOROUTPUT1KMBAND7"
#define ATTRIBUTENAMEGRANULENUMBERDEFAULT "GRANULENUMBER"
#define BRDFCORRECTIONATTEMPTEDDEFAULT "no"
#define TAIDEFAULT -1.0
#define QUALITYCLASSPERCENTAGEDEFAULT -1
#define MODLANDQADEFAULT -1
#define ECSMANDATORYQADEFAULT -1
#define LOCALGRANULEIDVERSIONDEFAULT "000"
#define LOCALGRANULEIDDEFAULT "LOCALGRANULEID metadata field was not set"
#define PGEVERSIONDEFAULT "PGEVERSION metadata field was not set"
#define AUTOQUALITYDEFAULT "AUTOMATICQUALITYFLAG was not set"
#define FIREPIXELSDEFAULT 0
#define ATTRIBUTENAMEFIREPIXELSDEFAULT "FIREPIXELS"
#define SCI_STATE_DEFAULT -1
#define SCI_ABNORM_DEFAULT -1
#define PERCENTNOTSEADEFAULT -1
#define OPERATIONMODEDEFAULT "OPERATIONMODE was not set"
#define PERCENTLANDDEFAULT -1
#define GRANULENUMBERDEFAULT -1
#define EQUATORCROSSINGDATEDEFAULT "EQUATORCROSSINGDATE not read from input file"
#define EQUATORCROSSINGLONGITUDEDEFAULT -1.0
#define EQUATORCROSSINGTIMEDEFAULT "EQUATORCROSSINGTIME not read from input file"
/*
 * #define AUTOMATICQUALITYFLAGEXPLANATIONREFDEFAULT "'Failed' if there is no valid land data contained within; 'Suspect' if the majority of pixels are at a high solar zenith angle."
*/ 
#define AUTOMATICQUALITYFLAGEXPLANATIONREFDEFAULT "Always 'Passed'."
#define AUTOMATICQUALITYFLAGEXPLANATIONFIREDEFAULT "'Suspect' if coarse resolution geolocation was used."
#define ANCILLARYINPUTTYPEDEFAULT "Geolocation"
#define PARAMETERNAMESRDEFAULT "MODIS Atmospherically Corrected Surface Reflectance"  /* KSTai, 2/26/1999*/
#define PARAMETERNAMECSRDEFAULT "MOD09CRS Coarse Atmo Corrected Surf Refl"
#define PARAMETERNAMEFIREDEFAULT "MODIS L2 Active Fire Detection"
#define PROCESSVERSIONDEFAULT "PROCESSVERSION not set"
#define ADDITIONALATTRIBUTENAMEPROCESSVERSIONDEFAULT "PROCESSVERSION"

/* static definitions */
#define CORES "CoreMetadata.0"
#define ARCHIVES "ArchiveMetadata.0"
#define METAHANDLES mdhandles
#define COREHANDLES mdhandles[1]
#define ARCHIVEHANDLES mdhandles[2]

#define NOVERS 1        /* same reason as above */

#define DAYNIGHTFLAGDAY "Day"
#define DAYNIGHTFLAGNIGHT "Night"
#define DAYNIGHTFLAGMIXED "Mixed"
#define DAYNIGHTFLAGOTHER "Other"
#define DAYNIGHTFLAGBOTH "Both" 
#define GRINGPOINTS 4
#define YES "yes"
#define NO "no"

/* logical number definitions for things that will be read from the PCF */
/* LUNs for the following two attributes have been changed as requested 05/01/01 Hong Xiang 
#define REPROCESSINGACTUALLUN 209700
#define REPROCESSINGPLANNEDLUN 209701  */
#define REPROCESSINGACTUALLUN 800504 
#define REPROCESSINGPLANNEDLUN 800503
#define LOCALGRANULEIDVERSIONLUN 209702
#define PGEVERSIONNUMBERLUN 800500     /* 09-FEB-00 */
#define SATELLITEINSTRUMENTLUN 800510     /* 03-AUG-00 */
#define LOCALGRANULEIDVERSIONFIRELUN 209704

/* global string buffers used for writing metadata */
#define FNAMEBUFFERS 2
#define BUILDBUFFER 0
#define OLDBUFFER 1

/* All these externs refer to global.h */
extern d_sds **sds_outputlist;
extern int currentlyopen, current_n_CMGwindows,n_ret;
extern int make_this_cmg[3];
extern char CMG_lgid_template[256];
extern float leftlon;
extern input_metadata **CMGmetadata;
extern commands_t commands;
extern int n_bands_res[3]; /* 250m, 500m, 1km */
extern char REF_ESDTname[10];
extern char CRF_ESDTname[10];
extern char VersionID[10];
extern char *ProcessingEnvironment, *Platform; 
extern input_metadata *metadata;
extern PGSt_MET_all_handles METAHANDLES; /* ECS metadata handles */



#endif

