#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "modis_srefl.h"
#include "protos.h" 
#include "param.h" 
#include "global.h" 
#include "pol_global.h" 
#include "versions.h"
int main(void)
/*
!C****************************************************************************

!Description: Redesign of PGE11's MOD_PR09 code.

!Input Parameters:
  none
 
!Input/Output Parameters:
  none

!Output Parameters:
  none

!Returns:
  Success: 0
  Failure: nonzero
  
!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
     Scale-up of MOD09C's functionality (06-FEB-03).
     
!END**************************************************************************/
{  /* BEGINNING OF MAIN */
dims_struct ***inputs;  /* A pointer to a two-dimensional array of dims_struct objects */
dims_struct ***outputs;  /* A pointer to a two-dimensional array of dims_struct objects */
int i, max, n_inputs[ALL_L1B_FILES], input_luns[ALL_L1B_FILES];
int retval = 0;
char location[256] = "modis_srefl.c, main\0";
extern int do_climatology;

/* Initializations */
retval |= init_processing(&max, n_inputs, input_luns);

/* set up array of structures holding information about inputs */
if (retval == 0)
   retval |= setup_inputs(&inputs, n_inputs, input_luns);

/* sort the array of structures holding information about inputs into temporal order */
if (retval == 0)
   retval |= sort_inputs(&inputs, n_inputs);

/* reset max to the actual number of time-points (not simply max of a type of file in PCF!) */
max = n_inputs[0];

/* set up array of structures holding information about outputs */
if (retval == 0)
   get_input_str_2D_memory(&outputs, 2, max);

/* report to STDIO for now * /
report(inputs, max); */

/* Do the first pass; atmospheric correction and generation of AOT in the daytime outputs. */
if (retval == 0)
   retval |= atmospheric_correction(1, inputs, outputs, max);
   
/*for(i=0;i<max;i++) printf("*** %s %s\n", outputs[0][i]->filename, outputs[1][i]->filename);*/

/* 06-JUN-06: Get cloud adjacency info into QAs */
cloud_adjacency(max, outputs); 

/* 31-MAR-06: Interpolate the TR/PR/SA arrays */
if ( tr_interpolation(max, outputs) )  modsmf(MODIS_E_AOT_INTERPOLATION, " ", location); 

if (0==1) {   /* 31-MAR-06: Don't do this for the new TR/PR/SA version of code */
/**********  NO  ***************************************************************/
/* Filter and interpolate the AOT */
if (do_climatology != 1) {
   aot_filter(max, outputs);
   if ( aot_interpolation(max, outputs) )  modsmf(MODIS_E_AOT_INTERPOLATION, " ", location);
   }
   
/* Do the second pass. */
/* As of 04-FEB-05: the second pass will now redo the AOT and associated QA masks, and rewrite AOT/etc to outputs.... */
if (retval == 0)
   retval |= atmospheric_correction(0, inputs, outputs, max);

/* Interpolate the AOT again */  
if (do_climatology != 1) {
   if ( aot_interpolation(max, outputs) )  modsmf(MODIS_E_AOT_INTERPOLATION, " ", location);
   }
/**********  NO  ***************************************************************/
   }   /* 31-MAR-06: Don't do this for the new TR/PR/SA version of code */

/* Now do a third pass with the latest AOT. */
if (retval == 0)
   retval |= atmospheric_correction(2, inputs, outputs, max);

/* 06-JUN-06: Get cloud adjacency info into QAs, again... */
cloud_adjacency(max, outputs); 

/* Make cloud shadows */
if (commands.make_cloud_shadow)
   if ( cld_shadow_mask(max, inputs, outputs) ) modsmf(MODIS_E_CLOUD_SHADOW, " ", location);  

/* Clean up structures... */
close_io_struct(inputs, ALL_L1B_FILES, max);
close_io_struct(outputs, 2, max);

/* ...and the rest */
clean_up_ancillary_data();
clean_up();

/* leave */
return(retval);
}  /* END OF MAIN*/

void clean_up(void)
/*
!C****************************************************************************

!Routine: clean_up()

!Description: Frees memory

!Input Parameters:
  none
 
!Input/Output Parameters:
  none

!Output Parameters:
  none

!Returns:
  none
  
!Revision History:
  Revision 1.0  24-MAR-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
  
!END**************************************************************************/
{
int i;
char msg[256];
extern float **Rho_ra_term, **Rho_r_term, **Step1_term, **Step2_term, **Step3_term, **SRA_term, **btau;
extern float **hkmRho_ra_term, **hkmRho_r_term, **hkmStep1_term, **hkmStep2_term, **hkmStep3_term, **hkmSRA_term;
extern float **qkmRho_ra_term, **qkmRho_r_term, **qkmStep1_term, **qkmStep2_term, **qkmStep3_term, **qkmSRA_term;
/*extern int n_bands_res[3]; /  * 250m, 500m, 1km */


if (aotr_struct.aotr_offsets != NULL) free(aotr_struct.aotr_offsets);
if (aotr_struct.aotr_lengths != NULL) free(aotr_struct.aotr_lengths);

if ((commands.makegcm == 1)&&(strlen(CMG_lgid_template) > 0)) {   /* 23-MAR-04: (strlen(CMG_lgid_template) > 0) added */

      for(i=0;i<NCMGFILES;i++) {
          if(PGS_MET_Init((MCF_IDN+i), METAHANDLES)!=PGS_S_SUCCESS) 
	      modsmf((PGSt_SMF_code)MODIS_E_SDP_ERROR, "PGS_MET_Init", "clean_up()"); 

          adapt_metadata_for_CMG(i, CMGmetadata[i], metadata, CMG_lgid_template);
	  
          if (make_this_cmg[i]) {
             set_ECS_metadata(IDSPROD, CMGmetadata[i]);
             PGS_MET_Write(COREHANDLES, "CoreMetadata.0", sds_outputlist[0]->sdid[i]);
             PGS_MET_Write(ARCHIVEHANDLES, "ArchiveMetadata.0", sds_outputlist[0]->sdid[i]);
	      }
	  
          PGS_MET_Remove();
	  
          freeECSmetadata(CMGmetadata[i]); 
          }
      free(CMGmetadata); 

     end_coarse_grid(sds_outputlist, currentlyopen, current_n_CMGwindows);
     
     if (n_ret > 100) {   /* 100 because there is a certain amount of slop in the mapping; I want this message to indicate
                             significant problems */
       sprintf(msg, "Possible bad mapping of data to nonpolar IDN file, in routine update_coarse_grid(); %d lines were bad", n_ret);
       modsmf((PGSt_SMF_code)MODIS_W_FUNCTION_ERROR, msg, "clean_up()"); 
     }
  }
  
/* 04-APR-01: for polarization correction */
if(commands.pol_corr == 1) {

   free_polarization_data(&aoi);
   free_polarization_data(&am12_0);
   free_polarization_data(&am13_0);
   if (am12_1.data != NULL) free_polarization_data(&am12_1);
   if (am13_1.data != NULL) free_polarization_data(&am13_1);
   free_polarization_data(&sun);
   free_polarization_data(&ang);
   free_polarization_data(&sigma);
   free_polarization_data(&eye);
   free_polarization_data(&cue);
   free_polarization_data(&you);
   
  }

if (ProcessingEnvironment != (char *) NULL) free(ProcessingEnvironment);
if (metadata != NULL) freeECSmetadata(metadata);  

/* 24-APR-06 
for(i=0;i<7;i++){
   free((void *) sref[i]);
   free((void *) ta_tab[i]);
  }
*/
}

void report(dims_struct ***inputs, int n_timepoints)
/*
!C****************************************************************************

!Routine: report()

!Description: Writes sequenced input information to STDIO

!Input Parameters:
  dims_struct ***inputs       the time-sequenced inputs
  int n_timepoints            the number of timepoints
 
!Input/Output Parameters:
  none

!Output Parameters:
  none

!Returns:
  none
  
!Revision History:
  Revision 1.0  06-FEB-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
     Meant primarily for debugging and code development.
  
!END**************************************************************************/
{
char x[23];
int i,j,ii;

for(j=0;j<n_timepoints;j++)  {        /* for each time point */
   printf("time %d : ", j);
   for(i=0;i<ALL_L1B_FILES;i++)  {   /* for each L1B input */
      printf(" %d:%d ", i,inputs[i][j]->OK);
      if (inputs[i][j]->OK == 0) {
         for(ii=0;ii<22;ii++) x[ii] = inputs[i][j]->xxxxx->localgranuleid[ii];
         x[ii] = '\0';
 	 }
      else {
         for(ii=0;ii<21;ii++) x[ii] = ' ';
         x[ii] = '\0';
 	 }
      printf("%s ", x);
      }
    printf("\n");
  }
}


int init_processing(int *max, int *n_in, int *input_luns)
/*
!C****************************************************************************

!Routine: init_processing()

!Description: Whole-process initializations.

!Input Parameters:
  none 
 
!Input/Output Parameters:
  none

!Output Parameters:
  int *input_luns          Input file LUNs and  
  int *n_in                ... the number of input files at each LUN
  int max                  ... the maximum number of input files at each LUN (which should approximate 
                                the number of timepoints being covered)

!Returns:
  Success: 0
  Failure: nonzero
  
!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
  
!END**************************************************************************/
{
char pcf_file[256];
int i, j, n_inputs, local_max, ichannel;
PGSt_integer n;
char msg[1000];
char location[256] = "modis_srefl.c, init_processing\0";
int bandindex;
short TESTLINE, TESTPIX;
char report[2000] = {'\0'};
char version_pge[10] = {'\0'};
PGSt_PC_Logical tafile[7]   = {TA_TABLE1, TA_TABLE2, TA_TABLE3, TA_TABLE4, TA_TABLE5, TA_TABLE6, TA_TABLE7};
PGSt_PC_Logical sreffile[7] = {FOSTAB1, FOSTAB2, FOSTAB3, FOSTAB4, FOSTAB5, FOSTAB6, FOSTAB7};
extern int32 locator500[], locator500_250[];

char sbandname[NBANDS][6] = {"modb1 ","modb2 ","modb3 ","modb4 ","modb5 "
      ,"modb6 ","modb7 ","modb8 ","modb9 ","modb10","modb11","modb12"
      ,"modb13","modb14","modb15","modb16"};  /* kept just in case... */

int taurayln = TAURAY_LUTLUN;      /* LUNs for new LUTs  */
int gscoefln = GASCOEF_LUTLUN;
int anglehdfln = ANGLES_LUTLUN;
int intrefln, transmln, spheraln;

char CAMOD[80];
char err_msg[800];
char tauraynm[256];     /* File names for new LUTs    */
char gscoefnm[256];
char anglehdf[256];
char intrefnm[256];      
char transmnm[256];
char spheranm[256];
int ver;
int strlencamod, retval;
extern float32 tauray[NBANDS];
extern float32 oztransa[NBANDS];
extern float32 wvtransa[NBANDS],wvtransb[NBANDS],wvtransc[NBANDS];
extern float32 ogtransa0[NBANDS],ogtransa1[NBANDS];
extern float32 ogtransb0[NBANDS],ogtransb1[NBANDS];
extern float32 ogtransc0[NBANDS],ogtransc1[NBANDS];
extern float32 nbfi[NTS][NXX],tsmax[NTS][NXX],tsmin[NTS][NXX];
extern float32 ttv[NTS][NXX],nbfic[NTS][NXX];
extern float32 tts[NTS];
extern int indts[NTS];

extern float32 rolutt[NBANDS][NPR][NAOT][8000];
extern float32 transt[NBANDS][NPR][NAOT][NTS];  /* just for now */
extern float32 sphalbt[NBANDS][NPR][NAOT];  

extern float32 roluttSMKL[NBANDS][NPR][NAOT][8000];
extern float32 transtSMKL[NBANDS][NPR][NAOT][NTS];
extern float32 sphalbtSMKL[NBANDS][NPR][NAOT];  
extern float32 roluttSMKH[NBANDS][NPR][NAOT][8000];
extern float32 transtSMKH[NBANDS][NPR][NAOT][NTS];
extern float32 sphalbtSMKH[NBANDS][NPR][NAOT];   
extern float32 roluttDUST[NBANDS][NPR][NAOT][8000];
extern float32 transtDUST[NBANDS][NPR][NAOT][NTS];
extern float32 sphalbtDUST[NBANDS][NPR][NAOT];  
extern float32 roluttURBPOL[NBANDS][NPR][NAOT][8000];
extern float32 transtURBPOL[NBANDS][NPR][NAOT][NTS];
extern float32 sphalbtURBPOL[NBANDS][NPR][NAOT];  
extern float32 roluttURBCLN[NBANDS][NPR][NAOT][8000];
extern float32 transtURBCLN[NBANDS][NPR][NAOT][NTS];
extern float32 sphalbtURBCLN[NBANDS][NPR][NAOT];  

extern int sat_type;


extern float **Rho_ra_term, **Rho_r_term, **Step1_term, **Step2_term, **Step3_term, **SRA_term, **btau;
extern float **hkmRho_ra_term, **hkmRho_r_term, **hkmStep1_term, **hkmStep2_term, **hkmStep3_term, **hkmSRA_term;
extern float **qkmRho_ra_term, **qkmRho_r_term, **qkmStep1_term, **qkmStep2_term, **qkmStep3_term, **qkmSRA_term;
/*extern int n_bands_res[3]; /  * 250m, 500m, 1km */

int32 index1km, index500m, index250m;
int32 sc1km[(NBSCANS_1KM*NBSPOTS_1KM)],sp1km[(NBSCANS_1KM*NBSPOTS_1KM)];
int32 sc500m[(NBSCANS_500M*NBSPOTS_500M)];
int32 scan500m, spot500m;
int32 scan250m, spot250m;
int32 k500=0, k250=0, g250=0;

leftlon = 999.0;     /* 07-APR-05: initialize this variable!  This variable had apparently
                        been initialized to zero and had been used as such, incorrectly. */

metadata = (input_metadata *)malloc(sizeof(input_metadata));
initECSmetadata(metadata);  /* this is global */


      for (index1km = 0; index1km < (NBSCANS_1KM*NBSPOTS_1KM); index1km++){
	/* store the spot and scan master indices */
	sp1km[index1km] = index1km % NBSPOTS_1KM;
	sc1km[index1km] = index1km / NBSPOTS_1KM;
	
        for(i=0;i<4;i++){
	    sc500m[index1km*4+i] = sc1km[index1km]*2 +(int32)(i/2);
	  }	
	
	/* set up locator arrays */
	for(i=0;i<2;i++){
	    for(j=0;j<2;j++){
	      spot500m=sp1km[index1km]*2+j;
	      locator500[k500]=sc500m[k500]*NBSPOTS_500M+spot500m;
	      k500++;
	    }
	  }
         
	}
	
        index250m = 0;
        for (i = 0; i < NBSCANS_500M;i++) {
           for (j = 0; j < NBSPOTS_500M; j++){
              locator500_250[index250m++]=(i*NBSPOTS_500M*4) + (2*j);
              locator500_250[index250m++]=(i*NBSPOTS_500M*4) + (2*j) + 1;
              locator500_250[index250m++]=((i*NBSPOTS_500M*4) + (2*j))+(NBSPOTS_500M*2);
              locator500_250[index250m++]=((i*NBSPOTS_500M*4) + (2*j))+(NBSPOTS_500M*2) +1;
           }
        }

/* these are global */
ALPHAF   = DEF_ALPHAF;
BETAF    = DEF_BETAF;
GAMMAF   = DEF_GAMMAF;
DELTAF   = DEF_DELTAF;
MIRAFTOG = DEF_MIRAFTOG;
MIRAFTIG = DEF_MIRAFTIG;
TATOG    = DEF_TATOG;
TATIG    = DEF_TATIG;
ROGT     = DEF_ROGT;

#ifdef SUBSET
T2M_value = DEF_T2M;
NCEPH2O_value = DEF_NCEPH2O;
#endif

/* as are these */
Platform = (char *) malloc(12);
ProcessingEnvironment = (char *) malloc(200);
for(i=0;i<BANDPROCESS;i++)process_bands[i]=0;

for(i=0;i<3;i++)make_this_cmg[i]=0;
currentlyopen = current_n_CMGwindows = n_ret = 0;
strcpy(CMG_lgid_template, "\0");

/* initialize tags for the ancillary data to generic values; will
   be made more specific later. */
strcpy(daoozone.tag, OZZONE);
strcpy(gdasozone.tag, GOZZONE);
strcpy(daospres.tag, SURFPR);
strcpy(daowv.tag,    WATERV);
strcpy(daotmp.tag,   TEMPR);
strcpy(daouwnd.tag,  UWND);
strcpy(daovwnd.tag,  VWND);
daoozone.data = NULL;
gdasozone.data = NULL;
daospres.data = NULL;
daowv.data    = NULL;
daotmp.data   = NULL;
daouwnd.data  = NULL;
daovwnd.data  = NULL;
daoozone.wasread = 1;
gdasozone.wasread = 1;
daospres.wasread = 1;
daowv.wasread    = 1;
daotmp.wasread   = 1;
daouwnd.wasread  = 1;
daovwnd.wasread  = 1;


/* put version of PGE, name of PCF file into LogStatus file */
if(PGS_PC_GetConfigData(PGEVERSIONNUMBERLUN, version_pge)!=PGS_S_SUCCESS){
    sprintf(msg, "LUN: %d", PGEVERSIONNUMBERLUN);
    modsmf((PGSt_SMF_code)MODIS_E_GETCONFIG_FAIL, msg, location);
	     }
else modsmf(MODIS_A_PGE_VERSION, version_pge, location);
strcpy(pcf_file, getenv("PGS_PC_INFO_FILE\0"));
modsmf(MODIS_A_PCF_FILE, pcf_file, location);

sprintf(report, "Running process version number %s, on a %s platform\n", PROCESSVERSIONNUMBER, PLATFORM);
sprintf(report+strlen(report), "\nUsing the following run-time parameters from PCF file '%s':\n\n", pcf_file);
get_params(&commands, &TESTLINE, &TESTPIX, 1, &scans_per_gran,
           &orbit_start_tai, &orbit_end_tai, process_bands, report);

PGS_SMF_GenerateStatusReport(report);
 
sprintf(report, "\nUsing the following internal fire mask values:\n");
sprintf(report+strlen(report), "\n\tALPHAF = %f ", ALPHAF);
if (ALPHAF==DEF_ALPHAF)sprintf(report+strlen(report), "(default)");
sprintf(report+strlen(report), "\n\tBETAF = %f ", BETAF);
if (BETAF==DEF_BETAF)sprintf(report+strlen(report), "(default)");
sprintf(report+strlen(report), "\n\tGAMMAF = %f ", GAMMAF);
if (GAMMAF==DEF_GAMMAF)sprintf(report+strlen(report), "(default)");
sprintf(report+strlen(report), "\n\tDELTAF = %f ", DELTAF);
if (DELTAF==DEF_DELTAF)sprintf(report+strlen(report), "(default)");
sprintf(report+strlen(report), "\n\tMIRAFTOG = %f ", MIRAFTOG);
if (MIRAFTOG==DEF_MIRAFTOG)sprintf(report+strlen(report), "(default)");
sprintf(report+strlen(report), "\n\tMIRAFTIG = %f ", MIRAFTIG);
if (MIRAFTIG==DEF_MIRAFTIG)sprintf(report+strlen(report), "(default)");
sprintf(report+strlen(report), "\n\tTATOG = %f ", TATOG);
if (TATOG==DEF_TATOG)sprintf(report+strlen(report), "(default)");
sprintf(report+strlen(report), "\n\tTATIG = %f ", TATIG);
if (TATIG==DEF_TATIG)sprintf(report+strlen(report), "(default)");
sprintf(report+strlen(report), "\n\tROGT = %f ", ROGT);
if (ROGT==DEF_ROGT)sprintf(report+strlen(report), "(default)");

PGS_SMF_GenerateStatusReport(report);

/* 06-JUN-01: adjust processing flags and commands to be internally consistent */
set_processing_flags(&commands, process_bands);

/* Set the global ESDT variables */
if (sat_type == 0) sprintf(prefix, "MOD\0");  /* sat_type is global */
if (sat_type == 1) sprintf(prefix, "MYD\0");
we_have_ESDTname = get_ESDT_names(sat_type, REF_ESDTname, CRF_ESDTname, VersionID);

if (we_have_ESDTname != 0) { /* It failed, make your own */
    sprintf(REF_ESDTname, "%s09\0", prefix);
    sprintf(CRF_ESDTname, "%s09CRS\0", prefix);
    }
      
/* Get memory for atmospheric correction variables; get it ONCE, and overwrite that memory repeatedly. */
for (i=0;i<3;i++) n_bands_res[i] = 0;
bandindex = 0;
  
for (i=0;i<BAND8;i++) {
    if ((i==BAND1_500M)||(i==BAND1_1KM))bandindex++;
    if (process_bands[i] == 1) n_bands_res[bandindex]++;
    }

  
/* set up array of LUNs for inputs */
input_luns[RAD1_FILE] = RAD1;
input_luns[RAD500_FILE] = RAD500;
input_luns[RAD250_FILE] = RAD250;
input_luns[GEO_FILE] = GEO;
input_luns[CLD_FILE] = CLD;

/* Set up flags for processing bands */
/* All reflectance bands 1-7 at all resolutions will be: DAY_ONLY and FCBOTH */
/*for(i=0;i<BAND8;i++) {*/
/* Change for version 5.1.0 */
for(i=0;i<BAND18;i++) {
   output_type[i].DN = DAY_ONLY;
   output_type[i].FC = FCBOTH;
   output_type[i].type = DFNT_INT16;
   output_type[i].lowrange  = (int16)-100;
   output_type[i].hirange   = (int16)16000;
   output_type[i].fillvalue = (int16)DATAFILLVALUE;
   output_type[i].scale = (float64)REFSCALE;
   strcpy(output_type[i].units, "reflectance\0");
   }

/*
i = BAND8;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_INT16;
output_type[i].lowrange  = (int16)-100;
output_type[i].hirange   = (int16)16000;
output_type[i].fillvalue = (int16)DATAFILLVALUE;
output_type[i].scale = (float64)REFSCALE;
strcpy(output_type[i].units, "reflectance\0");
*/

output_type[BAND18].DN = DNNEITHER;
output_type[BAND18].FC = FCNEITHER;
output_type[BAND18].type = DFNT_UINT16;

output_type[BAND19].DN = DNNEITHER;
output_type[BAND19].FC = FCNEITHER;
output_type[BAND19].type = DFNT_UINT16;

i = BAND20;
output_type[i].DN = DNBOTH;
output_type[i].FC = FCBOTH;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0;
output_type[i].hirange   = (int16)33300;
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)100.0;
strcpy(output_type[i].units, "degrees K\0");

i = BAND21;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0;
output_type[i].hirange   = (int16)42900;
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)100.0;
strcpy(output_type[i].units, "degrees K\0");

i = BAND22;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0;
output_type[i].hirange   = (int16)32900;
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)100.0;
strcpy(output_type[i].units, "degrees K\0");

i = BAND23;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0;
output_type[i].hirange   = (int16)32900;
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)100.0;
strcpy(output_type[i].units, "degrees K\0");

i = BAND26;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FCBOTH;
output_type[i].type = DFNT_INT16;
output_type[i].lowrange  = (int16)-100;
output_type[i].hirange   = (int16)16000;
output_type[i].fillvalue = (int16)DATAFILLVALUE;
output_type[i].scale = (float64)REFSCALE;
strcpy(output_type[i].units, "reflectance\0");

i = BAND29;
output_type[i].DN = DNBOTH;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0;
output_type[i].hirange   = (int16)33300;
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)100.0;
strcpy(output_type[i].units, "degrees K\0");

i = BAND31;
output_type[i].DN = DNBOTH;
output_type[i].FC = FCBOTH;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0;
output_type[i].hirange   = (int16)37800;
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)100.0;
strcpy(output_type[i].units, "degrees K\0");

i = BAND32;
output_type[i].DN = DNBOTH;
output_type[i].FC = FCBOTH;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0;
output_type[i].hirange   = (int16)41900;
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)100.0;
strcpy(output_type[i].units, "degrees K\0");

i = BAND20ALB;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FCBOTH;
output_type[i].type = DFNT_INT16;   
output_type[i].lowrange  = (int16)0;      /*  ??  */
output_type[i].hirange   = (int16)5000;   /*  ??  */
output_type[i].fillvalue = (int16)DATAFILLVALUE;
output_type[i].scale = (float64)REFSCALE;
strcpy(output_type[i].units, "reflectance\0");

i = AOT_1;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FCBOTH;
output_type[i].type = DFNT_INT16;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)5000; 
output_type[i].fillvalue = (int16)AOT1_FILL;
output_type[i].scale = (float64)100000.0;
strcpy(output_type[i].units, "unitless\0");

i = AOT_3;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FCBOTH;
output_type[i].type = DFNT_INT16;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)5000; 
output_type[i].fillvalue = (int16)AOT3_FILL;
output_type[i].scale = (float64)1000.0;
strcpy(output_type[i].units, "unitless\0");

i = AOT_8;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FCBOTH;  /* new, 08-SEP-06 */
output_type[i].type = DFNT_INT16;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)5000; 
output_type[i].fillvalue = (int16)AOT8_FILL;
output_type[i].scale = (float64)1000.0;
strcpy(output_type[i].units, "unitless\0");

i = AOT_M;   /* new, 31-OCT-05 */
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FCBOTH;
output_type[i].type = DFNT_UINT8;
output_type[i].lowrange  = (int16)1; 
output_type[i].hirange   = (int16)5; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)1.0;
strcpy(output_type[i].units, "unitless\0");

i = SURFP;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_UINT8;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)255; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)200.0;
strcpy(output_type[i].units, "atm\0");

i = SLEVP;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_UINT8;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)255; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)200.0;
strcpy(output_type[i].units, "atm\0");

i = WATERVAPOR;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FCBOTH;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)5000; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)100.0;
strcpy(output_type[i].units, "g/cm^2\0");

i = NCEPWATERVAPOR;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_UINT8;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)255; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)50.0;
strcpy(output_type[i].units, "g/cm^2\0");

i = OZONE;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_UINT8;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)255; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)400.0;
strcpy(output_type[i].units, "cm atm\0");

i = AIR_TEMP;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)65535; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)100.0;
strcpy(output_type[i].units, "degrees K\0");

i = LAT;
output_type[i].DN = DNBOTH;
output_type[i].FC = FCBOTH;
output_type[i].type = DFNT_FLOAT;
output_type[i].lowrange  = (int16)-90; 
output_type[i].hirange   = (int16)90; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)1.0;
strcpy(output_type[i].units, "degrees\0");

i = LON;
output_type[i].DN = DNBOTH;
output_type[i].FC = FCBOTH;
output_type[i].type = DFNT_FLOAT;
output_type[i].lowrange  = (int16)-180; 
output_type[i].hirange   = (int16)180; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)1.0;
strcpy(output_type[i].units, "degrees\0");

i = SOLZ;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_FLOAT;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)180; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)1.0;
strcpy(output_type[i].units, "degrees\0");

i = SATZ;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_FLOAT;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)180; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)1.0;
strcpy(output_type[i].units, "degrees\0");

i = RPHI;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_FLOAT;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)180; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)1.0;
strcpy(output_type[i].units, "degrees\0");

i = U_WND;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_FLOAT;
output_type[i].lowrange  = (int16)-50; 
output_type[i].hirange   = (int16)50; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)1.0;
strcpy(output_type[i].units, "m/s\0");

i = V_WND;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_FLOAT;
output_type[i].lowrange  = (int16)-50; 
output_type[i].hirange   = (int16)50; 
output_type[i].fillvalue = (int16)0;
output_type[i].scale = (float64)1.0;
strcpy(output_type[i].units, "m/s\0");

i = QKMQA;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FINE_ONLY;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)16383; 
output_type[i].fillvalue = (int16)65535;

i = HKMQA;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FINE_ONLY;
output_type[i].type = DFNT_UINT32;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)16383;   /* this is wrong */
output_type[i].fillvalue = (int16)3;

i = OKMQA;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FINE_ONLY;
output_type[i].type = DFNT_UINT32;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)16383;   /* this is wrong */
output_type[i].fillvalue = (int16)3;

i = STATE;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FINE_ONLY;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)49151; 
output_type[i].fillvalue = (int16)65535;

i = AOTCM;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FINE_ONLY;
output_type[i].type = DFNT_UINT16;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)65535;
output_type[i].fillvalue = (int16)0;

i = AOTQA;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = FINE_ONLY;
output_type[i].type = DFNT_UINT8;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)19;
output_type[i].fillvalue = (int16)0;

i = CRSCM;
output_type[i].DN = DAY_ONLY;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_UINT8;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)6;
output_type[i].fillvalue = (int16)0;

i = CRSQA;
output_type[i].DN = DNBOTH;
output_type[i].FC = COARSE_ONLY;
output_type[i].type = DFNT_UINT8;
output_type[i].lowrange  = (int16)0; 
output_type[i].hirange   = (int16)65535;
output_type[i].fillvalue = (int16)0;

/* 31-MAR-06 */
#ifndef TOTEMP
for(i=TRBAND1;i<=TGBAND16;i++) {
   output_type[i].DN = DAY_ONLY;
   output_type[i].FC = FINE_ONLY;
   output_type[i].type = DFNT_INT16;
   output_type[i].lowrange  = (int16)0;
   output_type[i].hirange   = (int16)16000;
   output_type[i].fillvalue = (int16)DATAFILLVALUE;
   output_type[i].scale = (float64)REFSCALE;
   strcpy(output_type[i].units, "unitless\0");
   }
#endif

/* Set up SDSname strings for the "later" SDSs */
strcpy(sdsname[BAND20ALB-BAND20ALB], "Surface Reflectance Band 20\0");
strcpy(sdsname[AOT_1-BAND20ALB], "Atmospheric Optical Depth Band 1\0");
strcpy(sdsname[AOT_3-BAND20ALB], "Atmospheric Optical Depth Band 3\0");
strcpy(sdsname[AOT_8-BAND20ALB], "Atmospheric Optical Depth Band 8\0");
strcpy(sdsname[AOT_M-BAND20ALB], "Atmospheric Optical Depth Model\0");
strcpy(sdsname[SURFP-BAND20ALB], "Surface Pressure\0");
strcpy(sdsname[SLEVP-BAND20ALB], "Mean Sea Level Pressure\0");
strcpy(sdsname[WATERVAPOR-BAND20ALB], "Water Vapor\0");
strcpy(sdsname[NCEPWATERVAPOR-BAND20ALB], "NCEP Water Vapor\0");
strcpy(sdsname[OZONE-BAND20ALB], "Ozone\0");
strcpy(sdsname[AIR_TEMP-BAND20ALB], "Air Temperature (2m)\0");
strcpy(sdsname[LAT-BAND20ALB], "Latitude\0");
strcpy(sdsname[LON-BAND20ALB], "Longitude\0");
strcpy(sdsname[SOLZ-BAND20ALB], "Solar Zenith\0");
strcpy(sdsname[SATZ-BAND20ALB], "Sensor Zenith\0");
strcpy(sdsname[RPHI-BAND20ALB], "Relative Azimuth\0");
strcpy(sdsname[U_WND-BAND20ALB], "U-component of Wind\0");
strcpy(sdsname[V_WND-BAND20ALB], "V-component of Wind\0");
strcpy(sdsname[QKMQA-BAND20ALB], "Reflectance Band Quality\0");
strcpy(sdsname[HKMQA-BAND20ALB], "Reflectance Band Quality\0");
strcpy(sdsname[OKMQA-BAND20ALB], "Reflectance Band Quality\0");
strcpy(sdsname[STATE-BAND20ALB], "Reflectance Data State QA\0");
strcpy(sdsname[AOTCM-BAND20ALB], "Atmospheric Optical Depth Band QA\0");   /* 5.0.2 */
strcpy(sdsname[AOTQA-BAND20ALB], "Atmospheric Optical Depth Band CM\0");   /* 5.0.2 */
strcpy(sdsname[CRSQA-BAND20ALB], "QA\0");
strcpy(sdsname[CRSCM-BAND20ALB], "CM\0");

/* 31-MAR-06 */
#ifndef TOTEMP
for(i=TRBAND1;i<=TRBAND16;i++) 
   sprintf(sdsname[i-BAND20ALB], "Transmission Band %d\0", (i-TRBAND1)+1);
for(i=PRBAND1;i<=PRBAND16;i++) 
   sprintf(sdsname[i-BAND20ALB], "Path Radiance Band %d\0", (i-PRBAND1)+1);
for(i=SABAND1;i<=SABAND16;i++) 
   sprintf(sdsname[i-BAND20ALB], "Spherical Albedo Band %d\0", (i-SABAND1)+1);
for(i=TGBAND1;i<=TGBAND16;i++) 
   sprintf(sdsname[i-BAND20ALB], "Gaseous Transmission Band %d\0", (i-TGBAND1)+1);
#endif
  

/* 17-JUL-01: read avg_z data */
/*********************************************
 *** 13-FEB-06: Use of PRMSL no longer *******
 *** requires correction by avg_z.   *********
use_avg_z = init_height_pressure(&avg_z); 
if (use_avg_z != 0)              
    modsmf((PGSt_SMF_code)MODIS_F_FILE_READ_ERROR, "Avg exp(-z/8) ancillary table", location);
 *********************************************
 *********************************************/
 
for(i=0;i<7;i++) {
   sref[i]=NULL;     /* initialize sref */
   ta_tab[i]=NULL;   /* initialize ta_tab */
   }
for(i=0;i<7;i++){
   for(j=0;j<10;j++)Sra_tab[i][j]=0.0F;   /* initialize Sra_tab*/
   }

/* 04-APR-01: for polarization correction */
n_sides = -1;
n_AOIs = -1;
n_dets = -1;
n_bands = -1;
n_lambda = -1;
n_rad_ray = -1;
n_sun_ray = -1;
n_wind_ray = -1;
n_order_ray = -1;
    
commands.pol_corr=0;  /* assume NO polarization correction */

/* Read polarization data */
if (read_polarization_data(POLARIZATION_TABLE_LUN, POLARIZATION_TABLE_LUN2) != 0) {
   modsmf((PGSt_SMF_code)MODIS_E_FILE_ACCESS, "polarization correction lookup tables", location);
   }
else {
   if (read_Rayleigh_data(RAYLEIGH_TABLE_LUN) != 0) {
      modsmf((PGSt_SMF_code)MODIS_E_FILE_ACCESS, "Rayleigh coefficient lookup table", location);
      }
   else commands.pol_corr=1;
   }

/* routine for avoiding AOTR over certain places */
if (init_aotr_rejection(AOTR_REJ_LUN)) aotr_struct.n_aotr_offsets = 0;

/*printf("%p %p %f %d\n",  aotr_struct.aotr_offsets, aotr_struct.aotr_lengths,  aotr_struct.aotr_resolution, aotr_struct.n_aotr_offsets);*/

/* read the new LUTs 
 *********************/
 /* We're going to need some way of switching between models... */
strcpy(CAMOD, "SMOKELABS\0");
strlencamod = strlen(CAMOD);

/* but first, get the names of the files from the PCF */
ver = 1;
tauraynm[0] = '\0';
if(PGS_PC_GetReference(taurayln, &ver, tauraynm)!=PGS_S_SUCCESS){
    modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, "rayleigh o.d file", location);
    return(1);
    }
/*printf("%s\n", tauraynm);*/

ver = 1;
gscoefnm[0] = '\0';
if(PGS_PC_GetReference(gscoefln, &ver, gscoefnm)!=PGS_S_SUCCESS){
    modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, "gaseous transmission coefficient file", location);
    return(1);
    }
/*printf("%s\n", gscoefnm);*/

ver = 1;
anglehdf[0] = '\0';
if(PGS_PC_GetReference(anglehdfln, &ver, anglehdf)!=PGS_S_SUCCESS){
    modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, "ANGLE.hdf file", location);
    return(1);
    }
/*printf("%s\n", anglehdf);*/

for(i=SMKL;i<MODELS;i++) { /* 31-OCT-05: multiple models */

  switch(i) {
     case (SMKL):
        intrefln = SMKL_RLUTLUN; transmln = SMKL_TLUTLUN; spheraln = SMKL_ALUTLUN;
	break;
     case (SMKH):
        intrefln = SMKH_RLUTLUN; transmln = SMKH_TLUTLUN; spheraln = SMKH_ALUTLUN;
	break;
     case (DSST):
        intrefln = DUST_RLUTLUN; transmln = DUST_TLUTLUN; spheraln = DUST_ALUTLUN;
	break;
     case (URBP):
        intrefln = URBP_RLUTLUN; transmln = URBP_TLUTLUN; spheraln = URBP_ALUTLUN;
	break;
     case (URBC):
        intrefln = URBC_RLUTLUN; transmln = URBC_TLUTLUN; spheraln = URBC_ALUTLUN;
	break;
     }
     
  ver = 1; 
  intrefnm[0] = '\0';
  if(PGS_PC_GetReference(intrefln, &ver, intrefnm)!=PGS_S_SUCCESS){
    modsmf((PGSt_SMF_code)MODIS_F_GETREF_FAIL, "intrinsic reflectance file", location);
    return(1);
    }
    
  ver = 1;
  transmnm[0] = '\0';
  if(PGS_PC_GetReference(transmln, &ver, transmnm)!=PGS_S_SUCCESS){
    modsmf((PGSt_SMF_code)MODIS_F_GETREF_FAIL, "transmission file", location);
    return(1);
    }
   
  ver = 1;
  spheranm[0] = '\0';
  if(PGS_PC_GetReference(spheraln, &ver, spheranm)!=PGS_S_SUCCESS){
    modsmf((PGSt_SMF_code)MODIS_F_GETREF_FAIL, "spherical albedo file", location);
    return(1);
    }
   /*printf("%s %s %s %s %s %s\n", tauraynm, gscoefnm, anglehdf, intrefnm, transmnm, spheranm);*/

  switch(i) {
     case (SMKL):
        readluts(CAMOD, &strlencamod, tauray, oztransa, wvtransa, wvtransb, wvtransc, ogtransa0, 
	         ogtransa1, ogtransb0, ogtransb1, ogtransc0, ogtransc1, tsmax, tsmin, 
                 ttv, tts, nbfi, nbfic, indts, roluttSMKL, transtSMKL, sphalbtSMKL, sbandname,
 	         err_msg, &retval, tauraynm, gscoefnm, anglehdf, intrefnm, transmnm, 
	         spheranm);
	break;
     case (SMKH):
        readluts(CAMOD, &strlencamod, tauray, oztransa, wvtransa, wvtransb, wvtransc, ogtransa0, 
	         ogtransa1, ogtransb0, ogtransb1, ogtransc0, ogtransc1, tsmax, tsmin, 
                 ttv, tts, nbfi, nbfic, indts, roluttSMKH, transtSMKH, sphalbtSMKH, sbandname,
	         err_msg, &retval, tauraynm, gscoefnm, anglehdf, intrefnm, transmnm, 
	         spheranm);
	break;
     case (DSST):
        readluts(CAMOD, &strlencamod, tauray, oztransa, wvtransa, wvtransb, wvtransc, ogtransa0, 
	         ogtransa1, ogtransb0, ogtransb1, ogtransc0, ogtransc1, tsmax, tsmin, 
                 ttv, tts, nbfi, nbfic, indts, roluttDUST, transtDUST, sphalbtDUST, sbandname,
	         err_msg, &retval, tauraynm, gscoefnm, anglehdf, intrefnm, transmnm, 
	         spheranm);
	break;
     case (URBP):
        readluts(CAMOD, &strlencamod, tauray, oztransa, wvtransa, wvtransb, wvtransc, ogtransa0, 
	         ogtransa1, ogtransb0, ogtransb1, ogtransc0, ogtransc1, tsmax, tsmin, 
                 ttv, tts, nbfi, nbfic, indts, roluttURBPOL, transtURBPOL, sphalbtURBPOL, sbandname,
	         err_msg, &retval, tauraynm, gscoefnm, anglehdf, intrefnm, transmnm, 
	         spheranm);
	break;
     case (URBC):
        readluts(CAMOD, &strlencamod, tauray, oztransa, wvtransa, wvtransb, wvtransc, ogtransa0, 
	         ogtransa1, ogtransb0, ogtransb1, ogtransc0, ogtransc1, tsmax, tsmin, 
                 ttv, tts, nbfi, nbfic, indts, roluttURBCLN, transtURBCLN, sphalbtURBCLN, sbandname,
	         err_msg, &retval, tauraynm, gscoefnm, anglehdf, intrefnm, transmnm, 
	         spheranm);
	break;
     }

  if (retval != 0){
      err_msg[strlen(err_msg)] = '\0';
      sprintf(msg, "Error: '%s' from routine readluts: cannot continue \n", err_msg);
      modsmf(MODIS_F_GETFILEINFO, msg, location);
      }
  
  }  /* 31-OCT-05: multiple models */


  
/* determine how many inputs are in the PCF */
n_inputs = 0;
local_max = 0;
for(j=0;j<ALL_L1B_FILES;j++) {
   n = (PGSt_integer)0;
   if(PGS_PC_GetNumberOfFiles((PGSt_integer)input_luns[j], &n)!=PGS_S_SUCCESS) {
       sprintf(msg, "%d", input_luns[j]);
       modsmf(MODIS_E_GETNUM_FILES, msg, location);
     }
  n_in[j] = n;
  if (n > local_max)local_max = n;
  n_inputs += n;
  }
if (n_inputs == 0) { 
    sprintf(msg, "No input files were staged, exiting");
    modsmf((PGSt_SMF_code)MODIS_E_GETNUM_FILES, msg, location);
    return(-1);
  }

*max = local_max;
return(0);

}

int get_variable_memory(float ***Rho_ra_term, float ***Rho_r_term, float ***Step1_term, float ***Step2_term, float ***Step3_term, float ***SRA_term, float ***tau,
                        float ***hkmStep1_term, float ***hkmStep2_term, float ***hkmStep3_term, float ***hkmSRA_term,
                        float ***qkmStep1_term, float ***qkmStep2_term, float ***qkmStep3_term, float ***qkmSRA_term, int n_bands[3])
/*
!C****************************************************************************

!Routine: get_variable_memory()

!Description: Gets memory for atmospheric correction variables.

!Input Parameters:
  int n_bands[3]       The number of REFLECTANCE bands to be processed at each resolution (250m, 500m, 1km)
 
!Input/Output Parameters:
  All the necessary atmospheric correction variables

!Output Parameters:
  none

!Returns:
  Success: 0
  Failure: nonzero (yet to be implemented)
  
!Revision History:
  Revision 1.0  09-APR-03
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
  
!END**************************************************************************/
{
int i;
if (n_bands[0] > 0) {   /* QKM */
    *qkmStep1_term  = (float **)malloc(2*sizeof(float *));
    *qkmStep2_term  = (float **)malloc(2*sizeof(float *));
    *qkmStep3_term  = (float **)malloc(2*sizeof(float *));
    *qkmSRA_term    = (float **)malloc(2*sizeof(float *));
}
if (n_bands[1] > 0) {   /* HKM */
    *hkmStep1_term  = (float **)malloc(7*sizeof(float *));
    *hkmStep2_term  = (float **)malloc(7*sizeof(float *));
    *hkmStep3_term  = (float **)malloc(7*sizeof(float *));
    *hkmSRA_term    = (float **)malloc(7*sizeof(float *));
}
if (n_bands[2] > 0) {   /* 1KM */
    *Rho_ra_term = (float **)malloc(7*sizeof(float *));
    *Rho_r_term    = (float **)malloc(7*sizeof(float *));
    *Step1_term  = (float **)malloc(7*sizeof(float *));
    *Step2_term  = (float **)malloc(7*sizeof(float *));
    *Step3_term  = (float **)malloc(7*sizeof(float *));
    *SRA_term    = (float **)malloc(7*sizeof(float *));
    *tau         = (float **)malloc(7*sizeof(float *));
}

for (i=BAND1;i<=BAND7;i++) {     
     if (n_bands[2] > 0) {   /* 1KM */
        (*Rho_ra_term)[i] = (float *)malloc(NBSCANS_1KM*NBSPOTS_1KM*sizeof(float));
        (*Rho_r_term)[i] = (float *)malloc(NBSCANS_1KM*NBSPOTS_1KM*sizeof(float));
        (*Step1_term)[i] = (float *)malloc(NBSCANS_1KM*NBSPOTS_1KM*sizeof(float));
        (*Step2_term)[i] = (float *)malloc(NBSCANS_1KM*NBSPOTS_1KM*sizeof(float));
        (*Step3_term)[i] = (float *)malloc(NBSCANS_1KM*NBSPOTS_1KM*sizeof(float));
        (*SRA_term)[i] = (float *)malloc(NBSCANS_1KM*NBSPOTS_1KM*sizeof(float));
        (*tau)[i] = (float *)malloc(NBSCANS_1KM*NBSPOTS_1KM*sizeof(float));
        }
     if (n_bands[1] > 0) {   /* HKM */
        (*hkmStep1_term)[i] = (float *)malloc(NBSCANS_500M*NBSPOTS_500M*sizeof(float));
        (*hkmStep2_term)[i] = (float *)malloc(NBSCANS_500M*NBSPOTS_500M*sizeof(float));
        (*hkmStep3_term)[i] = (float *)malloc(NBSCANS_500M*NBSPOTS_500M*sizeof(float));
        (*hkmSRA_term)[i] = (float *)malloc(NBSCANS_500M*NBSPOTS_500M*sizeof(float));
        }
	
    }   
for (i=BAND1;i<=BAND2;i++) {
    if (n_bands[0] > 0) {   /* QKM */
        (*qkmStep1_term)[i] = (float *)malloc(NBSCANS_250M*NBSPOTS_250M*sizeof(float));
        (*qkmStep2_term)[i] = (float *)malloc(NBSCANS_250M*NBSPOTS_250M*sizeof(float));
        (*qkmStep3_term)[i] = (float *)malloc(NBSCANS_250M*NBSPOTS_250M*sizeof(float));
        (*qkmSRA_term)[i] = (float *)malloc(NBSCANS_250M*NBSPOTS_250M*sizeof(float));
      }
}
return(0);
}


int free_variable_memory(float **Rho_ra_term, float **Rho_r_term, float **Step1_term, float **Step2_term, float **Step3_term, float **SRA_term, float **tau,
                         float **hkmStep1_term, float **hkmStep2_term, float **hkmStep3_term, float **hkmSRA_term,
                         float **qkmStep1_term, float **qkmStep2_term, float **qkmStep3_term, float **qkmSRA_term, int n_bands[3])
/*
!C****************************************************************************

!Routine: free_variable_memory()

!Description: Frees memory for atmospheric correction variables.

!Input Parameters:
  int n_bands[3]       The number of REFLECTANCE bands to be processed at each resolution (250m, 500m, 1km)
 
!Input/Output Parameters:
  All the necessary atmospheric correction variables

!Output Parameters:
  none

!Returns:
  Success: 0
  
!Revision History:
  Revision 1.0  09-APR-03
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
      08-DEC-04: changed around the steps for freeing QKM, HKM variables.
     
!END**************************************************************************/
{
int i;

if (qkmStep1_term != NULL) {
    for (i=BAND1;i<=BAND2;i++) {
       if (qkmStep1_term[i] != NULL) free(qkmStep1_term[i]);
       }
    free(qkmStep1_term);
   }
if (qkmStep2_term != NULL) {
    for (i=BAND1;i<=BAND2;i++) {
       if (qkmStep2_term[i] != NULL) free(qkmStep2_term[i]);
       }
    free(qkmStep2_term);
   }
if (qkmStep3_term != NULL) {
    for (i=BAND1;i<=BAND2;i++) {
       if (qkmStep3_term[i] != NULL) free(qkmStep3_term[i]);
       }
    free(qkmStep3_term);
   }
if (qkmSRA_term != NULL) {
    for (i=BAND1;i<=BAND2;i++) {
       if (qkmSRA_term[i] != NULL) free(qkmSRA_term[i]);
       }
    free(qkmSRA_term);
   }
   
if (hkmStep1_term != NULL) {
    for (i=BAND1;i<=BAND7;i++) {
       if (hkmStep1_term[i] != NULL) free(hkmStep1_term[i]);
       }
    free(hkmStep1_term);
   }
if (hkmStep2_term != NULL) {
    for (i=BAND1;i<=BAND7;i++) {
       if (hkmStep2_term[i] != NULL) free(hkmStep2_term[i]);
       }
    free(hkmStep2_term);
   }
if (hkmStep3_term != NULL) {
    for (i=BAND1;i<=BAND7;i++) {
       if (hkmStep3_term[i] != NULL) free(hkmStep3_term[i]);
       }
    free(hkmStep3_term);
   }
if (hkmSRA_term != NULL) {
    for (i=BAND1;i<=BAND7;i++) {
       if (hkmSRA_term[i] != NULL) free(hkmSRA_term[i]);
       }
    free(hkmSRA_term);
   }
   
if (n_bands[2] > 0) {   /* 1KM */
   for (i=BAND1;i<=BAND7;i++) {
       free(Rho_ra_term[i]);
       free(Rho_r_term[i]);
       free(Step1_term[i]);
       free(Step2_term[i]);
       free(Step3_term[i]);
       free(SRA_term[i]);
       free(tau[i]);
    }
    free(Rho_ra_term);
    free(Rho_r_term);
    free(Step1_term);
    free(Step2_term);
    free(Step3_term);
    free(SRA_term);
    free(tau);
   }
return(0);
}

int init_height_pressure(d_ancillary *avg_z) 
/*
!C*****************************************************************************

!Description:   init_height_pressure; initializes average exp(-z/8) structure.

!Input Parameters:       none

!Output Parameters:      none

!Input/Ouput Parameters: avg_z (updated)

!Returns:   0 for success
            1 for error

!Revision History:

1. Original Version, Jim Ray, May 21, 2001

!Team-unique Header:

    This software developed by the MODIS Land Science Team for
    the National Aeronautics and Space Administration, Goddard
    Space Flight Center, under contract NAS5-96062 and NAS5-31752

!References and Credits:
  
    MODIS Science Team Members:
			
	Eric F Vermote
	Univerisity of MD.
   	NASA's GSFC Code 923
   	Greenbelt, MD , 20771
	eric@kratmos.gsfc.nasa.gov
	301-286-6232

    Developers:
      Jim Ray
      MODIS Land Science Team           Science Systems and Applications Inc.
      jim@cratmos.gsfc.nasa.gov         NASA's GSFC Code 923 
      phone (301) 614-6684              Greenbelt, MD 20771
 

!Design Notes
     File 1_km_z8.hdf contains SDS "average exp(-z/8)", which contains height data from
     PGS Toolkit DEM files.  Here is a blurb from the metadata of file 1_km_z8.hdf:
      
       Values in SDS 'Elevation' from input [DEM] files are added to corresponding
       values in SDS 'Geoid', to represent meters above the geoid -- variable z.
       This SDS contains the values of exp(-z/8000) averaged over the corresponding
       1 km space.  SDS 'n' contains the number of 1 km values averaged. SDS 'geoid'
       is the geoid used, taken from the first of the input files (all input files
       have identical 'Geoid' SDSs).

!END
****************************************************************************/			
{
PGSt_integer ver;
char tf_name[PGSd_PC_FILE_PATH_MAX]={'\0'};
char location[100]="modis_srefl.c, init_height_pressure\0";
char msg[PGS_SMF_MAX_MSG_SIZE]; /* buffer to hold SMF message */
int32 sdid, sds, rank, start[2], ddims[5], number_type, nattr;
char name[256];
PGSt_PC_Logical avgzheight_lun;

/* Get the relevant avg exp(-z/8) data from ancillary table */
ver = 1;
avgzheight_lun = (PGSt_PC_Logical)DEGDEMFILE;  
if(PGS_PC_GetReference(avgzheight_lun, &ver, tf_name)!=PGS_S_SUCCESS){
    modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, "avg exp(-z/8) file", location);
    return(1);
    }

/* just for now * /
strcpy(tf_name, "/data2/home/jim/1_km_z8.hdf\0"); */
if ( (sdid = SDstart(tf_name, DFACC_RDONLY)) == -1) {
    sprintf(msg, "LUN %d, version %d", (int32)avgzheight_lun, (int32)1);
    modsmf((PGSt_SMF_code)MODIS_E_OPEN_HDF_FILE, msg, location);
    return(1);
    }

if ( (sds = SDselect(sdid, SDnametoindex(sdid, "average exp(-z/8)"))) == -1) {
    modsmf((PGSt_SMF_code)MODIS_E_FILE_READ_ERROR, "defined by avg exp(-z/8)", location);
    SDend(sdid);   
    return(1);
    }
    
if ( SDgetinfo(sds, name, &rank, ddims, &number_type, &nattr) == -1) {
    SDendaccess(sds);
    SDend(sdid);   
    modsmf((PGSt_SMF_code)MODIS_E_FILE_READ_ERROR, "defined by avg exp(-z/8)", location);
    return(1);
    }

if ((avg_z->data = (float32 *)calloc(ddims[0]*ddims[1], sizeof(float32))) == NULL){
    SDendaccess(sds);
    SDend(sdid);   
    return(1);
    }

start[0] = start[1] = 0;
if ( SDreaddata(sds, start, NULL, ddims, (VOID *)avg_z->data) == -1) {
    SDendaccess(sds);
    SDend(sdid);   
    modsmf((PGSt_SMF_code)MODIS_E_FILE_READ_ERROR, "defined by avg exp(-z/8)", location);
    return(1);
    }
SDendaccess(sds);
SDend(sdid);   
avg_z->dims[1] = ddims[0];
avg_z->dims[2] = ddims[1];

return(0);
}



void zero_out(void *inputs, int32 type, int32 count)
/*
!C****************************************************************************

!Routine: zero_out()

!Description: Generic void-pointer initializing routine.

!Input Parameters:
  int32 type             The type of data to be zeroed
  int32 count            Length of the array to be zero-outed
  
!Input/Output Parameters:
  void *inputs           Array to be zeroed (updated)       

!Output Parameters:
  none    

!Returns:
  none
  
!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
 
!END**************************************************************************/
{
int32 *i32_array;
uint32 *ui32_array;
float64 *f64_array;
float32 *f32_array;
int8 *i8_array;
uint8 *ui8_array;
int16 *i16_array;
uint16 *ui16_array;
char *char_array;
uchar8 *uchar_array;
       switch(type) {
           case 3:
              uchar_array = (uchar8 *)inputs;
	      memset(uchar_array, '\0', count*sizeof(uchar8));
              break;
           case 4:
              char_array = (char *)inputs;
	      memset(char_array, '\0', count*sizeof(char));
              break;
           case 5:
              f32_array = (float *)inputs;
	      memset(f32_array, '\0', count*sizeof(float));
              break;
           case 6:
              f64_array = (double *)inputs;
	      memset(f64_array, '\0', count*sizeof(double));
              break;
           case 20:
              i8_array = (int8 *)inputs;
	      memset(i8_array, '\0', count*sizeof(int8));
              break;
           case 21:
              ui8_array = (uint8 *)inputs;
	      memset(ui8_array, '\0', count*sizeof(uint8));
              break;
           case 22:
              i16_array = (int16 *)inputs;
	      memset(i16_array, '\0', count*sizeof(int16));
              break;
           case 23:
              ui16_array = (uint16 *)inputs;
	      memset(ui16_array, '\0', count*sizeof(uint16));
              break;
           case 24:
              i32_array = (int32 *)inputs;
	      memset(i32_array, '\0', count*sizeof(int32));
              break;
           case 25:
              ui32_array = (uint32 *)inputs;
	      memset(ui32_array, '\0', count*sizeof(uint32));
              break;

             }
}

int32 get_memory(void **data, int32 type, int32 count)
/*
!C****************************************************************************

!Routine: get_memory()

!Description: Generic void-pointer memory allocation routine.

!Input Parameters:
  int32 type             The type of data to be allocated for
  int32 count            Length of the array
  
!Input/Output Parameters:
  void **data            Array to be allocated for (updated)       
                          

!Output Parameters:
  none    

!Returns:
  success: 0
  failure: -13  (memory failure)
  
!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
    
!END**************************************************************************/
{
int32 *i32_array;
uint32 *ui32_array;
float64 *f64_array;
float32 *f32_array;
int8 *i8_array;
uint8 *ui8_array;
int16 *i16_array;
uint16 *ui16_array;
char *char_array;
uchar8 *uchar_array;
int32 ret = 0;
char msg[256] = {"\0"};
char location[100]="modis_srefl.c, get_memory";

       switch(type) {
           case 4:
              char_array = (char *)malloc(count*sizeof(char));
	      if (char_array != NULL)  *data = (void *)char_array;
	      else ret = -13;
              break;
           case 3:
              uchar_array = (uchar8 *)malloc(count*sizeof(uchar8));
	      if (uchar_array != NULL) *data = (void *)uchar_array;
	      else ret = -13;
              break;
           case 5:
              f32_array = (float32 *)malloc(count*sizeof(float32));
	      if (f32_array != NULL) *data = (void *)f32_array;
	      else ret = -13;
              break;
           case 6:
              f64_array = (float64 *)malloc(count*sizeof(float64));
	      if (f64_array != NULL) *data = (void *)f64_array;
	      else ret = -13;
              break;
           case 20:
              i8_array = (int8 *)malloc(count*sizeof(int8));
	      if (i8_array != NULL) *data = (void *)i8_array;
	      else ret = -13;
              break;
           case 21:
              ui8_array = (uint8 *)malloc(count*sizeof(uint8));
	      if (ui8_array != NULL) *data = (void *)ui8_array;
	      else ret = -13;
              break;
           case 22:
              i16_array = (int16 *)malloc(count*sizeof(int16));
	      if (i16_array != NULL) *data = (void *)i16_array;
	      else ret = -13;
              break;
           case 23:
              ui16_array = (uint16 *)malloc(count*sizeof(uint16));
	      if (ui16_array != NULL) *data = (void *)ui16_array;
	      else ret = -13;
              break;
           case 24:
              i32_array = (int32 *)malloc(count*sizeof(int32));
	      if (i32_array != NULL) *data = (void *)i32_array;
	      else ret = -13;
              break;
           case 25:
              ui32_array = (uint32 *)malloc(count*sizeof(uint32));
	      if (ui32_array != NULL) *data = (void *)ui32_array;
	      else ret = -13;
              break;
           default:
              sprintf(msg, "Bad data type value, '%d'; will skip", type);
	      modsmf(MODIS_E_MEM_FAILURE, msg, location);
              break;
                         }

return(ret);
} 

void *get_image(int32 id, int32 *dims, int32 type, int *size)
{
int32 *i32_array;
uint32 *ui32_array;
float64 *f64_array;
float32 *f32_array;
int8 *i8_array;
uint8 *ui8_array;
int16 *i16_array;
uint16 *ui16_array;
char *char_array;
uchar8 *uchar_array;
int i,count;
int32 start[5];

start[0] = start[1] = start[2] = start[3] = start[4] = *size = 0;
i=0;
count = dims[i++];
while(dims[i] != 0) count *= dims[i++];
    

       switch(type) {
           case 4:
	      *size = count;
              char_array = (char *)malloc(*size);
              SDreaddata(id, start, NULL, dims, char_array);
	      return((void *)char_array);
              break;
           case 3:
	      *size = count;
              uchar_array = (uchar8 *)malloc(*size);
              SDreaddata(id, start, NULL, dims, uchar_array);
	      return((void *)uchar_array);
              break;
           case 5:
	      *size = count*sizeof(float);
              f32_array = (float *)malloc(*size);
              SDreaddata(id, start, NULL, dims, f32_array);
	      return((void *)f32_array);
              break;
           case 6:
	      *size = count*sizeof(double);
              f64_array = (double *)malloc(*size);
              SDreaddata(id, start, NULL, dims, f64_array);
	      return((void *)f64_array);
              break;
           case 20:
	      *size = count*sizeof(int8);
              i8_array = (int8 *)malloc(*size);
              SDreaddata(id, start, NULL, dims, i8_array);
	      return((void *)i8_array);
              break;
           case 21:
	      *size = count*sizeof(uint8);
              ui8_array = (uint8 *)malloc(*size);
              SDreaddata(id, start, NULL, dims, ui8_array);
	      return((void *)ui8_array);
              break;
           case 22:
	      *size = count*sizeof(int16);
              i16_array = (int16 *)malloc(*size);
              SDreaddata(id, start, NULL, dims, i16_array);
	      return((void *)i16_array);
              break;
           case 23:
	      *size = count*sizeof(uint16);
              ui16_array = (uint16 *)malloc(*size);
              SDreaddata(id, start, NULL, dims, ui16_array);
	      return((void *)ui16_array);
              break;
           case 24:
	      *size = count*sizeof(int32);
              i32_array = (int32 *)malloc(*size);
              SDreaddata(id, start, NULL, dims, i32_array);
	      return((void *)i32_array);
              break;
           case 25:
	      *size = count*sizeof(uint32);
              ui32_array = (uint32 *)malloc(*size);
              SDreaddata(id, start, NULL, dims, ui32_array);
	      return((void *)ui32_array);
              break;
                         }
}

