#include "PGS_TD.h"
#include "PGS_PC.h"
#include "PGS_IO.h"
#include "PGS_SMF.h"
#include "struct.h"

/* in file modis_srefl.c */
void report(dims_struct ***inputs, int n_timepoints);
int init_processing(int *max, int *n_in, int *input_luns);
void zero_out(void *inputs, int32 type, int32 count);
int32 get_memory(void **data, int32 type, int32 n_cols); 
int init_height_pressure(d_ancillary *avg_z); 
void clean_up(void);
int get_variable_memory(float ***Rhop_r_term, float ***T_r_term, float ***Step1_term, float ***Step2_term, float ***Step3_term, float ***SRA_term, float ***tau,
                        float ***hkmStep1_term, float ***hkmStep2_term, float ***hkmStep3_term, float ***hkmSRA_term,
                        float ***qkmStep1_term, float ***qkmStep2_term, float ***qkmStep3_term, float ***qkmSRA_term, int n_bands[3]);
int free_variable_memory(float **Rhop_r_term, float **T_r_term, float **Step1_term, float **Step2_term, float **Step3_term, float **SRA_term, float **tau,
                        float **hkmStep1_term, float **hkmStep2_term, float **hkmStep3_term, float **hkmSRA_term,
                        float **qkmStep1_term, float **qkmStep2_term, float **qkmStep3_term, float **qkmSRA_term, int n_bands[3]);

/* in file modis_srefl_io.c */
int lgid_parser(char *lgid, char *esdt, int *jday);
void get_input_str_2D_memory(dims_struct ****inputs, int dim1, int dim2);
void get_input_str_1D_memory(dims_struct ***inputs, int dim1);
int setup_inputs(dims_struct ****inputs, int *n_in, int *input_luns);
int sort_inputs(dims_struct ****inputs, int *n_in);
int get_info(int lun, PGSt_integer version, dims_struct *inputs);
void init_struct(dims_struct *inputs, char *call);
int close_HDF_interface(dims_struct *inputs);
int close_io_struct(dims_struct ***input, int dim1, int dim2);
int close_io_struct2(dims_struct **input, int dim1);
void set_core_metadata(dims_struct ***inputs, int *n_inputs);
void set_archive_metadata(void);
PGSt_SMF_status rmpath(char *buffer);
void modsmf(PGSt_SMF_code mnemonicstring, char *infostring, char *functionstring);
int get_granule_times(dims_struct ****inputs, int *n_in, PGSt_double orbit_start_tai, 
                      PGSt_double orbit_end_tai, PGSt_double *granule_start_tai, PGSt_double 
		      *granule_end_tai);
int get_unique_granule_times(dims_struct ****inputs, int *n_in, PGSt_double **all_granule_start_tais, 
                             int *n_unique_tais);
int compare_tai(const void *i, const void *j);
int open_inputs(dims_struct **local_input, int *nlines_1km, int *ncols_1km, int *local_process_bands);
int get_L1B_sds(int index, dims_struct **local_input, char what_to_get[10][25], int n_SDSs_to_look_for,
                int *local_process_bands);
void get_bands_and_scales(int32 sds, int32 *bands, int32 *n_bands, float32 *refoffset, 
                          float32 *refscale, float32 *radoffset, float32 *radscale);
void get_2D_scale(int32 sds, float64 *scale);
int open_outputs(int mode, int output_lun, dims_struct **local_output, dims_struct *input, int nlines_1km, 
                 int ncols_1km, int *local_process_bands, int *band_numbers);
int open_coarse_outputs(int output_lun, dims_struct **output, int version, int nlines_1km,
                        int ncols_1km, int *local_process_bands, int *band_numbers, int daynightflag);
void structcpy( dims_struct *to, dims_struct *from );
int set_attributes(int32 sdsid, char *longname, char *rez, output_type_t output_type, 
                   int32 type, int8 qa_flag);
int32 read_dead_det(int32 sdid, int32 dead_bands[BANDPROCESS][4 * NBSCANS_1KM], int32 noisy_bands[BANDPROCESS][4 * NBSCANS_1KM]);

/* in file params.c */
int get_params(commands_t *commands, short *TESTLINE, short *TESTPIX, short int fullstemp,
		int32 *scans_per_gran, PGSt_double *orbit_start_tai, 
		PGSt_double *orbit_end_tai, int32 *process_bands, char *report);
void lun_description_lut(char *report, PGSt_PC_Logical number, char *param_value);
void set_processing_flags(commands_t *commands, int32 *process_bands);

/* in file atmcor.c */
int get_scan_memory(int AOTR_flag, dims_struct **local_input, dims_struct **local_output, int total_1km_pixels, 
                    int *local_process_bands);
int free_scan_memory(int AOTR_flag, dims_struct **local_input, dims_struct **local_output, int *local_process_bands);
int atmospheric_correction(int AOTR_flag, dims_struct ***inputs, dims_struct ***outputs, int n_timepoints);
void get_ancillary_data(int mode, char *rbdt);
void ucoarsen(uint16 *fine, int nlines_fine, int nbcols_fine, int factor, int nlines_coarse, int nbcols_coarse, uint16 *coarse);
void u8coarsen(uint8 *fine, int nlines_fine, int nbcols_fine, int factor, int nlines_coarse, int nbcols_coarse, uint8 *coarse);
void coarsen(int16 *fine, int nlines_fine, int nbcols_fine, int factor, int nlines_coarse, int nbcols_coarse, int16 *coarse);
void fcoarsen(float *fine, int nlines_fine, int nbcols_fine, int factor, int nlines_coarse, int nbcols_coarse, float *coarse);
void make_night_coarse_qa(uint8 *fine, int nlines_fine, int nbcols_fine, int factor, int nlines_coarse, int nbcols_coarse, uint8 *coarse);
int set_height_pressure(float *lat, float *lon, int16 *height, float *sp, float *slp);
int32 initatmospars(t_atmospars *pars);
static float32 PrincipalAngle(float32 angle);
static int32 NearlyZero(float32 x);

/* these should be in thermal.c */
void cnvt_data_to_temp(uint16 *input, float64 offset, float64 scale, int n_pixels, float sat_value, float64 lambda);
void cnvt_to_temp(dims_struct *input, int offset, int n_pixels, float sat_value, float64 lambda);
float64 Tb(float64 B, float64 lambda);
float64 PlanckFunction(float64 c1, float64 c2, float64 T);
float64 newPlanckFunction(float64 T, float64 lambda);

/* in file meta.c */
void set_ECS_metadata(filetype_t filetype, input_metadata *metadata);
void grow_Mstringarray(marray_t *string, int32 length, char *value);
void initECSmetadata(input_metadata *input);
void freeECSmetadata(input_metadata *input);
void read_ancillary_metadata(int mode, char *rbd, d_ancillary *dao, char *logreport_string);
void concatenate_fields(PGSt_PC_Logical file_lun, int k, char *field1, char *field2, char *d_and_t);
void clean_up_ancillary_data(void);
void split_datetime(char *datetime, char *date, char *time);
void ECSmetadatacopy(input_metadata *to, input_metadata *from);
void ECScpy(input_metadata *to, input_metadata *from, int mode);
PGSt_SMF_status getECSfileinfo(PGSt_PC_Logical file_lun, PGSt_integer ver, input_metadata *metadata);
int get_ESDT_names(int sattype, char *REF_ESDTname, char *CRF_ESDTname, char *VersionID);
char *create_localgranuleid(char *timecodea, char *version, char *esdt);
void cpy_inputpointer(input_metadata *to, input_metadata *from);
void adapt_metadata_for_CMG(int type, input_metadata *CMGmetadata, input_metadata *metadata, char *lgid_template);
int32 getjulianday(char *timecodea);

/* in file ancillary.c */
float32 local_interpolate(int32 iline, int32 lines, int32 isamp, int32 samps, 
                          float32 rowfactor, float32 colfactor, float32 *data);
PGSt_SMF_status read_dao_plane(PGSt_PC_Logical file_lun, 
			       char *rdbt, char *daovar, d_ancillary *ancillary);
PGSt_SMF_status read_plane(char *rdbt, d_ancillary *ancillary);
int map_ancillary_data(float *lat, float *lon, int total_1km_pixels, float *O3, float *wv, float *sp, float *t2m, float *uw, float *vw, 
                       int *bands, uint16 *qa250, uint32 *qa500, uint32 *qa1);

/* in file utils.c */
void *AllocMem(size_t bytes, char *calling_routine);
void *CAllocMem(size_t num, size_t size);
void *ReAllocMem(void *ptr, size_t bytes);
PGSt_IO_Gen_FileHandle *OpenFileRead(PGSt_PC_Logical logical, PGSt_integer version);

/* in file polarization.c */
#include "pol_driver.h"
int read_polarization_data(int lun1, int lun2);
int get_polarization_data(int32 sd, pol_param_t *param, char *name);
void init_polarization_data(pol_param_t *param, char *name);
void free_polarization_data(pol_param_t *param);
int read_Rayleigh_data(int lun);
int get_Rayleigh_data(int32 sd, pol_param_t *param, char *name);
int polarization_coefficient( float32 SUN, float32 THETA, 
                             float32 PHI, int N_DET, float32 ZAOI, float scan_line_orientation,
		             int N_SIDE, int N_SUN, float32 *RAD);

/* in sra.c */
int32 readsratab(float32 sra[7][10], PGSt_PC_Logical file);
int32 interSra(float32 *Sra_tab,float32 tau_a,float32 *S_ra);
int32 readaerotab(float32 *refl,PGSt_PC_Logical fostab);
int32 interrefl(float32 *refltab,float32 *reflval,float32 ts,
              float32 tv,float32 phi,float32 aot);
int32 readatranstab(float32 *ta_tab, PGSt_PC_Logical file);
int32 interatrans(float32 *table, float32 *t_a, float32 theta, float32 tau500);
int32 index_aot_tab(float32 aot);

/* in aerosols.c */
int init_aotr_rejection(int lun);
int aotr_rejection(float lat, float lon);
int aot_filter(int refout, dims_struct ***outputs);

/* in ids.c */
void writeHDFSDSattributes(int32 sds_id, int16 fillv, int16 vrangeh,
			   int16 vrangel, float64 scale, float64 offset, char *errstring, 
			   int32 cal, char *units, char *long_name);
void which_ids(float *lat);
PGSt_SMF_status update_ids(int scan_number, dims_struct **input, dims_struct *local_output);
int start_coarse_grid(int make_this_cmg[3], d_sds **sds);
int update_coarse_grid(int k, dims_struct *geo_input, dims_struct *local_output, int *which_file);
void preview_coarse_grid(int k, float *lat, int *which_file);
void end_coarse_grid(d_sds **sds, int currentlyopen, int current_n_CMGwindows);
int32 get_set_sds(int getsetinit, d_sds *sdsinfo, int line, int n_lines);
void ids_zero_out(d_sds *sdsinfo, int n_lines);
int transfer_sds(d_sds **sds_outputlist);
int get_max_min(float *array, int start, int length, float *max, float *min);
int which_and_where(int32 sdss[2], int32 dimms[2][2], int32 sstart[2][2], 
                    int line, int n_lines, int32 n_col);

/* in qa.c */
void initqavalues(qavals_t *qavalues);
void count_quality(int32 *bands, dims_struct *output, 
		   qavals_t *qavalues, int32 process_sea, int32 process_clouds);
