/* @@ 016144909 /clear_vobs1/MODIS/STORE/PGE11/MOD_PR09/sra.c /main/1 Jennifer Davis (STX) 08-Jan-98.12:51:42 @@ */
/*
!C******************************************************************

!Description: 
		sra.c


!Input Parameters:
!Output Parameters: 
!Revision History:
!Team-unique Header:
!References and Credits 
!Design Notes


DESCRIPTION


!END*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* PGS seed files */

#include "PGS_MODIS_37092.h"
#include "mfhdf.h"
#include "modis_srefl.h"

#ifdef LINUX
#include <byteswap.h>
#endif

/* PGS I/O & SMF module *  /
#include "smfio.h"

#include "proc_define.h"
#include "proc_struct.h" */
#include "protos.h"


float32 algorithm_aot_tab[10]={0.05F, 0.15F, 0.3F, 0.5F, 0.7F, 0.9F, 1.1F, 1.4F, 1.7F, 2.0F};


int32 readsratab(float32 sra[7][10], PGSt_PC_Logical file)
{
/*
!C******************************************************************

!Description: The function reads the radiative transfer
tables containing atmospheric spherical albedo (10 optical depths)
seven bands. The status is return (0 success, -1 problem)

!Input Parameters:

sra[7][10]: the addresses of the spherical albedo tables.
file: file pointer (pgs type)

!Output Parameters:

readsratab:  0 success, -1 problem.


!Revision History:
$LOG: sra.c,v1.0$

Original version Fri Sep  8 16:10:32 EDT 1995
E. VERMOTE (eric@kratmos.gsfc.nasa.gov)


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758

!Design Notes


!END****************************************************************************
*/
	PGSt_IO_Gen_FileHandle *fd;
	int32 i,j;
	float32 aot;

	fd = OpenFileRead(file, (PGSt_integer)1);
	if (!fd) return -1;

	for (i=0;i<10;i++) {
		if (fscanf(fd,"%f %f %f %f %f %f %f %f",&aot,
			&sra[0][i],&sra[1][i],&sra[2][i],&sra[3][i],&sra[4][i],&sra[5][i],
			&sra[6][i])!=8) {

			for (j=0; j<7; j++) sra[j][i] = 0.0F;
			}
		}

	return CloseFile(fd);
}


int32 interSra(float32 *Sra_tab,float32 tau_a,float32 *S_ra)
{

/*
!C******************************************************************

!Description: The function performs interpolation at
a specific optical of the atmospheric spherical albedo table.
The status is return but is always 0 (success) in that version.

!Input Parameters:

sra_tab[10]: the addresses of the spherical albedo tables at a given
             wavelenght.

tau_a:  the input value of optical depth for which we want to compute
        atmospheric spherical albedo


!Output Parameters:

S_ra:  the value of atmospheric spherical albedo
interSra:  0 success.

!Revision History:
Revision 1.1, P. Fisher
1.  Allows extrapolation at extremely large optical depths.

Original version Fri Sep  8 16:10:32 EDT 1995
E. VERMOTE (eric@kratmos.gsfc.nasa.gov)


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits

!Design Notes


!END****************************************************************************
*/
	float32 c_aot;
	int32 iaot;

	iaot = index_aot_tab(tau_a);

	if(iaot<9){
	  c_aot = (tau_a - algorithm_aot_tab[iaot]) / (algorithm_aot_tab[iaot+1] - algorithm_aot_tab[iaot]);
          *S_ra = (1.0F - c_aot)*Sra_tab[iaot] + c_aot*Sra_tab[iaot+1];
	}else{
	  c_aot = (tau_a - algorithm_aot_tab[iaot]) / (algorithm_aot_tab[iaot] - algorithm_aot_tab[iaot-1]);
          *S_ra = (1.0F  + c_aot)*Sra_tab[iaot];
	}

        if ( *S_ra > 1.0000F) *S_ra = 1.0000F;

 /*       *S_ra = (1.0F - c_aot)*Sra_tab[iaot] + c_aot*Sra_tab[iaot+1]; */

	return 0;
}

int32 readaerotab(float32 *refl,PGSt_PC_Logical fostab)
{
/*
!C******************************************************************

!Description: This function reads in the table of pre-computed
atmospheric intrinsic reflectance.

!Input Parameters:
fostab:	  name of the file containing the atmospheric intrinsic
	  reflectance


!Output Parameters:
refl:    table of float32 containing the pre-computed top of the atmosphere
	  intrinsic reflectance.
readaerotab: int32 status (0: OK)

!Revision History:
$LOG: aerosols.c,v1.0$

Original version
E. VERMOTE (eric@kratmos.gsfc.nasa.gov)


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes


!END****************************************************************************
*/
	PGSt_IO_Gen_FileHandle *fd;
#ifdef LINUX
        unsigned int iskp, icnt, iptr, itmp;
#endif

	fd = OpenFileRead(fostab, (PGSt_integer)1);
	if (!fd) return -1;

	if (fread(refl,73*22*22*10*sizeof(float32),1,fd)!=1){
		modsmf(MODIS_E_FILE_READ_ERROR, "", "sra.c, readaerotab\0");
		return -1;
		}
#ifdef LINUX
        /* Swap every 4 bytes in refl so that the resulting byte order is 4, 3, 2, 1. */
        iskp = sizeof(float32);
        for ( icnt = 0 ; icnt < 73*22*22*10*iskp ; icnt = icnt + iskp ) { 
            iptr = (unsigned int) refl + icnt;
            itmp = bswap_32 ( * (int32*) iptr );
            * (int32*) iptr = itmp;
        }
#endif


	return CloseFile(fd);
}


int32 interrefl(float32 *refltab,float32 *reflval,float32 ts,
              float32 tv,float32 phi,float32 aot)
{
/*
!C******************************************************************

!Description: This function computes the atmospheric intrinsic
reflectance given geometrical conditions, optical depth by
 interpolating in a pre-computed table.

!Input Parameters:
refltab:   the table of float32 containing pre-computed atmospheric
intrinsic reflectances

ts:	  float32 value of the solar zenith angle
tv:	  float32 value of the view zenith angle
phi:	  float32 value of the relative azimuth angle

aot:	the aerosol optical depth (float32)


!Output Parameters:
interrefl:     the atmosphere intrinsic reflectance.



!Revision History:
        Revision 1.1, P. Fisher, SSAI
	1.  Modified to allow extrapolation at extremely high optical thicknesses.

	Original version
	E. VERMOTE (eric@kratmos.gsfc.nasa.gov)


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes


!END****************************************************************************
*/

   float32 c_ts,c_tv,c_phi,c_aot,refl1,refl2,frefl[2];
   int32 its,itv,iphi,iaot,tiaot,i,ind1,ind2,ind3,ind4;
   float32 ts0,tv0,phi0,step_ts,step_tv,step_phi;

   ts0=A_TS0;
   tv0=A_TV0;
   phi0=A_PHI0;
   step_ts=A_DELTS;
   step_tv=A_DELTV;
   step_phi=A_DELPHI;

   its=(ts-ts0)/step_ts;
   if (its<0) its=0;
   if (its>(A_NBTS-2)) its=A_NBTS-2;
   itv=(tv-tv0)/step_tv;
   if (itv<0) itv=0;
   if (itv>(A_NBTV-2)) itv=A_NBTV-2;
   iphi=(phi-phi0)/step_phi;
   if (iphi<0) iphi=0;
   if (iphi>(A_NBPHI-2)) iphi=A_NBPHI-2;

   iaot = index_aot_tab(aot);

   c_ts=(ts-ts0-its*step_ts)/step_ts;
   c_tv=(tv-tv0-itv*step_tv)/step_tv;
   c_phi=(phi-phi0-iphi*step_phi)/step_phi;
   if(iaot<9){
     c_aot=(aot-algorithm_aot_tab[iaot])/(algorithm_aot_tab[iaot+1]-algorithm_aot_tab[iaot]);
     tiaot=iaot;
   }else{
     c_aot=(aot-algorithm_aot_tab[iaot])/(algorithm_aot_tab[iaot]-algorithm_aot_tab[iaot-1]);
     tiaot=iaot-1;
   }

   for (i=tiaot; i<=tiaot+1; i++)
    {
    ind1=i+A_NBAOT*(itv+A_NBTV*(its+A_NBTS*iphi));
    ind2=i+A_NBAOT*(itv+A_NBTV*((its+1)+A_NBTS*iphi));
    ind3=i+A_NBAOT*((itv+1)+A_NBTV*(its+A_NBTS*iphi));
    ind4=i+A_NBAOT*((itv+1)+A_NBTV*((its+1)+A_NBTS*iphi));
    refl1=(1.0F-c_ts)*(1.0F-c_tv)*refltab[ind1]+
          c_ts*(1.0F-c_tv)*refltab[ind2]+
          (1.0F-c_ts)*c_tv*refltab[ind3]+
          c_ts*c_tv*refltab[ind4];
    ind1=i+A_NBAOT*(itv+A_NBTV*(its+A_NBTS*(iphi+1)));
    ind2=i+A_NBAOT*(itv+A_NBTV*((its+1)+A_NBTS*(iphi+1)));
    ind3=i+A_NBAOT*((itv+1)+A_NBTV*(its+A_NBTS*(iphi+1)));
    ind4=i+A_NBAOT*((itv+1)+A_NBTV*((its+1)+A_NBTS*(iphi+1)));
    refl2=(1.0F-c_ts)*(1.0F-c_tv)*refltab[ind1]+
          c_ts*(1.0F-c_tv)*refltab[ind2]+
          (1.0F-c_ts)*c_tv*refltab[ind3]+
          c_ts*c_tv*refltab[ind4];
    frefl[i-tiaot]=(1.0F-c_phi)*refl1+c_phi*refl2;
    }


   /* if(iaot<9) */
     *reflval=(1.0F-c_aot)*frefl[0]+c_aot*frefl[1];
   /* else
     *reflval=(1.0F+c_aot)*frefl[1]; */
   return 0;
   }

int32 readatranstab(float32 *ta_tab, PGSt_PC_Logical file)
{
/*
!C******************************************************************

!Description: This function reads in the table of pre-computed
atmospheric transmittance.

!Input Parameters:
file:	  name to the file containing the atmospheric intrinsic
	  transmittance


!Output Parameters:
ta_tab:    table of float32 containing the pre-computed top of the atmosphere
	  intrinsic transmittance.

readatranstab: int32 status (0: OK)

!Revision History:
$LOG: aerosols.c,v1.0$

Original version
E. VERMOTE (eric@kratmos.gsfc.nasa.gov)

Version 2.0, Paul Fisher, SSAI, November 1998
1. Modified to use defined constants to read the table vs. hard coded file sizes.


!Team-unique Header:

MODIS TEAM MEMBER Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	      
DEVELOPERS    Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	      Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	      Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758
	      Paul Fisher, fisher@ltpmail.gsfc.nasa.gov, 301-286-2980

!References and Credits:


!Design Notes:


!END****************************************************************************
*/
	PGSt_IO_Gen_FileHandle *fd;
#ifdef LINUX
        unsigned int iskp, icnt, iptr, itmp;
#endif

	fd = OpenFileRead(file, (PGSt_integer)1);
	if (!fd) return -1;

	if (fread(ta_tab,T_NBTHETA*T_NBAOT*sizeof(float32),1,fd)!=1){
		modsmf(MODIS_E_FILE_READ_ERROR, "", "sra.c, readatranstab\0");
		return -1;
		}
#ifdef LINUX
        /* Swap every 4 bytes in ta_tab so that the resulting byte order is 4, 3, 2, 1. */
        iskp = sizeof(float32);
        for ( icnt = 0 ; icnt < T_NBTHETA*T_NBAOT*iskp ; icnt = icnt + iskp ) {
            iptr = (unsigned int) ta_tab + icnt;
            itmp = bswap_32 ( * (int32*) iptr );
            * (int32*) iptr = itmp;
        }
#endif

	return CloseFile(fd);
}



int32 interatrans(float32 *table, float32 *t_a, float32 theta, float32 tau500)
/*
!C******************************************************************

!Description: This function interpolates the value of the atmospheric
transmssion tables to get the value of transmission at a particular
angle and optical depth.

!Input Parameters:
ta_tab:   the table of float32 containing pre-computed atmospheric
transmittances

theta:	  float32 value of the zenith angle (solar or view)

aot:	the aerosol optical depth (float32)


!Output Parameters:
t_a:     the atmosphere transmittance.
interatrans: int32 status (0:OK)


!Revision History:
        Revision 2.0, P. Fisher, SSAI
	1. Modified to use a "greater than" high class.

        Revision 1.1, P. Fisher, SSAI
	1.  Modified to allow extrapolation at extremely high optical thicknesses.

	Original version
	E. VERMOTE (eric@kratmos.gsfc.nasa.gov)


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
             Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	     Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	     Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758
	     Paul Fisher, fisher@ltpmail.gsfc.nasa.gov, 301-286-2980


!References and Credits


!Design Notes


!END****************************************************************************
*/{
  
  int i, index;
  int bin_theta, c_tab_aot;
  float c_index_aot, percent_theta;
  float ta[2];
  int ind1, ind2;
  float trup_a;

  index = index_aot_tab(tau500);
  
  bin_theta=(int)(theta/5.0F);
  percent_theta=(theta-(bin_theta*5.0F))/5.0F;

  if(index<9){
    c_index_aot=(tau500-algorithm_aot_tab[index])/(algorithm_aot_tab[index+1]-algorithm_aot_tab[index]);
    c_tab_aot=index;
  }else{
    c_index_aot=(tau500-algorithm_aot_tab[index])/(algorithm_aot_tab[index]-algorithm_aot_tab[index-1]);
    c_tab_aot=index-1;
  }

  if(bin_theta<T_NBTHETA-1){
    for (i=c_tab_aot;i<=c_tab_aot+1;i++){
      ind1=i+10*bin_theta;
      ind2=i+10*(bin_theta+1);
      ta[i-c_tab_aot]=(1-percent_theta)*table[ind1]+(percent_theta)*table[ind2];
    }
  }else{
    for (i=c_tab_aot;i<=c_tab_aot+1;i++)ta[i-c_tab_aot]=table[i+10*(T_NBTHETA-1)];
  }
  
  /* if(index<9) */
    *t_a=(1.0F-c_index_aot)*ta[0]+c_index_aot*ta[1];
  /* else
    *t_a=(1.0F+c_index_aot)*ta[1]; */
  
  return(0);
  
}

int32 index_aot_tab(float32 aot)
/*
!C*****************************************************************************
!Description:
	Return index to bracketing values in atmospheric extinction table.

!Input Parameters:

!Output Parameters:

!Revision History:
        Revision 1.1, P. Fisher, SSAI
	1.  Modified to allow extrapolation at extremely high optical thicknesses.

	Original version Sept. 1995, Eric Vermote, eric@kratmos.gsfc.nasa.gov
	Put into separate function Feb. 1996, L. Giglio (giglio@hades.gsfc.nasa.gov).

!Team-unique Header:
        References and Credits
        Design Notes
!END***************************************************************************
*/
{
	int32 i;

	for(i=0;i<10;i++)if(aot<=algorithm_aot_tab[i])break;

	if(i>0)return(i-1);
	return(i);

}
